/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */

import jasco.Jasco;
import jasco.util.CommandLine;
import jasco.util.CommandRunner;
import jasco.util.logging.Logger;

/*
 * Test comment, this can be removed
 */
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.TreeMap;
import java.util.Vector;

public class Main 
{

	public static void main(String [] args) throws Exception
	{	
		//jasco.util.Logger.setInstance(new jasco.util.ExceptionLogger());
		jasco.options.Options.loadOptions(false);
		
		CommandLine commandLine = new CommandLine(args,2);
		boolean correct = commandLine.parse();
		
		TreeMap flags = commandLine.getFlags();
		
		String version = (String) flags.get("version");
		if (version!=null)  {
			System.out.println(jasco.options.Options.getVersion());
			return;
		}
		
		String init = (String) flags.get("init");
		if (init!=null) {
			jasco.options.Options.generateScripts();
			return;
		}
		
		if (commandLine.getArgs().size() == 0) {
			Logger.getInstance().showError( "Command required" );
			return;
		}
		String command = (String) commandLine.getArgs().get(0);
		
		String classPath = (String)flags.get("classpath");
		String outputDir = (String)flags.get("outputdir");
		
		//default stuff: Generated
		if(classPath==null&&outputDir==null&&!jasco.options.Options.isContainedInClasspath("Generated"))
			classPath="Generated";
		
		//System.out.println("boeps classpath:"+classPath);
		
		if (classPath!=null)  {
			
			jasco.options.Options.addToClassPath(classPath);
			jasco.util.CommandRunner runner = new jasco.util.CommandRunner(false);
			String props=" -D\"jasco.installpath="+jasco.options.Options.getInstallPath()+"\"";
			props+=" -D\"jasco.debug="+jasco.options.Options.showDebugOutput()+"\"";
			String cmd = "java "+props+" -classpath \""+jasco.options.Options.getClassPath()+"\" Main";
			for(int i=0;i<args.length;++i)  
				if(!args[i].startsWith("-classpath:"))
					cmd+=" \""+args[i]+"\" ";
			//System.out.println(cmd);
			
			CommandRunner.runCommand(cmd,false);
			return;
		}
		//System.out.println(jasco.options.Options.getClassPath());
		
		String target = (String)flags.get("target");
		if(target!=null)
			jasco.options.Options.setTarget(target);
	
		
		jasco.Jasco aspect = new jasco.Jasco();
		
		if (command.equals("Introspect")) 
		{
			Jasco.createIntrospectorWindow();
			return;
		}
		
		boolean X = false;
		//System.out.println(flags.get("X"));
		if(flags.get("X")!=null)
			X=true;	
		
		
		boolean A = false;
		//System.out.println(flags.get("X"));
		if(flags.get("A")!=null)
			A=true;	
		boolean D = false;
		//System.out.println(flags.get("X"));
		if(flags.get("D")!=null)
			D=true;
		boolean noevents = false;
		//System.out.println(flags.get("X"));
		if(flags.get("noevents")!=null)
			noevents=true;	
		boolean k = false;
		//System.out.println(flags.get("X"));
		if(flags.get("k")!=null)  
			k=true;	
		if(flags.get("debug")!=null)
			jasco.options.Options.showDebugOutput(true);	
		
		if(k)
			jasco.options.Options.setDeleteTempFiles(false);
		if(noevents)
			jasco.options.Options.doTransformEvents(false);
		else jasco.options.Options.doTransformEvents(true);
		
		String file = "";
		if (!correct)  {
			printUsage(command);
			return;
		}
		else file=(String) commandLine.getArgs().elementAt(1);
		
		if(outputDir!=null)
			jasco.options.Options.setOutputDir(new File(outputDir).getCanonicalPath());	
		
			
		if(command.equals("CompileAspect"))  {
			jasco.options.Options.setTransformHooks(X);
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			}
			
			Jasco.compileAspect(file);
		}
		else if(command.equals("CompileRefinement"))  {
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			}
			
			Jasco.compileRefinement(file);
		}
		else if(command.equals("CompileConnector"))  {
			//jasco.options.Options.loadClass("jasco.connectorparser.Cutpoint");
			
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			} 
			
			if(X)
				jasco.options.Options.setTransformBean(true);	
			Jasco.compileConnector(file);
		}
		else if (command.equals("RemoveConnector")) 
		{
			Jasco.removeConnectorInOutputDir(file);
		}
		else if (command.equals("TransformBean"))  {
			jasco.options.Options.addSinglePathToClassPathFirst(jasco.options.Options.getOutputDir());
			Vector excludedmethods = new Vector();
			if (X) 
			{
				String value = (String)flags.get("X");
				String excludedmethodstring = value;
				excludedmethods = parseExcludedMethods(excludedmethodstring);
			}
			if(A)
				jasco.options.Options.setTransformAllMethods(true);
			
			if (flags.get("l")!=null)  {
				try {
				
				File filelist = new File(file);
				Logger.getInstance().showOutput("START TRANSFORIMG, please wait....");
				BufferedReader reader = new BufferedReader(new FileReader(file));
				String line = reader.readLine();
				while (line!=null)  {
					
					Jasco.transformBeanInOutputDir(line,excludedmethods);
					line=reader.readLine();
				}
				reader.close();
				Logger.getInstance().showOutput("TRANSFORIMG DONE");
				}catch(Exception ex)  {Logger.getInstance().showError("Exception occured while accessing file list:" +file);
				}
			}
			else if (D)   {
				Logger.getInstance().showOutput("Searching all java beans in "+jasco.options.Options.getOutputDir()+"....");
				
				Vector v = getListOfBeansInDir(new File(file));
				Logger.getInstance().showOutput("START TRANSFORIMG, please wait....");
				Iterator i = v.iterator();
				while (i.hasNext())  {
					String next = (String) i.next();
					//Logger.getInstance().showOutput("tranforming "+next);
					Jasco.transformBeanInOutputDir(next,excludedmethods);
				}
			}
			else {
			Jasco.transformBeanInOutputDir(file,excludedmethods);
			}
		}
		else if (command.equalsIgnoreCase("CompileTraversal"))  {
			
			if (!checkFile(file))  {
				Logger.getInstance().showError( "Non-existing file: "+file);
				return;
			}
			Jasco.compileTraversal(file);
		}
		else System.err.println("Unknown command! "+command);
 	
		
		
		/*
		
		if (args[0].equals("Introspect")) 
		{
			aspect.createIntrospectorWindow();
			return;
		}
		
		if (args.length<2)  {
			System.out.println("Usage: "+args[0]+" filename "+"[classpath] [outputdir]");
			return;
		}
		
		if(args.length>2)
			jasco.options.Options.setOutputDir(new File(args[2]).getCanonicalPath());
		
		if(args.length>3)
			jasco.options.Options.setTransformBean(true);
		
		
					
		if (args[0].equals("CompileAspect")) 
		{
			if (args.length>1&&!(new File(args[1]).exists()))  {
			System.err.println( "Not existing file: "+args[1]);
			return;
		}
			aspect.compileAspect(args[1]);
		}
		
		else if (args[0].equals("CompileConnector")) 
		{	
			if (args.length>1&&!(new File(args[1]).exists()))  {
			System.err.println( "Not existing file: "+args[1]);
			return;
		}
			aspect.compileConnector(args[1]);
		}
		
		else if (args[0].equals("RemoveConnector")) 
		{
			aspect.removeConnectorInOutputDir(args[1]);
		}
		
		else if (args[0].equals("TransformBean")) 
		{
			aspect.transformBeanInOutputDir(args[1]);
		}
		else if (args[0].equals("Introspect")) 
		{
			aspect.createIntrospectorWindow();
		}
		
		else System.err.println("Unknown command!");
		*/
		
		System.exit(0);
	}
	
	protected static Vector getListOfBeansInDir(File dir)  {
		
		if (!dir.exists())  {
			Logger.getInstance().showError("File "+dir.getAbsolutePath()+" does not exist!");
			return new Vector();
		}
		if (!dir.isDirectory())  {
			Logger.getInstance().showError("File "+dir.getAbsolutePath()+" is not a directory!");
			return new Vector();
		}
			
 		Vector res= new Vector();
		String[] files = dir.list();
		for(int i=0; i<files.length; ++i)  {
			File f = new File(dir,files[i]);
			
			if(!f.isDirectory())
				continue;
			
			res.addAll(getListOfBeansInDirAndPackage(f,files[i]));
		}
		return res;					
	}
	protected static Vector getListOfBeansInDirAndPackage(File dir, String p)  {
	
		Vector res= new Vector();
		String[] files = dir.list();
		for(int i=0; i<files.length; ++i)  {
			File f = new File(dir,files[i]);
			if (f.isDirectory())  
				res.addAll(getListOfBeansInDirAndPackage(f,p+"."+files[i]));
			else  {
				String classname = files[i];
				if(!classname.endsWith(".class"))
					continue;
				if(classname.indexOf("$")!=-1)
					continue;	
				int j = classname.lastIndexOf('.');
				if(j!=-1)
					classname=classname.substring(0,j);
				String cClass = p+"."+classname;
				res.add(cClass);	
			}
		}
		return res;
	}
	
	protected static Vector parseExcludedMethods(String theMethods) 
	{	
		
		Vector output = new Vector();
		StringTokenizer t = new StringTokenizer(theMethods,jasco.options.Options.PATH_SEPARATOR);
		while (t.hasMoreTokens())
		{
			String res=t.nextToken();
			
			output.add(res);
		}
		return output;
	}
	
	protected static boolean checkFile(String f)  
	{		
		return new File(f).exists();
	}
			
	public static void printUsage(String command)  {
	
		System.err.println();
		String filename = "filename";
		if(command.equalsIgnoreCase("transformBean"))
			filename="fullclassname";
		System.err.println("Usage: "+command+" [flags] "+filename+" ");
		System.err.println();
		System.err.println("Flags:");
		System.err.println("-classpath:");
		System.err.println("-outputdir:");
		System.err.println("-X:");
		System.err.println("-version:");
		System.err.println("-target:");
		if (command.equalsIgnoreCase("transformBean"))  {
			System.err.println("-A");
			System.err.println("-D");
			System.err.println("-noevents");
			
		}
		System.err.println();
		System.err.println("More info: http://ssel.vub.ac.be/jasco/tools.php");
	}
		
	
	
	

	
}
