/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.options;

import jasco.tools.BeanTransformerJavassist;
import jasco.util.logging.ErrorOutputLogger;
import jasco.util.logging.Logger;
import jasco.util.logging.SilentLogger;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;

public class Options  {
    
	public static final String version = "0.8.5.M3";
	
	private static final String[] systemProps = new String[]  {"os.name","os.arch","os.version","file.separator","path.separator"
	,"line.separator","user.name","user.home","user.dir","java","user","sun", "awt", "file"
	};
	
	public final static String FILE_SEPARATOR = System.getProperty("file.separator");
	public final static String PATH_SEPARATOR = System.getProperty("path.separator");
	public final static String NEWLINE = System.getProperty("line.separator");
	public final static String OS_NAME = System.getProperty("os.name");
	public final static String JAVA_VERSION_NAME = System.getProperty("java.version");
	
	public static final int HOOKS_MAX_VALUE = 10000;
	public static final int MAX_HOOKS_AT_ONE_JOINPOINT;
	
	public static final int JASCO_EMBEDDED = 0;
	public static final int JASCO_PC = 1;
	
	private static String outputdir = "Generated";
	private static String connectorPackage = "Connector";
	private static String installPath = "";
	private static Vector classPath = new Vector(); 
	private static boolean deleteTmpFiles = true;
	private static boolean transformBean = false;
	//private static DynamicClassLoader loader;
	public final static String ASPECT_FIELD = "_Jasco_cutpointconstructorRep";
	public static final String saveFile = "jasco.opt";
	public static String javaCompiler = "javac";
	public static boolean showDebug = false;
	private static boolean doExit = true;
	private static boolean throwExitException=false;
	private static boolean transformAllMethods = false;
	private static boolean transformEvents = false;
	private static boolean transformHooks = false;
	private static int jascoPlatform = JASCO_PC;
	private static int JAVA_VERSION_MAJOR=0;
	private static int JAVA_VERSION_MINOR=0;
	private static int priority = Thread.NORM_PRIORITY;
	private static String target = "1.5";
	
	
	private static boolean isJuttaEnabled = true;
	private static boolean doHotSwap = false;
	private static boolean juttaPredefinedCache = true;
	private static int juttacachesize = 10000;
	private static int loadInterval = 1000;
	private static String connectorLoadPath;
	private static int juttaserverport=8898;
	private static boolean optimizeLoadTime=false;
	private static ClassLoader classLoader= ClassLoader.getSystemClassLoader();
	public static File FINAL_TEMP_DIR = new File("."+File.separator+"_jasco_temp");
	
	private static final String[] libdirs = new String[]  {"jasco.jar", "jasco-libs.jar"
	};
	private static final String[] commands = new String[]  {"TransformBean","RemoveConnector",
		"CompileConnector", "CompileAspect", "Introspect", "CompileTraversal",  "CompileRefinement"
	};
	
	private static File tempDir;
	private static File logDir = new File("./_jasco_logs");
	
	private static boolean useReflectionForProceed = false;
	
	private static void initTempAndLogDir() {
		
		File tmpDir = new File(".");
		emptyFolder(FINAL_TEMP_DIR);
		if (!hotswap1_5Enabled())  {
			tmpDir=FINAL_TEMP_DIR;
			tmpDir.mkdir();
			tmpDir.deleteOnExit();
		} else {
			FINAL_TEMP_DIR.mkdir();
			FINAL_TEMP_DIR.deleteOnExit();
		}
		tempDir=tmpDir;
		
		emptyFolder(logDir);
		logDir.mkdirs();
		logDir.deleteOnExit();
	}
	
	protected static void emptyFolder(File dir)  {
		
			String[] files = dir.list();
			if(files==null) return;
			for(int i=0; i<files.length; ++i)  {
				File f = new File(dir,files[i]);
				if (f.isDirectory()) 
					emptyFolder(f);
				f.delete();
			}
			dir.delete();
		}
	
	public static void systemHalt() {
		
		
	}
	
	
		
	static  {
		
		initSafeClassPath();
		initTempAndLogDir();
		
		String tmp = "jasco.installpath";
		tmp=System.getProperty(tmp);
		if(tmp!=null && !tmp.equals(""))
			setInstallPath(tmp);
		
		tmp = "jasco.debug";
		tmp=System.getProperty(tmp);
		if(tmp!=null && new Boolean(tmp).booleanValue())
			showDebug=true;
		
		tmp = "jasco.output";
		tmp=System.getProperty(tmp);
		if(tmp!=null && !tmp.equals("")) 
			setOutputLevel(tmp);	
		
		tmp = "jasco.jutta";	
		tmp=System.getProperty(tmp);
		if(tmp!=null  && !new Boolean(tmp).booleanValue())
			setIsJuttaEnabled(false);
		
		tmp = "jasco.keep.tempfiles";	
		tmp=System.getProperty(tmp);
		if(tmp!=null  && new Boolean(tmp).booleanValue())
			setDeleteTempFiles(false);
		
		tmp = "jasco.jutta.predefinedcache";	
		tmp=System.getProperty(tmp);
		if(tmp!=null  && ! new Boolean(tmp).booleanValue())
			setJuttaPredefinedCache(false);	
		
		tmp = "jasco.connector.loadpath";	
		tmp=System.getProperty(tmp);
		if(tmp!=null  && !tmp.equals(""))
			connectorLoadPath=tmp;	
		else connectorLoadPath=System.getProperty("java.class.path");
		
		
		tmp = "jasco.hotswap.optimize.loadtime";	
		tmp=System.getProperty(tmp);
		if(tmp!=null  && !tmp.equals(""))
			optimizeLoadTime=new Boolean(tmp).booleanValue();
		else optimizeLoadTime=true;
		
		try {
		tmp = "jasco.jutta.cachesize";	
		tmp=System.getProperty(tmp);
		if(tmp!=null && !tmp.equals(""))
			setJuttaCacheSize(new Integer(tmp).intValue());	
		}catch(NumberFormatException ex)  {
			Logger.getInstance().showError("illegal jutta cache size: "+tmp);
		}
		
		try {
		tmp = "jasco.hotswap.tcpport";	
		tmp=System.getProperty(tmp);
		if(tmp!=null && !tmp.equals(""))
			juttaserverport(new Integer(tmp).intValue());	
		}catch(NumberFormatException ex)  {
			Logger.getInstance().showError("illegal hotswap server port: "+tmp);
		}
		
		
		int t = 20;
		try {
		
		tmp = "jasco.jutta.maxhooksperjoinpoint";	
		tmp=System.getProperty(tmp);
		if (tmp!=null && !tmp.equals(""))  
			t = new Integer(tmp).intValue();
		}catch(NumberFormatException ex)  {
			Logger.getInstance().showError("illegal max hooks per joinpoint: "+tmp);
		}
		finally  {
			MAX_HOOKS_AT_ONE_JOINPOINT=t;
		}
		
		try {
		tmp = "jasco.connector.loadinterval";	
		tmp=System.getProperty(tmp);
		if(tmp!=null && !tmp.equals("")) 
			setConnectorLoadInterval(new Integer(tmp).intValue());	
		}catch(NumberFormatException ex)  {
			Logger.getInstance().showError("illegal connector load interval: "+tmp);
		}
		
		setConnectorLoadingPriority();
		
		testOs();
		
		initJavaVersion();
	}
	
	public static void setTempDir(File dir) {
		
		tempDir=dir;
	}
	
	protected static void setConnectorLoadingPriority()  {
		
		String tmp = "jasco.connector.loadpriority";
		try {
			
		tmp=System.getProperty(tmp);
		if(tmp!=null  && !tmp.equals(""))
			priority=(new Integer(tmp).intValue());
		
		
		setConnectorLoadingPriority(priority);
		} catch(NumberFormatException ex)  {
		
			Logger.getInstance().showError("illegal connector load priority: "+tmp);
		}
	}
	
	protected static void setConnectorLoadingPriority(int priotiy)  {
	
		if (priority<Thread.MIN_PRIORITY||priority>Thread.MAX_PRIORITY)  {
			Logger.getInstance().showError("illegal connector load interval: "+ priority + " (" +Thread.MIN_PRIORITY+"< value < "+Thread.MAX_PRIORITY+")");
			priority=Thread.NORM_PRIORITY;
		}
	}
	
	public static void modifyConnectorLoadingPriority(int newPriority)  {
		
		setConnectorLoadingPriority(priority);
		jasco.runtime.ConnectorRegistry.restartLoadThread();
	}
	
	public static int getConnectorLoadingPriority()  {
	
		return priority;
	}
		
	
	protected static void setOutputLevel(String level)  {
	
		if(level.equals("silent"))
			Logger.setInstance(new SilentLogger());
		else if(level.equals("nowarnings"))
			Logger.setInstance(new ErrorOutputLogger());
		else if(level.equals("normal"))
			;
		else Logger.getInstance().showWarning("Unrecognized jasco.output level: "+level);
	}	
		
	
	public static void doDebugException(Exception ex)  {
	
		if (showDebugOutput())  {
			Logger.getInstance().showError("unexpected exception occured (use RuntimeExceptions) :"+ex.getMessage());
			ex.getStackTrace();
		}
		if(ex instanceof RuntimeException)
			throw (RuntimeException) ex;
		else throw new jasco.runtime.JAsCoAdviceException("Exception in advice:"+ex.getMessage(),ex);
	}
	
	public static boolean optimizeLoadTime()  {
		
		return optimizeLoadTime;
	}
	
	public static void juttaserverport(int i)  {
	
		juttaserverport=i;
	}
	
	public static int juttaserverport()  {
	
		return juttaserverport;
	}
	
	public static boolean hotswap1_5Enabled()  {
		
		try  {
		return jasco.runtime.hotswap.JAsCoHotSwap.hotswap1_5Enabled();
		} catch(Throwable t)  {
			return false;
		}
	}
	
	
	
	public static String getOptionsLaunchString()  {
		
		/*	
		String res = "";
		res+="-Djasco.jutta.cachesize="+getJuttaCacheSize()+" ";
		res+="-Djasco.jutta="+isJuttaEnabled()+" ";
		res+="-Djasco.connector.loadinterval="+getConnectorLoadInterval()+" ";
		res+="-Djasco.debug="+showDebugOutput()+" ";
		res+="-Djasco.connector.loadpath=\""+getConnectorLoadPath()+"\" ";
		res+="-Djasco.jutta.predefinedcache="+hasJuttaPredefinedCache()+" ";
		res+="-Djasco.hotswap.tcpport="+juttaserverport()+" ";
		res+="-Djasco.hotswap.keeptemp="+!deleteTempFiles()+" ";
		res+="-Djasco.hotswap.trapall="+transformAllMethods()+" ";
		
		return res;
		*/
		
		String res = "";
		Iterator i = System.getProperties().entrySet().iterator();
		outer:
		while (i.hasNext())  {
			Map.Entry next = (Map.Entry) i.next();
			String key = (String) next.getKey();
			
			for(int j=0;j<systemProps.length;++j)
				if (key.startsWith(systemProps[j]))  
					continue outer;
			
			
			String value = (String) next.getValue();
			res+="-D"+key+"=\""+value+"\" ";
		}
	//	System.out.println(res);
		return res;
	}
		
		
	
	public static boolean doTransformEvents()  {
	
		return transformEvents;
	}
	
	public static void doTransformEvents(boolean b)  {
	
		transformEvents=b;
	}
	
	public static boolean doHotSwapJDI()  {
	
		return doHotSwap;
	}
	
	public static void doHotSwapJDI(boolean b)  {
	
		doHotSwap=b;
	}
	
	public static String getConnectorLoadPath()  {
	
		return connectorLoadPath;
	}
	
	public static void setConnectorLoadPath(String path)  {
		
		connectorLoadPath=path;
	}
		
	
	public static int getConnectorLoadInterval()  {
	
		return loadInterval;
	}
	
	public static void setConnectorLoadInterval(int b)  {
		
		if(b>500||b==0||b==-1)
			loadInterval=b;
		else Logger.getInstance().showError("JASCO ERROR: Connector load interval cannot be smaller than 500ms, using defult now");	
	}
	
	
	public static boolean hasJuttaPredefinedCache()  {
	
		return juttaPredefinedCache;
	}
	
	public static void setJuttaPredefinedCache(boolean b)  {
	
		juttaPredefinedCache=b;
	}
	
	public static boolean isJuttaEnabled()  {
	
		return isJuttaEnabled;
	}
	
	public static File getTempDir()  {
		
		return tempDir;
	}
	
	public static File getFinalTempDir()  {
		
		return tempDir;
	}
	
	public static File getLogDir()  {
		
		return logDir;
	}
		
	
	public static void setIsJuttaEnabled(boolean b)  {
		
		 isJuttaEnabled=b;
	}
	
	public static int getJuttaCacheSize()  {
	
		return juttacachesize;
	}
	
	public static void setJuttaCacheSize(int b)  {
		
		juttacachesize=b;
	}
	
	protected static void initJavaVersion()  {
		try {
		int index1 = JAVA_VERSION_NAME.indexOf(".");
		if(index1==-1)
			return;
		
		String m = JAVA_VERSION_NAME.substring(0,index1);
		JAVA_VERSION_MAJOR = new Integer(m).intValue();
		
		
		m=JAVA_VERSION_NAME.substring(index1+1);
		int index2 = m.indexOf(".");
		if(index2==-1)
			return;
		String i = JAVA_VERSION_NAME.substring(index1+1,index1+1+index2);
		JAVA_VERSION_MINOR = new Integer(i).intValue();
		} 
		catch(Exception ex) {
			ex.printStackTrace();
		}
	}
	
	public static boolean isEmbeddedJasco()  {
	
		return getJAsCoPlatform()==JASCO_EMBEDDED;
	}
	
	public static int getJAsCoPlatform()  {
	
		return jascoPlatform;
	}
	
	public static boolean transformHooks()  {
	
		return transformHooks;
	}
	
	public static boolean isJavaAtLeastVersion(int major,int minor)  {
	
		return isAtLeastVersion(major,minor);
	}
	
	public static boolean isAtLeastVersion(int major,int minor)  {
	
		if(JAVA_VERSION_MAJOR>major)
			return true;
		
		else if(JAVA_VERSION_MAJOR==major)
			return JAVA_VERSION_MINOR >= minor;
		else return false;
	}
	
	public static void setTransformHooks(boolean b)  {
	
		transformHooks=b;
	}
	
	public static String getJavaCompiler()  {
		
		return javaCompiler;
	}
	
	public static Vector getCompileOptions() {
		
		Vector v = new Vector();
		v.add("-target");
		v.add(target);
		v.add("-source");
		v.add(target);
		//v.add("-nowarn");
		
		return v;
	}
	
	public static String getJavaVMTarget() {
		
		return target;
	}
	
	public static boolean isJavaVMTargetAtLeast(int major, int minor) {
		
		try{
			int index = target.indexOf('.');
			if(index<0)
				return false;
			String s = target.substring(0,index);
			int majorV = new Integer(s).intValue();
			
			if(majorV>major)
				return true;
			else if(major>majorV)
				return false;
			
			s = target.substring(index+1);
			int minorV = new Integer(s).intValue();
			
			if(minorV>=minor)
				return true;
			else return false;
		}
		catch(Exception x) {}
		return false;
	}
	
	public static void setJavaCompiler(String s)  {
	
		javaCompiler=s;
	}
	
	public static void setJAsCoPlatform(int version)  {
		
		if(version!=JASCO_EMBEDDED&&version!=JASCO_PC)
			Logger.getInstance().showError("Wrong JAsCo platform version: "+version);
		else jascoPlatform=version;
	}
	
	public static boolean transformAllMethods()  {
		
		//return true;
		return transformAllMethods;
	}
	
	public static void setTransformAllMethods(boolean b)  {
	
		transformAllMethods=b;
	}
	
	public static void exit(int exitcode)  {
		
		if(doExit) 
			System.exit(exitcode);
		else if(throwExitException)
			throw new JAsCoStopException();	
	}	
	
	public static boolean systemExitAllowed()  {
	
		return doExit;
	}
	
	public static boolean systemExitException()  {
	
		return throwExitException;
	}
	
	public static void setSystemExitAllowed(boolean v)  {
	
		doExit=v;
	}
	
	public static void setSystemExitException(boolean v)  {
	
		throwExitException=v;
	}
	
	public static boolean isWindows()  {
	
		return OS_NAME.toLowerCase().indexOf("windows")!=-1;
	}
	
	public static void testOs()  {
	
		if(OS_NAME.toLowerCase().indexOf("windows")==-1&&OS_NAME.toLowerCase().indexOf("linux")==-1)
			System.err.println("Warning: Unsupported OS detected, JAsCo might fail! ("+OS_NAME+")");
	}	
	
	public static String getVersion()  {
	
		return "JAsCo "+version+".EN";
	}
	
	public static void initSafeClassPath()  {
		
		classPath.removeAllElements();
		addToClassPath(".");
		StringTokenizer tokenizer = new StringTokenizer(System.getProperty("java.class.path"),PATH_SEPARATOR,false);
		while (tokenizer.hasMoreTokens())  {
		//	System.out.println("adding to loader: "+aPath);
			addToClassPath(tokenizer.nextElement().toString());
		}
		addToClassPath(getInstallPath());
		
	}
	
	public static boolean useReflectionForProceed() {
		
		return useReflectionForProceed;
	}
	
	public static void setUseReflectionForProceed(boolean b) {
		
		useReflectionForProceed=b;
	}
	
	public static boolean showDebugOutput()  {
	
		return showDebug;
	}
	
	public static void showDebugOutput(boolean b)  {
	
		showDebug=b;
	}
	
	public static void clearLoader()  {
		
	
	}
	
	public static ClassLoader createClassLoader() {
		
		URL[] urls =  new URL[getClassPathVector().size()];
		Iterator i = getClassPathVector().iterator();
		int j = 0;
		while(i.hasNext()) {
			try {
			urls[j]=new File(i.next().toString()).toURL();
			}catch(MalformedURLException ex) {Logger.getInstance().showError(ex);}
			++j;
		}
			

		URLClassLoader classloader = new URLClassLoader(urls,getClassLoader());
		return classloader;
	}
	
	public static Class loadClass(String name)  throws ClassNotFoundException {
		
		return createClassLoader().loadClass(name);
		/*
		//System.err.println("---------------------------------"+name);
		try {
		try {
		Class clz = getClassLoader().loadClass(name);
		return clz;
		}
		catch(ClassNotFoundException ex) {
			
		//ex.printStackTrace();
		}
		
		try {
		ClassPool pool = ClassPool.getDefault();
		pool.appendPathList(getClassPath()); 
		
		CtClass clz  = pool.get(name); 
		
		Class clazz = clz.toClass(getClassLoader());
		
		return clazz;
		}
		catch(NotFoundException ex) {}
		
		}
		catch(Exception ex) {System.err.println(ex.getMessage());ex.printStackTrace();}
		
		throw new  ClassNotFoundException(name);*/
	}
	
	public static boolean transformBean() {
	
		return transformBean;
	}
	
	public static boolean isContainedInClasspath(String aPath)  {
	
		return classPath.contains(aPath);
	}
	
	
	public static void addSinglePathToClassPathFirst(String aPath)  {
			
		aPath=aPath.replace('\\','/');
		classPath.insertElementAt(aPath,0);	
	}
	
	
	public static void addToClassPath(String aPath)
	{	
		StringTokenizer tokenizer = new StringTokenizer(aPath,PATH_SEPARATOR,false);
		while (tokenizer.hasMoreTokens())  {
			addSinglePathToClassPath((String)tokenizer.nextElement());
		}
	}
	
	public static void addSinglePathToClassPath(String aPath)  {
	
		aPath=aPath.replace('\\','/');
		if (!classPath.contains(aPath))  {
			classPath.addElement(aPath);
		}
	}
	
	protected static void setClassPathVector(Vector v)  {
		
		classPath=v;
	}
	
	public static void setClassPathVectorSafe(Vector v)  {
		
		initSafeClassPath();
		Iterator i = v.iterator();
		while (i.hasNext())  {
			String s = (String) i.next();
			addToClassPath(s);
		}
	}
		
	public static Vector getClassPathVector()  {
		
		return classPath;
	}
	
	public static void setClassPathSafe(String s)  {
		
		initSafeClassPath();
		StringTokenizer tokenizer = new StringTokenizer(s,PATH_SEPARATOR,false);
		while (tokenizer.hasMoreTokens())  {
		//	System.out.println("adding to loader: "+aPath);
			addToClassPath(tokenizer.nextElement().toString());
		}
	}
	
	public static String getClassPath()  {
		String res = "";
		Iterator i = classPath.iterator();
		while (i.hasNext())  {
			String s = (String) i.next();
			if(!s.equals("")&&!s.endsWith(PATH_SEPARATOR))
				res+=s+PATH_SEPARATOR;
		}
		return res;
	}
	
	public static String getRealHookClassName(String type)  {
		int index = type.lastIndexOf(".");
		String aspectType = type.substring(0,index)+"_"+type.substring(index+1);
		return aspectType;
	}	
	
	public static void setTransformBean(boolean b)  {
	
		transformBean=b;
	}
	
	public static String getBinDir()  {
	
		return installPath	+"/bin/";
	}
	
	public static ClassLoader getClassLoader() {
		
		return classLoader;
	}
	
	public static void setClassLoader(ClassLoader cl) {
		
		classLoader=cl;
	}
	
	public static String getOutputDir()  {
	
		return outputdir;
	}
	
	public static String getInstallPath()  {
	
		return installPath;
	}
	
	public static void setInstallPath(String path)  {
	
		
		if(path.endsWith("/")||path.endsWith("\\"))
			path=path.substring(0,path.length()-1);
		installPath=path;	
//		decompiler = installPath+"/libs/djava/D-Java";
		addToClassPath(path+"/jasco/classes/");
	}
	
	public static void setOutputDir(String dir)  {
	
		outputdir=dir;
	}
	
	public static String getConnectorPackage()  {
	
		return connectorPackage;
	}
	
	public static boolean deleteTempFiles()  {
	
		return deleteTmpFiles;
		//return false;
	}
	
	public static void setDeleteTempFiles(boolean b)  {
	
		deleteTmpFiles = b;
	}
	
	public static void generateBatchFiles()  {
	
		for(int i=0;i<commands.length;++i)
			generateBatchFile(commands[i]);
	}
	
	public static void generateScripts() {
		new File(getBinDir()).mkdirs();
		if (OS_NAME.equalsIgnoreCase("Linux"))
			for (int i = 0; i < commands.length; ++i)
				generateScript(commands[i]);
		else
			for (int i = 0; i < commands.length; ++i)
				generateBatchFile(commands[i]);
	}

	/**
	 * Return JAsCo libs, suitable for classpath string, with PATH_SEPARATOR at the end
	 */
	public static String getLibPaths() {
		String path = "";
		for (int i = 0; i < libdirs.length; ++i)
			path += installPath + FILE_SEPARATOR +  libdirs[i] + PATH_SEPARATOR;
		return path;
	}

	protected static void generateBatchFile(String name) {

		String impl = "";
		impl += "@ECHO OFF" + NEWLINE;
		impl += "Echo Working ..." + NEWLINE;
		impl += "java -Djasco.installpath=\"" + installPath + "\" -classpath \"." + PATH_SEPARATOR;
		
		impl += getLibPaths();
		
		impl += "%CLASSPATH%"+PATH_SEPARATOR;

		//impl += installPath+"/jasco/classes/";

		impl += "\" Main ";
		impl += name + " %1 %2 %3 %4 %5 %6 %7 %8 %9 " + NEWLINE;
		impl += "Echo Done" + NEWLINE;

		File file = new File(getBinDir(), name + ".bat");
		try {
			FileWriter writer = new FileWriter(file);
			writer.write(impl);
			writer.close();
		} catch (Exception ex) {
			System.err.println("Error during write of file:" + name);
			System.err.println(ex.getMessage());
		}
	}

	// GNU/*/Linux script
	protected static void generateScript(String name) {

		String impl = "";
		impl += "#!/bin/sh" + NEWLINE;
		impl += "echo Working ..." + NEWLINE;
		impl += "java -Djasco.installpath=\"" + installPath + "\" -classpath \"." + PATH_SEPARATOR;

		impl += getLibPaths();

		impl += "\"$CLASSPATH\""+PATH_SEPARATOR;

		impl += "\" Main ";
		impl += name + " $1 $2 $3 $4 $5 $6 $7 $8 $9 " + NEWLINE;
		impl += "echo Done" + NEWLINE;

		File file = new File(getBinDir(), name);
		String setPermissions = "chmod +x " + file.getPath();
		try {
			FileWriter writer = new FileWriter(file);
			writer.write(impl);
			writer.close();
			Runtime.getRuntime().exec(setPermissions);
			} catch (Exception ex) {
			System.err.println("Error during write of file:" + name);
			System.err.println(ex.getMessage());
		}
	}
	
	public static boolean isAspectBean(String aspectIdentifier)  
     { 
          boolean result = false; 
          
          java.lang.Class javaClass = null; 
          
          try   
          { 
               javaClass = Options.loadClass(aspectIdentifier); 
       
          
		  java.lang.reflect.Method m = javaClass.getDeclaredMethod("_Jasco_getHooks",new Class[0]);
          if(m!=null)
		  	result=true;
         
          }
		  catch(Exception e)  
          { 
               
			   result= false; 
          } 
          return result; 
     }
	 
	 public static boolean isAspectBean(javassist.CtClass clz)  
     { 
          boolean result = false; 
          
		  try {
          
		  javassist.CtMethod m = clz.getDeclaredMethod("_Jasco_getHooks");
          if(m!=null)
		  	result=true;
         
          }
		  catch(Exception e)  
          { 
               
			   result= false; 
          } 
          return result; 
     }
	 
	 public static boolean isTraversalConnector(javassist.CtClass clz)  
     { 
          boolean result = false; 
          
		  try {
          
		  javassist.CtClass[] m = clz.getInterfaces();
		  for(int i=0;i<m.length;++i)  
		  	if (m[i].getName().equals("jasco.runtime.connector.traversal.Traversal"))	  {
         		result=true;
				break;
		  	}
          }
		  catch(Exception e)  
          { 
               
			   result= false; 
          } 
          return result; 
     }
	 
	 
	 
	 public static boolean isAspect(javassist.CtClass clz)  
     { 
          boolean result = false; 
          //System.out.println(aspectIdentifier+result);
          
		  try {
          if (clz.getField(ASPECT_FIELD)!=null)   
          { 
               result = true; 
          } 
          else  
          { 
               result = false; 
          } 
		  //System.out.println(aspectIdentifier+result);
		  }
		 catch(Exception e)  
          { 
               
			   result= false; 
          } 
		  return result; 
     }
	
	public static boolean isAspect(String aspectIdentifier)  
     { 	  
		 
          boolean result = false; 
          //System.out.println(aspectIdentifier+result);
          java.lang.Class javaClass = null; 
          
          try   
          { 
               javaClass = Options.loadClass(aspectIdentifier); 
          } 
          catch(Exception e)  
          { 
               
			   return false; 
          } 
           
          if (hasAspectFields(javaClass))   
          { 
               result = true; 
          } 
          else  
          { 
               result = false; 
          } 
		  //System.out.println(aspectIdentifier+result);
          
		  return result; 
     }
	
	public static boolean definesNoAspects(javassist.CtClass javaClass) 
    { 
         try {
		 javassist.CtClass[] interfaces = javaClass.getInterfaces();
         for(int i=0;i<interfaces.length;++i)  {
	
		  	if(interfaces[i].getName().equals("jasco.runtime.NoAspects"))
				return true;
         }	
         }
		  catch(Exception ex) {
		  }
		  return false;	
    }
	 
	 public static boolean isCombinationStrategy(javassist.CtClass javaClass) 
     { 
          try {
		  javassist.CtClass[] interfaces = javaClass.getInterfaces();
          for(int i=0;i<interfaces.length;++i)  {
	
		  	if(interfaces[i].getName().equals("jasco.runtime.connector.CombinationStrategy"))
				return true;
          }	
          }
		  catch(Exception ex) {
		  }
		  return false;	
     }
	 
	 public static boolean isCombinationStrategy(String classIdentifier)  
     { 
          java.lang.Class javaClass = null; 
           
          try   
          { 
               javaClass = Options.loadClass(classIdentifier); 
          } 
          catch(Exception e)  
          { 
               return false; 
          } 
          
		  Class[] interfaces = javaClass.getInterfaces();
          for(int i=0;i<interfaces.length;++i)
		  	if(interfaces[i].getName().equals("jasco.runtime.connector.CombinationStrategy"))
				return true;
			
			
		  return false;	
     }
      
     private static boolean hasAspectFields(java.lang.Class theClass)  
     { 
          try { 
	
               theClass.getField(ASPECT_FIELD); 	  
               return true; 
          } 
          catch(NoSuchFieldException x)  { 
		  	
               return false; 
          } 
     } 
      
     public static boolean isClass(String name) 
     {      
     	try {
    		loadClass(name);
    		return true;
     	}
     	
     	catch(ClassNotFoundException ex) {}
     	
     	catch(Exception ex) {
     		Logger.getInstance().showError(ex);
     	}
    	
     	return false;
     }
	 
	 public static boolean isJAsCoBean(javassist.CtClass clz) 
     { 
	 	return BeanTransformerJavassist.isAJascoBean(clz);
     }
	 
	 public static boolean isJAsCoBean(String name) 
     { 	
	 	try {
	 	ClassPool pool = ClassPool.getDefault();
	 	CtClass clz = pool.get(name);
        return isJAsCoBean(clz);
	 	}catch(Exception x) {
	 		return false;
	 	}
     }
	 
	 public static void setTarget(String target2)  {
	 
	 	target=target2;
	 	
	 	//test target:
	 	try {
	 	int index = target.indexOf('.');
		if(index<0)
			throw new IllegalArgumentException("illegal VM target: "+target);
		String s = target.substring(0,index);
		int majorV = new Integer(s).intValue();
		s = target.substring(index+1);
		int minorV = new Integer(s).intValue();
	 	}catch(Exception ex) {
	 		throw new IllegalArgumentException("illegal VM target: "+target);	
	 	}
	 }
	 
	 public static void saveClearClassPath()  {
	 
	 	setClassPathVector(new Vector());
		saveOptions();
	 }
	 
	 public static void saveOptions()  {
	 	File file = new File(getInstallPath()+File.separator+saveFile);
		if(!file.exists())
			return;
			
	 	try {
			Properties properties = new Properties();
			properties.setProperty("jasco.classpath",getClassPath().replace('\\','/'));
			properties.setProperty("jasco.javacompiler",getJavaCompiler());
			properties.setProperty("jasco.platform",getJAsCoPlatform()+"");
	 		FileOutputStream s = new FileOutputStream(getInstallPath()+File.separator+saveFile);
			properties.store(s,null);
			s.close();
	 
	 	}
		catch(Exception ex)  {
		
			System.err.println("Error saving options: "+ex.getMessage());
			ex.printStackTrace();
		}
	 }
	 
	 public static void loadOptions()  {
	 
	 	loadOptions(true);
	 }
	 
	 public static void loadOptions(boolean loadClassPath)  {
	 
	 	try {
		
		Properties properties = new Properties();
		File file = new File(getInstallPath()+File.separator+saveFile);
		if(!file.exists())
			return;
	 	FileInputStream s = new FileInputStream(getInstallPath()+File.separator+saveFile);
		properties.load(s);
		String p = properties.getProperty("jasco.classpath");
		if(p!=null&&loadClassPath)
			setClassPathSafe(p);
		p = properties.getProperty("jasco.javacompiler");
		if (p!=null)  {
			javaCompiler = p;	
	 	}
		p = properties.getProperty("jasco.platform");
		if (p!=null)  {
			setJAsCoPlatform(new Integer(p).intValue());	
	 	}
	 	}
		catch(Exception ex)  {
		
			System.err.println("Error loading options: "+ex.getMessage());
			ex.printStackTrace();
		}
	 }
	
}
