/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime;

import jasco.runtime.connector.Connector;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.Hashtable;

public class ConnectorPersistance  {
	
	private static Hashtable connectorConfMap = new Hashtable();

	public static void saveConnector(Connector connector)  {
		
		if(connector.getFilename()==	null)
			return;
		try {
		
		File confFile =  getConfFile(connector);
		BufferedWriter  writer = new BufferedWriter(new FileWriter(confFile));
		writer.write(connector.isEnabled()+jasco.util.generators.JavaGenerator.NEWLINE);
		writer.write(connector.adaptOnClasses()+jasco.util.generators.JavaGenerator.NEWLINE);
		writer.close();
		
		Long timeClass=new Long(confFile.lastModified());
		connectorConfMap.put(connector,timeClass);
		
		ConnectorRegistry.fireConnectorRegistryChanged(new CRChangeEvent(CRChangeEvent.PROPERTY_CHANGE,connector));
		}
		catch(Exception ex)  {
			System.err.println("Error in saving connector "+connector.getName()+": "+ex.getMessage());
			ex.printStackTrace();
		}
	}
	
	public static File getConfFile(Connector connector)  {
	
		String filename = connector.getFilename();
		return new File(filename+".conf");
	} 
	
	public static void removeConfFile(File connfile)  {
		
		try {
		String filename = connfile.getCanonicalPath();
		File f = new File(filename+".conf");
		//System.out.println(filename);
		if(f.exists())
			f.delete();
		}
		catch(Exception ex)  {
			System.err.println("Error in deleting connector "+connfile.getName()+": "+ex.getMessage());
			ex.printStackTrace();
		}	
	}	
	
	public static boolean loadConnector(Connector connector)  {
		
		if(connector.getFilename()==null)
			return false;
		try {
		File confFile =  getConfFile(connector);
		if(!confFile.exists())
			return false;
		
		Long timeClass = (Long) connectorConfMap.get(connector);
		if (timeClass==null) 	 {//props, and before no props->change!
		
		}
		else  {
			long time = timeClass.longValue();
			
			if(time==confFile.lastModified()) //not yet adapted props, no change!
				return false;
		}
		
		BufferedReader  reader = new BufferedReader(new FileReader(confFile));
		Boolean b = new Boolean(reader.readLine());
		connector.setEnabledImpl(b.booleanValue());
		b = new Boolean(reader.readLine());
		connector.setAdaptOnClassesImpl(b.booleanValue());
		reader.close();
		
		timeClass=new Long(confFile.lastModified());
		connectorConfMap.put(connector,timeClass);
		
		return true;
		}
		catch(Exception ex)  {
			System.err.println("Error in loading connector "+connector.getName()+": "+ex.getMessage());
			ex.printStackTrace();
			return false;
		}
		
	}
}
	
		