/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime;
 
import jasco.options.Options;
import jasco.runtime.aspect.IHook;
import jasco.runtime.aspect.HookHelper;
import jasco.runtime.aspect.HookPropertyChangeEvent;
import jasco.runtime.aspect.StatefulAspect;
import jasco.runtime.cache.CombinedHooksFactory;
import jasco.runtime.connector.CombinationStrategy;
import jasco.runtime.connector.Connector;
import jasco.runtime.connector.HookElement;
import jasco.runtime.hotswap1.HotSwap;
import jasco.util.logging.Logger;

import java.io.File;
import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.TreeSet;
import java.util.Vector;

/*
 * 
 * @author Wim
 *
 * TODO To change the template for this generated type comment go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
public class ConnectorRegistry
{	
	private static Vector connectors = new Vector();
	
	private static Vector listeners = new Vector();
	
	private static Thread loadThread;
	
	private static Vector combinations = new Vector();
	private static long not = 0;
	private static boolean noConnectors = true;
	private static Vector cache=new Vector();
	private static Hashtable cachedItems=new Hashtable();
	private static Set threadContext = new HashSet();
	
	private ConnectorRegistry()  {
	}

	private static final CombinedHooks EMPTY_CONBINED_HOOKS = new CombinedHooks()  
		{public Object invoke(JascoMethod method) throws Exception  {return method.invokeOriginalJAsCoMethod();}
		public void init(Vector v)  {}};
	private static final Comparator CONNECTOR_COMPARATOR = new ConnectorComparator();
	
	public static void restartLoadThread()  {
	
		loadThread.stop();
		loadThread=new LoadThread();
		loadThread.start();
	}
	
	public static void quit()  {
		
		try  {
		loadThread.stop();
		
		while(connectors.size()>0)
			unRegisterConnector((Connector)connectors.elementAt(0),false);
		}catch(Exception ex)  {
			Logger.getInstance().showDebug("Exception occured during killAll: "+ex.getMessage());
		}
	}
	protected static void invalidateCache()  {
		
		if(Options.showDebugOutput())
			Logger.getInstance().showOutput("JASCO DEBUG: invalidating cache...");		
		
		//synchronized(cache) {
		Enumeration enu = cache.elements();
		while (enu.hasMoreElements())  {
			JascoMethod m = (JascoMethod) enu.nextElement();
			m.setDoNotCache(false);
			m.setCombinedHooks(null);
		}
		cache.removeAllElements();
		//}
		//synchronized(connectors) {
		Iterator i = connectors.iterator();
		while (i.hasNext())  {
			Connector m = (Connector) i.next();
			compCachable(m);
		}
		// }
	}
	
	protected static Object executeHooks(java.util.Vector hooks, JascoMethod originalMethod) throws java.lang.Exception {
	
		Object output = null;
		java.util.Hashtable hasher = new java.util.Hashtable();
		VectorSorter.sort(hooks,new CutpointReplaceComparator());
		
		Vector linkedHooks = new Vector();
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpointElement = (HookElement)hooks.elementAt(i);
			IHook cp = cutpointElement.getCutpoint();
			if(cutpointElement.getReplace().getExecution()) {
				linkedHooks.add(cutpointElement.getCutpoint());
			}
		}
		IHook[] linkedHooksArray = new IHook[linkedHooks.size()];
		linkedHooks.toArray(linkedHooksArray);
		Vector aroundThrowing = new Vector();
		Vector aroundReturning = new Vector();
		for(int i=0;i<linkedHooksArray.length;++i) {
			IHook cp = linkedHooksArray[i];
			if(HookHelper.definesAdvice(cp,IHook.AROUND_THROWING))
					aroundThrowing.add(cp);
			if(HookHelper.definesAdvice(cp,IHook.AROUND_RETURNING))
					aroundReturning.add(cp);
		}
		
		VectorSorter.sort(hooks,new CutpointBeforeComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			if(cutpoint.getBefore().getExecution()) cutpoint.getCutpoint().before(originalMethod,originalMethod.getCalledObject(),originalMethod.getArgumentsArray());
		}
		originalMethod.setLinkedHooks(linkedHooksArray);
		try {
		try {
		output = originalMethod.invokeJAsCoMethod();
		
		for (int i = aroundReturning.size()-1; i >=0 ; --i) {
			IHook cutpoint = (IHook)aroundReturning.elementAt(i);
			output = cutpoint.around_returning(output,originalMethod,originalMethod.getCalledObject(),originalMethod.getArgumentsArray());
		}
		VectorSorter.sort(hooks,new CutpointAfterComparator());
		for (int i = 0; i < hooks.size() ; ++i) {
			HookElement cutpoint = (HookElement)hooks.elementAt(i);
			if(cutpoint.getAfter().getExecution()) {
				cutpoint.getCutpoint().after_returning(output,originalMethod,originalMethod.getCalledObject(),originalMethod.getArgumentsArray());
			}
		}
		}
		catch(Exception ex)  {
			for (int i = 0; i< aroundThrowing.size(); ++i) {
				IHook cutpoint = (IHook)aroundThrowing.elementAt(i);
				if(i==(aroundThrowing.size()-1)) //last to execute, no catch anymore...
					output = cutpoint.around_throwing(ex,originalMethod,originalMethod.getCalledObject(),originalMethod.getArgumentsArray());
				
				else {
				try {
					output = cutpoint.around_throwing(ex,originalMethod,originalMethod.getCalledObject(),originalMethod.getArgumentsArray());
					break;
				}catch(Exception ex2) {ex=ex2;}
				}
			}
			if(aroundThrowing.size()==0)
				throw(ex);
		}
		}
		catch(Exception ex)  {
			VectorSorter.sort(hooks,new CutpointAfterComparator());
			for (int i = 0; i < hooks.size() ; ++i) {
				HookElement cutpoint = (HookElement)hooks.elementAt(i);
				if(cutpoint.getAfter().getExecution()) cutpoint.getCutpoint().after_throwing(ex,originalMethod,originalMethod.getCalledObject(),originalMethod.getArgumentsArray());
			}
			throw(ex);
		}
		finally {
			VectorSorter.sort(hooks,new CutpointAfterComparator());
			for (int i = 0; i < hooks.size() ; ++i) {
				HookElement cutpoint = (HookElement)hooks.elementAt(i);
				if(cutpoint.getAfter().getExecution()) {
					cutpoint.getCutpoint().after(originalMethod,originalMethod.getCalledObject(),originalMethod.getArgumentsArray());
				}
			}
		}
		return output;
	}
	
	
	protected static void addToCachedItems(Vector result,CombinedHooks cached)  {
	
		cachedItems.put(result,cached);
	}
	
	protected static CombinedHooks getCachedItem(Vector result)  {
		
		java.util.Enumeration hasherI = cachedItems.keys();
		while(hasherI.hasMoreElements()) {
			Vector res = (Vector) hasherI.nextElement();
			boolean iss = true;
			if(result.size()!=res.size())
				continue;
				
			Enumeration enu = res.elements();
			Enumeration enum2 = result.elements();
			while (enu.hasMoreElements())  {
				if (enu.nextElement()!=enum2.nextElement())  {
					iss=false;
					break;
				}
			}
			if(iss)
				return (CombinedHooks) cachedItems.get(result);
		}
		return null;
	}
	
	protected static Enumeration allHooks()  {
		
		Vector res = new Vector();
		synchronized(connectors)  {
		
			Iterator i = connectors.iterator();
			while (i.hasNext())  {
				Connector c = (Connector) i.next();
				Enumeration enu = c.getHooks();
				while (enu.hasMoreElements())  {
					res.addElement(enu.nextElement());
				}
			}
		}
		return res.elements();
	}
	
	public static Object executeTrap(String aMethod, Object anInstance,JascoMethod aJascoMethod)  throws Exception {
		
		if (Options.isJuttaEnabled())  {
		
			//CombinedHooks combinedAspects = (CombinedHooks) cache.get(aMethod);
			CombinedHooks combinedAspects = aJascoMethod.getCombinedHooks();
			if (combinedAspects==null)  {
				
				if (aJascoMethod.doNotCache())  {
					Vector result = findCutpointForMethod(aMethod,anInstance,aJascoMethod,false);
					logJP(aMethod,result);
					return executeHooks(result,aJascoMethod);
				}
				Vector result = findCutpointForMethod(aMethod,anInstance,aJascoMethod,true);
				logJP(aMethod,result);
				if (!isCachable(result))  {
				
					cache.add(aJascoMethod); 
					aJascoMethod.setDoNotCache(true);
					result=findCutpointForMethod(aMethod,anInstance,aJascoMethod,false);
					return executeHooks(result,aJascoMethod);
				} 		
				else if (result.size()==0)  {
					combinedAspects=EMPTY_CONBINED_HOOKS;
					cache.add(aJascoMethod); 
					aJascoMethod.setCombinedHooks(combinedAspects);
				}
				else  {	
					if (jasco.runtime.inline.InlineCompiler.getDefault().inlineCompile(result,aJascoMethod))  {
						result=findCutpointForMethod(aMethod,anInstance,aJascoMethod,false);
						return executeHooks(result,aJascoMethod);
					}
					else  {
						combinedAspects=getCachedItem(result);
						if (combinedAspects==null)  {
						    try {
							  
						      combinedAspects= CombinedHooksFactory.getDefault().generateCachedAspectJavassist(result,aJascoMethod); 
							  addToCachedItems(result,combinedAspects);
					    	}catch(Exception ex)  {
								System.err.println(ex.getMessage());
								ex.printStackTrace();
							    Logger.getInstance().showError("JAsCo SYSTEM ERROR: Caching aspects failed, executing the old fashioned way...");
							    return executeHooks(result,aJascoMethod);
						    }
						}
					
						//cache.put(aMethod,combinedAspects);
						cache.add(aJascoMethod); 
						aJascoMethod.setCombinedHooks(combinedAspects);
					}
				}
			}
			
		    return combinedAspects.invoke(aJascoMethod);
		}
		else  {
			Vector result = findCutpointForMethod(aMethod,anInstance,aJascoMethod);
			logJP(aMethod,result);
			return executeHooks(result,aJascoMethod);
		}
	}
	
	/**
	 * @param result
	 */
	private static void logJP(String aMethod, Vector result) {
		if (jasco.options.Options.showDebugOutput())  {
			System.out.println("JASCO DEBUG: joinpoint encountered "+aMethod);
			String test = "JASCO DEBUG: matching hooks = ";
			Enumeration j = result.elements();
			if(!j.hasMoreElements())
				test+= "NONE";
			while (j.hasMoreElements())   {
				Object next = j.nextElement();
				if (next instanceof HookElement)  {
					HookElement cp = (HookElement) next;
					test+=cp.getCutpoint().getClass().getName()+cp.getBefore().getExecutionNumber()+" ";
				}
				else test+= j.nextElement().getClass().getName().toString()+" ";
			}
			System.out.println(test);	
		}
	}

	//fails to soon, better test has to be implemented...
	protected static boolean isCachable(Vector hooks)  {
		Enumeration enu = hooks.elements();
		while (enu.hasMoreElements())  {
			HookElement cpe = (HookElement) enu.nextElement();
			IHook cp = cpe.getCutpoint();
			if(cp._Jasco_getConnector().doNotCache())
				return false;
		}
		return true;
	}
	
	protected static void compCachable(Connector c)  {
		
		if (!c.adaptOnClasses())  {
			Logger.getInstance().showWarning("connector "+c.getName()+" is not cachable (non-static), Jutta system is not activated");
			c.setDoNotCache(true);
			return;
		}
		
		
		Enumeration hooks = c.getHooks();
		while (hooks.hasMoreElements())  {
			HookElement cpe = (HookElement) hooks.nextElement();
			IHook cp = cpe.getCutpoint();
			if(cp instanceof StatefulAspect||cp instanceof DoNotCache)
				if (c.getCombinationStrategies().hasMoreElements())  {
					Logger.getInstance().showWarning("hook "+cp.getClass().getName()+" is not cachable (dynamic condition in combination with combination strategy), Jutta system is not activated for connector "+c.getName());
					c.setDoNotCache(true);
					return;
				}
			
			if (cpe.doNotCache())  {
				Logger.getInstance().showWarning("hook "+cp.getClass().getName()+" is not cachable, Jutta system is not activated for connector "+c.getName());
				c.setDoNotCache(true);
				return;
			}
			
		}
		
		Enumeration combStrats = c.getCombinationStrategies();
		while (combStrats.hasMoreElements())  {
			CombinationStrategy cp = (CombinationStrategy) combStrats.nextElement();
			if (cp instanceof DoNotCache)  {
				Logger.getInstance().showWarning("combination strategy "+cp.getClass().getName()+" is not cachable, Jutta system is not activated for connector "+c.getName());
				c.setDoNotCache(true);
				return;
			}
		}
		
		c.setDoNotCache(false);
	} 
		
	public static Vector findCutpointForMethod(String aMethod, Object anInstance,MethodJoinpoint aJascoMethod)  {
		return  findCutpointForMethod(aMethod,anInstance,aJascoMethod,false);
	}
				

	public static Vector findCutpointForMethod(String aMethod, Object anInstance,MethodJoinpoint aJascoMethod, boolean forJutta) 
	{	
	
		
		synchronized(connectors)  {
		
		Vector result = new Vector();
		if(connectors.size()==0)
			return result;
		
		RuntimeContext context = new RuntimeContext(aMethod,aJascoMethod,aJascoMethod.isExecution());
		RuntimeContext context2 = new RuntimeContext(aMethod,aJascoMethod,aJascoMethod.isExecution());
		
		Vector connectorsCopy = getConnectorsVector();
		Collections.sort(connectorsCopy,CONNECTOR_COMPARATOR);
		
		Enumeration i = combinations.elements();
		while (i.hasMoreElements())  {
			ConnectorCombinationStrategy c = (ConnectorCombinationStrategy) i.nextElement();	
			connectorsCopy=c.validateCombinations(connectorsCopy,context2);
		}
		
		int startID = 0;
		i = connectorsCopy.elements();
		while (i.hasMoreElements()) 	
		{   
			Connector connector = (Connector) i.nextElement();	
			//System.out.println(connector.getName());
			Vector tmp = connector.findCutpointForMethod(anInstance,context,aJascoMethod,startID,forJutta);	
			//System.out.println(connector.getName()+tmp.size());
			if(tmp!=null)
				addAll(result,tmp);
			startID+=jasco.options.Options.HOOKS_MAX_VALUE+1;	
		}
		
		return result;
		}
	}
	
	protected static void addAll(Vector v, Vector toAdd)  {
	
		Enumeration e = toAdd.elements();
		while(e.hasMoreElements())
			v.addElement(e.nextElement());
	}	
	
	public static void addConnectorCombinationStrategy(ConnectorCombinationStrategy c)  {
	
		combinations.addElement(c);
		fireConnectorRegistryChanged(new CRChangeEvent(CRChangeEvent.COMBSTRAT_ADDED,c));
		
		if(c instanceof DoNotCache) {
			Options.setIsJuttaEnabled(false);
			Logger.getInstance().showOutput("JAsCo Warning: non cachable connector combination strategy detected, Jutta system is disabled...");
		}
	}
	
	public static void removeConnectorCombinationStrategy(ConnectorCombinationStrategy c)  {
	
		combinations.removeElement(c);
		fireConnectorRegistryChanged(new CRChangeEvent(CRChangeEvent.COMBSTRAT_REMOVED,c));
	}
		
	
	public static void addConnectorRegistryListener(ConnectorRegistryListener l)  {
		synchronized(listeners)  {
		listeners.addElement(l);
		}
	}
	
	public static void removeConnectorRegistryListener(ConnectorRegistryListener l)  {
	    synchronized(listeners)  {
		listeners.removeElement(l); 
	    }
	}
	
	protected static void fireConnectorRegistryChanged(CRChangeEvent event)  {
		
		synchronized(listeners)  {
	
		Enumeration i = listeners.elements();
		while (i.hasMoreElements())  {
		
			ConnectorRegistryListener l = (ConnectorRegistryListener) i.nextElement();
			l.registryChanged(event);
		}
		}
		
		if(Options.isJuttaEnabled())
			invalidateCache();
	}
			
	
	public static void registerConnector(Connector c) 
	{	
		synchronized(connectors)  {
		if (!connectors.contains(c))  {
			
			connectors.add(c);
			noConnectors=false;
			fireConnectorRegistryChanged(new CRChangeEvent(CRChangeEvent.CONNECTOR_ADDED,c));
			if(Options.doHotSwapJDI())
				HotSwap.hotswaptargetClasses(c);
		}	
		}
	}
	
	public static void unRegisterConnector(Connector c)  {
	
		unRegisterConnector(c,true);
	}
	
	public static void unRegisterConnector(Connector c, boolean notifyAspects) 
	{	
		synchronized(connectors)  {
		connectors.remove(c);	
		fireConnectorRegistryChanged(new CRChangeEvent(CRChangeEvent.CONNECTOR_REMOVED,c));
		if(connectors.size()==0)
			noConnectors=true;
		if(Options.doHotSwapJDI())
				HotSwap.unhotswaptargetClasses(c);	
		
		if(notifyAspects)
			c.firePropertyChangeEvent(new HookPropertyChangeEvent(HookPropertyChangeEvent.IS_REMOVED));
		}
	}
	
	public static void start()  {
	}
	
	private static final File[] classpathArray ;
	
	private static File[] initClasspath()  {
		
		Set res = new TreeSet();
		StringTokenizer tokenizer = new StringTokenizer(Options.getConnectorLoadPath(),Options.PATH_SEPARATOR);
		while (tokenizer.hasMoreTokens()) 
		{
			String token = tokenizer.nextToken();
			File file = new File(token);
			if (file.isDirectory())   {
				//System.out.println("adding ... "+file.getPath());
				try  {
				res.add(file.getCanonicalFile());
				}catch(IOException ex)  {
				}
			}
		}
		File[] f = new File[res.size()];
		res.toArray(f);
		return f;
	}
	
	public static void forceConnectorLoad()  {
	
		if(loadThread!=null)
			loadThread.interrupt();
		else load();	
	}	
	
	static 
	{	
		classpathArray=initClasspath();
		if (Options.isJuttaEnabled())  {
			cache = new Vector(Options.getJuttaCacheSize());
			if(!Options.doHotSwapJDI())
			
			Logger.getInstance().showOutput("--- JAsCo Jutta system activated ---");
		}
		
		else if(!Options.doHotSwapJDI()) 
			Logger.getInstance().showOutput("--- JAsCo Jutta system NOT activated ---");
		
		if (jasco.options.Options.getConnectorLoadInterval()>-1)  {
			ConnectorRegistry.load();
			ConnectorRegistry.checkPersistance();
		}
		
		if (jasco.options.Options.getConnectorLoadInterval()>0)  {
			loadThread=new LoadThread();
			loadThread.start();
		}
		else Logger.getInstance().showWarning("Automatic Connector loading disabled");
		
		if (jasco.options.Options.showDebugOutput())  {
			Logger.getInstance().showOutput("JASCO DEBUG: System initialized");
		}
	}
	
	public static void pauseRefreshing()  {
		
		loadThread.suspend();
	}
	
	public static void resumeRefreshing()  {
		
		loadThread.resume();
	}
	
	protected static HashMap getConnectorsCopy()  {
		
		synchronized(connectors)  {
		HashMap v = new HashMap();
		Iterator i = connectors.iterator();
		while (i.hasNext())  {
			Connector c = (Connector) i.next();
			
			/*if(Options.showDebugOutput())
				System.out.println("JASCO DEBUG: loaded connector: "+c.getName()+";");*/
			v.put(c.getName(),c);
		}
		//System.out.println("---");
		return v;
		}
	}
	
	public synchronized static Iterator getConnectors()  {
	
		return connectors.iterator();
	}
	
	public synchronized static Vector getConnectorsVector()  {
		
			Vector v =  (Vector) connectors.clone();
			
			return v;
		}
	
	protected static void debug()  {	
		
	}
	
	public static  boolean isAdapted(JascoMethod method)  {
		
		return  !noConnectors && method.getCombinedHooks()!=EMPTY_CONBINED_HOOKS;	
	}
	
	public static  boolean isAdaptedAndThreadFlag(JascoMethod method)  {
		
		boolean b =   !noConnectors && method.getCombinedHooks()!=EMPTY_CONBINED_HOOKS&&!threadflag;
		
		return b; 
	}
	
	public static  void setThreadFlag()  {
		
		threadflag=true;
		
		//threadContext.add(Thread.currentThread());
	}
	static boolean threadflag = false;
	
	public static  void releaseThreadFlag()  {
		
	
		threadflag=false;
		//threadContext.remove(Thread.currentThread().getName());
		
		//System.out.println("RELEASING");
	}
	
	protected static void load()  {
		
		//debug();
		//System.out.println("number of traps occured: ");
		
		Map connectorsCopy = getConnectorsCopy();
		Class[] tmp = new Class[]  {"".getClass()};
		
		for(int ji=0;ji<classpathArray.length;++ji)  {
			File file = classpathArray[ji];
			File connectorpath = new File(file,"Connector");
			if (connectorpath.exists() && connectorpath.isDirectory()) 
			{
				String[] files = connectorpath.list();
				for(int i=0; i<files.length; ++i)
					try
					{
					File tfile = new File(connectorpath,files[i]);
					String classname = tfile.getName();
					if(!classname.endsWith(".class"))
						continue;
					int j = classname.indexOf('.');
					if(j!=-1)
						classname=classname.substring(0,j);
					String cClass = "Connector."+classname;	
					
					
					if(connectorsCopy.remove(classname)==null)  { //not yet loaded
						Class clz = jasco.options.Options.loadClass(cClass);
						if(!Connector.class.isAssignableFrom(clz))
							continue;
						if (Options.showDebugOutput())  
							System.out.println("JASCO DEBUG: new connector found! "+classname+";");
						clz.getMethod("load",tmp).invoke(null,new Object[]  {tfile.getAbsolutePath()});
					}
					}
					catch(ExceptionInInitializerError e)  {
						
						Logger.getInstance().showDebug(e);
					}
					catch(NoSuchMethodException ex)  {
						
						Logger.getInstance().showDebug(ex);
					}
					catch(ClassFormatError error)  {
						Logger.getInstance().showDebug(error);
					}
					catch(Throwable ex) 
					{
						Logger.getInstance().showDebug(ex);
					}
			}
		}
		
		//all connectors still present in connector have been removed
		Iterator j = connectorsCopy.values().iterator();
		while (j.hasNext())  {
			Connector c = (Connector) j.next();
			unRegisterConnector(c);
		}
		
		synchronized(listeners)  {
	
		Enumeration i = listeners.elements();
		while (i.hasMoreElements())  {
		
			ConnectorRegistryListener l = (ConnectorRegistryListener) i.nextElement();
			l.registryUpdated();
		}
		}
	}
	
	public static void checkPersistance()  {
		
		synchronized(connectors)  {
		Iterator i = connectors.iterator();
		boolean change = false;
		while (i.hasNext()) 	
		{
			Connector connector = (Connector) i.next();	
			if(ConnectorPersistance.loadConnector(connector))
				change=true;
			
			if(change) 
				fireConnectorRegistryChanged(new CRChangeEvent(CRChangeEvent.PROPERTY_CHANGE,connector));	
		}
		
		}	
	}
	
	protected static void notifyAspects(Connector c, HookPropertyChangeEvent event)  {
	}
	
		
	public static void reOrderConnectors() {
		
		//clear jutta cache so that new order is employed...
		invalidateCache();
		fireConnectorRegistryChanged(new CRChangeEvent(CRChangeEvent.ALLCHANGE));	
	}
			
}

class ConnectorComparator implements Comparator {
	
	public int compare(Object o1, Object o2) {
		
		if(o1 instanceof Connector && o2 instanceof Connector) {
			
			Connector c1 = (Connector) o1;
			Connector c2 = (Connector) o2;
			
			if(c1.getPriority()<c2.getPriority())
				return 1;
			else if(c2.getPriority()<c1.getPriority())
				return -1;
			else return 0;
		}
		else return -1;
	}
}

class LoadThread extends Thread  {
	
	ThreadGroup parent;
	
	LoadThread()  {
		ThreadGroup group = getThreadGroup();
		while(group.getParent()!=null)
			group=group.getParent();
		parent=group;
		setDaemon(true);	
	}
	
	public void run()  {
		
		if (jasco.options.Options.showDebugOutput())  {
			System.out.println("JASCO DEBUG: jasco aspect framework operational");
		}
		setPriority(Options.getConnectorLoadingPriority());
		int sleeptime = jasco.options.Options.getConnectorLoadInterval();
		
		while (true)  {
			
			try {
			sleep(sleeptime);
			}catch(Exception ex)  {
			}
			ConnectorRegistry.load();
			ConnectorRegistry.checkPersistance();
			
			/*
			System.out.println(parent.activeCount());
			Thread[] t = new Thread[parent.activeCount()];
			parent.enumerate(t);
			for(int i=0;i<t.length;++i)
				System.out.println(t[i]);
			if(parent.activeCount()==1)
				break;*/
		}
	}

}
