/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.aspect;

import jasco.runtime.JascoMethod;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.RuntimeContext;

import java.lang.annotation.Annotation;
import java.util.Hashtable;
import java.util.Vector;

/**
 * Abstract super class for all pointcut elements in hook constructors.
 */
public abstract class PCutpointConstructorApplicationDesignator
{	
	public final static String NEWLINE = System.getProperty("line.separator");
    public final static String TAB = "\t";

	private static int index;
	private PCutpointConstructorApplicationDesignator param;
	private PCutpointConstructorApplicationDesignator param2;
	
	public PCutpointConstructorApplicationDesignator(PCutpointConstructorApplicationDesignator param)  {
	
		this.param = param;
	
	}
	
	public PCutpointConstructorApplicationDesignator(PCutpointConstructorApplicationDesignator param, PCutpointConstructorApplicationDesignator param2)  {
	
		this.param = param;
		this.param2 = param2;
	}
	
	public PCutpointConstructorApplicationDesignator getParameter()  {
	
		return param;
	}
	
	public PCutpointConstructorApplicationDesignator getParameter2()  {
	
		return param2;
	}
	
	public boolean dontLoadClasses(MethodJoinpoint jp) {
		
		return jp instanceof JascoMethod &&((JascoMethod)jp).dontLoadClass();
	}
	
	/**
	 * 
	 * Evaluates to true when the pointcut triggers
	 * 
	 * @param context runtime context available
	 * @param mappings from variables to concrete method signatures, fe: 1 -> * *(*), 2-> void test.main.Main.run()
	 * @param forJutta is this for analyzing jutta, meaning is this runtimecontext a static context. Fe methods in cflow might change, target might change. The current method does not change however.
	 * @return
	 */
	public abstract boolean eval(RuntimeContext context,Hashtable mappings,boolean forJutta);
	
	public abstract String generateCreateCode(String argname);
	
	public String generateCreateCode()  {
	
		return generateCreateCode(getNextArgName());
	}
	
	public boolean checkAnnotation(String signature, RuntimeContext context) {
		
		if(!signature.startsWith("@"))
			return true;
		
		int index = signature.indexOf(" ");
		if(index!=-1) {
			String annotationString = signature.substring(1,index);
			Annotation[] annotations = context.getAnnotations();
			
			for(int i=0;i<annotations.length;++i)
				if(annotations[i].annotationType().getName().equals(annotationString)) {
					//System.out.println(annotations[i].annotationType().getName());
					return true;
				}
		}
		 
		return false;
	}
	
	public String removeAnnotation(String signature) {
		
		if(!signature.startsWith("@"))
			return signature;
		
		int index = signature.indexOf(" ");
		if(index!=-1) {
			return signature.substring(index+1);
		}
		else return signature;
	}
	
	public String toString()  {
	
		return generateCreateCode();
	}
	
	public void resetArgNames()  {
		index=0;
	}
	
	protected String getNextArgName()  {
	
		return "arg"+index++;
	}
	
	protected static boolean isRegularExp(String s)  {
	
		return s.indexOf('*') != -1;
	}
	
	public abstract String toDescString();
	
	public boolean doNotCache()  {
		
		boolean res = false;
		if(param!=null)
			res=res||param.doNotCache();
		if(param2!=null)
			res=res||param2.doNotCache();	
		return res;
	}
	
	public Vector allPointcuts()  {
		Vector res = new Vector();
		if(getParameter()!=null)
			res.add(getParameter());
		if(getParameter2()!=null)
			res.add(getParameter2());
		return res;
	}	
	
}