/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.connector;

//import jasco.aspectparser.*;
import jasco.runtime.DoNotCache;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.RuntimeContext;
import jasco.runtime.aspect.IHook;
import jasco.runtime.aspect.factory.HookInstantiationFailed;
import jasco.runtime.aspect.factory.IAspectFactory;
import jasco.runtime.aspect.factory.IAspectInstanceCreator;
import jasco.runtime.refinements.RefinementManager;

import java.lang.reflect.Method;
import java.util.Enumeration;

public class DynamicHookElement extends HookElement implements IAspectInstanceCreator  {
	
	private MethodJoinpoint currentMethod = null;
	private IAspectFactory factory = null;
	
	public static final char INITIALIZER = '$';
	
	public DynamicHookElement(IHook aCutpoint, CutpointExecutionElement aBefore, CutpointExecutionElement aAfter, CutpointExecutionElement aReplace, String aName) 
	{
		super(aCutpoint,aBefore,aAfter,aReplace,aName);
	}
	 
	public void setAspectFactory(IAspectFactory c)  {
		factory=c;
	}

	public boolean doNotCache() {
		
		return factory instanceof DoNotCache;
	}
	
	public IHook getCutpoint() 
	{	
		return getCutpoint(currentMethod);
	}
	
	public IHook getCutpoint(MethodJoinpoint jp) {
		
		if(jp == null)
			return super.getCutpoint();
		
		IHook result =  factory.getAspectInstance(jp,this);
		if(result==null)
			return super.getCutpoint();
		else return result;
	}
	/*
	
	*/
	protected Method findMethod(String name, Class c)  {
	
		Method[] methods = c.getMethods();
		for(int i=0;i<methods.length;++i)
			if(methods[i].getName().equals(name))
				return methods[i];
		
		return null;
	}
	
	public IHook createNewInstance() throws HookInstantiationFailed{
		
		return createNewInstance(super.getCutpoint().getClass());
	}
	
	public IHook createNewInstance(ClassLoader classLoader) throws HookInstantiationFailed{
		
		try {
			return createNewInstance(classLoader.loadClass(super.getCutpoint().getClass().getName()));
		} catch (ClassNotFoundException e) {
			throw new HookInstantiationFailed("ClassLoader "+classLoader+" failed to load hook class: "+super.getCutpoint().getClass().getName());
		}
	}
	
	public IHook createNewInstance(Class clz) throws HookInstantiationFailed {
		
		IHook newInstance = super.getCutpoint();

		try {
		newInstance = (IHook) clz.newInstance();
		
		RefinementManager.getInstance().addConnectorInlineInstanceFor(newInstance,super.getCutpoint());
		
		
		Enumeration propI = getProperties().keys();
		while (propI.hasMoreElements())  {
		
			Object key = (Object) propI.nextElement();
			Object value = getProperties().get(key);
			
			
			String f = (String) key;
			
			if (f.charAt(0)==INITIALIZER)  {
				HookInitializer h = (HookInitializer) value;
				//System.out.println("initialiser: "+h.getClass().getName());
				h.init(newInstance);
				continue;
			}
			
			
			
			Object [] params = new Object[]  {value
			};
			Method m = findMethod(f,clz);
			if(m==null)
				throw new HookInstantiationFailed("Method "+value+" not found in hook class "+clz.getName());
			
			try {
			m.invoke(newInstance,params);
			}
			catch(java.lang.Exception e)  {
				throw new HookInstantiationFailed("@DynamicCutpointElement createNewInstance: "+e.getMessage(),e);
			}
		} 
		
		}
		catch(Throwable e) {
			throw new HookInstantiationFailed("@DynamicCutpointElement createNewInstance: "+e.getMessage(),e);
		}
			
		
		return newInstance;
	}
	
	//possiblle thread unsafe, keeps object and expects next to be getCutpoint...
	public boolean isApplicable(RuntimeContext runtimeContext, java.util.Hashtable mappings,MethodJoinpoint aJascoMethod, boolean forJutta)  {
	
		currentMethod=aJascoMethod;
		return super.isApplicable(runtimeContext,mappings,aJascoMethod,forJutta);
	}
	
	
}