/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap;

import jasco.runtime.CRChangeEvent;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.ConnectorRegistryListener;
import jasco.runtime.connector.Connector;
import jasco.runtime.inline.InlineCompiler;
import jasco.runtime.transform.JAsCoClassTransformer;
import jasco.util.logging.ErrorOutputLogger;
import jasco.util.logging.Logger;

import java.lang.instrument.ClassDefinition;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;

public class JAsCoTransformer implements ClassFileTransformer,ConnectorRegistryListener {
	
	//private JAsCoHotSwap hotswap;
	private Hashtable adaptedMethods = new Hashtable();
	private Logger logger = new ErrorOutputLogger();
	private Set loadedClasses = new TreeSet();
	private Vector listeners = new Vector();
	
	
	public JAsCoTransformer()  {
	
		
		ConnectorRegistry.addConnectorRegistryListener(this);
	}
	
	public JAsCoHotSwap getHotSwap()  {
	
		return JAsCoHotSwap.getSingleton();
	}
	
	
	public TransformClassChecker getTransformChecker()  {
	
		return TransformClassChecker.getSingleton();
	}
	

	public Hashtable getAdaptedMethods()  {
	
		return adaptedMethods;
	}
	
	
	public Hashtable getAdaptedMethodsForRemoval()  {
	
		return join(getAdaptedMethods(),jasco.runtime.inline.InlineCompiler.getDefault().getInlineCompiledMethods());
	}
	
	protected Hashtable join(Hashtable table1, Hashtable table2)  {
		
		Hashtable res= new Hashtable();
		addAll(res,table1);
		addAll(res,table2);
		return res;
	}
	
	protected void addAll(Hashtable res, Hashtable table)  {
			
		Iterator i = table.entrySet().iterator();
		while (i.hasNext())  {
			
			Map.Entry entry = (Map.Entry) i.next();
			String s = (String) entry.getKey();
			Vector v = (Vector) entry.getValue();
			Vector resV = (Vector) res.get(s);
			if (resV==null)  {
				resV=new Vector();
				res.put(s,resV);
			}
			resV.addAll(v);
		}
	}
		
	public boolean trapRemoved(String classname, String methodname)  {
	
		Vector v = (Vector)getAdaptedMethods().get(classname);
		if(v==null)
			return false;
		
		return v.remove(methodname);
	}
			
	
	public void registryChanged(CRChangeEvent event)  {
	
		switch (event.getType())  {
		
			case CRChangeEvent.CONNECTOR_ADDED:
				hotswaptargetClasses((Connector)event.getParam());
				break;
			case CRChangeEvent.CONNECTOR_REMOVED:
				unhotswaptargetClasses((Connector)event.getParam());
				break;
			case CRChangeEvent.PROPERTY_CHANGE:
				cpropertyChange((Connector) event.getParam());
				break;
			case CRChangeEvent.ALLCHANGE:
			case CRChangeEvent.COMBSTRAT_ADDED:
			case CRChangeEvent.COMBSTRAT_REMOVED:
				connectorCombStratChange();
				break;
			
		}
	}
	
	public void registryUpdated()  {
	}
	
	protected void connectorCombStratChange()  {
	
		Iterator i = ConnectorRegistry.getConnectors();
		while (i.hasNext())  {
			removeInlineCompiledMethods((Connector)i.next());
		}
	}
	
	protected void cpropertyChange(Connector c)  {
		
		removeInlineCompiledMethods(c);
	}

	protected void doSingleMethodUnhotswap(Hashtable removalMethods) throws Exception {
		
		Vector swaptable = new Vector();
		Iterator i = removalMethods.entrySet().iterator();
		while (i.hasNext())  {
			Map.Entry entry = (Map.Entry) i.next();
			String classname = (String) entry.getKey();
			Vector methods = (Vector) entry.getValue();
			
			byte[] code = removeTraps(classname,methods);
			
			if(code!=null)
				swaptable.add(new ClassDefinition(Class.forName(classname),code));
		}
		
		ClassDefinition[] clzs = new ClassDefinition[swaptable.size()];
		swaptable.copyInto(clzs);
		
		getHotSwap().redefineClasses(clzs);
	}
	
	
	protected byte[] removeTraps(String classname, Vector methods) throws Exception {
		
		CtClass clazz = JAsCoClassLoader.getSingleton().loadClass(classname);
		JAsCoClassTransformer transformer = new JAsCoClassTransformer(clazz);
			
		byte[] code = transformer.removeTrapsAtRuntimeByName(methods);
		if(code==null)
			return null;
		
		Vector v = (Vector) getAdaptedMethods().get(classname);
		if(v!=null)
			v.removeAll(methods);
		
		InlineCompiler.getDefault().removeInlineCompiledMethods(classname,methods);
		
		return transformer.getResultingByteCode();
	}
		
	public void unhotswaptargetClasses(Connector unload)  {
		
		removeInlineCompiledMethods(unload);
	try {
		Hashtable toremoveMethods = new Hashtable();
		Hashtable removalMethods = getAdaptedMethods();
		Enumeration adaptedMethodsEnum = removalMethods.keys();
		while (adaptedMethodsEnum.hasMoreElements())  {
			String matchingclassname = (String)adaptedMethodsEnum.nextElement();
			Vector matchingmethods = (Vector) removalMethods.get(matchingclassname);
		
			Set resultset = AspectApplicationComputer.getInstance().getMethodsWhereHooksAreApplicable(matchingclassname);

			Vector toremove = new Vector();
			Iterator matchingmethodsI = matchingmethods.iterator();
			while (matchingmethodsI.hasNext())  {
				
				String matchingmethod = (String) matchingmethodsI.next();
				if (!resultset.contains(matchingmethod))  {
					toremove.add(matchingmethod);
				}
			}	
			toremoveMethods.put(matchingclassname,toremove);
		} 
		
		doSingleMethodUnhotswap(toremoveMethods);
	}
	catch(Exception ex)  {
			Logger.getInstance().showError("Unloading failed: ");
			Logger.getInstance().showError(ex);		
	}
	}
	
	public void removeInlineCompiledMethods(Connector theconnector)  {
		
		if(!InlineCompiler.getDefault().isEnabled())
			return;
		
		try {
		Hashtable toremoveMethods = new Hashtable();
		Hashtable inlineCompiled = InlineCompiler.getDefault().getInlineCompiledMethods();
		Enumeration adaptedMethodsEnum = inlineCompiled.keys();
		while (adaptedMethodsEnum.hasMoreElements())  {
			String matchingclassname = (String)adaptedMethodsEnum.nextElement();
			Set matchedmethods = AspectApplicationComputer.getInstance().getMethodsWhereHooksAreApplicable(matchingclassname,theconnector);
			
			Vector result = new Vector();
			result.addAll(matchedmethods);
			toremoveMethods.put(matchingclassname,result);
		}
		doSingleMethodUnhotswap(toremoveMethods);
		checkWhetherConnectorsAppy(toremoveMethods);
		}
		catch(Exception ex)  {
			Logger.getInstance().showError("Unloading inline compiled methods failed: ");
			Logger.getInstance().showError(ex);		
		}
	}
	
	//TODO possible to optimize...
	public void checkWhetherConnectorsAppy(Hashtable removalMethods)  {
		
		try {
			hotswaptargetClasses();
		} catch (Exception e) {
			Logger.getInstance().showError(e);
		}
	}
	
	public void hotswaptargetClasses() throws Exception {
		
		try  {
			
				Vector swaptable = new Vector();
				
				Class[] classes = getHotSwap().getAllClasses();
				for(int i=0;i<classes.length;++i)  {
					Class nextClass = classes[i];
					try {
						if(getTransformChecker().checkValidType(nextClass.getName())==null)
							continue;
						
						Logger.getInstance().showDebug("Hotswapping: "+nextClass.getName());	
						CtClass clz = JAsCoClassLoader.getSingleton().loadClass(nextClass.getName());
						byte[] code = this.transformClass(clz,nextClass.getName(),nextClass.getClassLoader());
						
						if(code!=null)
							swaptable.add(new ClassDefinition(nextClass,code));
						
					}
					catch(Exception ex)  {
						Logger.getInstance().showError("Transforming failed: "+nextClass.getName());
						Logger.getInstance().showError(ex);
					}
				}
				
				ClassDefinition[] clzs = new ClassDefinition[swaptable.size()];
				swaptable.copyInto(clzs);
				
				getHotSwap().redefineClasses(clzs);
			}
			catch(Exception ex)  {
					Logger.getInstance().showError("Hotswapping failed for connector:");
					Logger.getInstance().showError(ex);
				}
	}
		
	
	public void hotswaptargetClasses(Connector c)  {
		
		try  {
			Vector swaptable = new Vector();
			
			Class[] classes = getHotSwap().getAllClasses();
			for(int i=0;i<classes.length;++i)  {
				Class nextClass = classes[i];
				try {
					if(getTransformChecker().checkValidType(nextClass.getName())==null)
						continue;
					
					
					byte[] code = this.transformClass(nextClass.getName(),c,nextClass.getClassLoader());
					if(code!=null)
						swaptable.add(new ClassDefinition(nextClass,code));
				}
				catch(Exception ex)  {
					Logger.getInstance().showError("Transforming failed: "+nextClass.getName());
					Logger.getInstance().showError(ex);
				}
			}
			
			
			ClassDefinition[] clzs = new ClassDefinition[swaptable.size()];
			swaptable.copyInto(clzs);
			getHotSwap().redefineClasses(clzs);
			
		}
		catch(Exception ex)  {
				Logger.getInstance().showError("Hotswapping failed for connector: "+c.getName());
				Logger.getInstance().showError(ex);
			}
	}
	
	
	protected ClassPool getClassPool()  {
	
		return JAsCoClassLoader.getSingleton().getClassPool();
	}
	
	protected void addAdaptedMethods(String classname, Vector methods)  {
	
		Vector adapted = (Vector) getAdaptedMethods().get(classname);
		if (adapted==null)  {
			adapted=new Vector();
			getAdaptedMethods().put(classname,adapted);
		}
			
		adapted.addAll(methods);
		
		InlineCompiler.getDefault().removeInlineCompiledMethods(classname,methods);
	}
	
	protected static String getMethodname(String method)  {
		
		method=removeAnnotation(method);
		
		int index = method.indexOf("(");
		if(index==-1)
			return null;
		method=method.substring(0,index);	
		index = method.lastIndexOf(".");
		if(index==-1)
			return null;
		method=method.substring(index+1);
		return method;
	}
	
	public static String removeAnnotation(String signature) {
		
		if(!signature.startsWith("@"))
			return signature;
		
		int index = signature.indexOf(" ");
		if(index!=-1) {
			return signature.substring(index+1);
		}
		else return signature;
	}
				
	
	protected static String getClassname(String method)  {
		
		method=removeAnnotation(method);
		
		int index = method.indexOf(" ");
		if(index==-1)
			return null;
		
		method=method.substring(index+1);
		index = method.indexOf("(");
		if(index==-1)
			return null;
		method=method.substring(0,index);	
		index = method.lastIndexOf(".");
		if(index==-1)
			return null;
		method=method.substring(0,index);
		return method;
	}
	
	protected synchronized byte[] transformClass(String classname, Connector c, ClassLoader loader) {
		
		Set matchingMethods = AspectApplicationComputer.getInstance().getMethodsWhereHooksAreApplicable(classname,c);
		if(matchingMethods.isEmpty())
			return null;
		
		CtClass cls;
		try {
			cls = JAsCoClassLoader.getSingleton().loadClass(classname);
			return transformClass(cls, classname,loader,matchingMethods);
		} catch (NotFoundException e) {
			Logger.getInstance().showError(e);
		}
		return null;
	}
	
	protected synchronized byte[] transformClass(CtClass cls, String classname, ClassLoader loader) {
		
		Set matchingMethods = AspectApplicationComputer.getInstance().getMethodsWhereHooksAreApplicable(cls,loader);
		if(matchingMethods.isEmpty())
			return null;
		
		return transformClass(cls, classname,loader,matchingMethods);
	}
	
	protected synchronized byte[] transformClass(CtClass cls, String classname, ClassLoader loader,Set matchingMethods)  {
		
		
		try  {
		JAsCoClassTransformer transformer = new JAsCoClassTransformer(cls); 
		Vector adaptedMethodsV = (Vector) getAdaptedMethods().get(classname);
		if(adaptedMethodsV==null) {
			adaptedMethodsV=new Vector();
		}
		
		
		Logger.getInstance().showDebug("hotswapping class: "+cls.getName());
		byte[] result = transformer.insertTrapsAtRuntime(adaptedMethodsV,matchingMethods,loader);
		if(result==null||transformer.getAdaptedMethods().size()==0)
			return null;
		
		addAdaptedMethods(classname,transformer.getAdaptedMethods());
		
		return result;
		}
		
		catch(Throwable x)  {
			Logger.getInstance().showError("Transforming failed :"+classname);
			Logger.getInstance().showError(x.getMessage());
			x.printStackTrace();
		}
		
		return null;
	}
	
	
	
	public Instrumentation getIntrumentation()  {
	
		return JAsCoHotSwap.getSingleton().getInstrumentation();
	}
	
	protected byte[] alterURLClassLoader(CtClass clz) throws Exception {

		CtMethod[] methods = clz.getMethods();
		for(int i=0;i<methods.length;++i)  {
			if (methods[i].getName().equals("addURL"))  {
				int m = methods[i].getModifiers();
				m=javassist.Modifier.setPublic(m);
			    methods[i].setModifiers(m);
			}
		}
		clz.stopPruning(true);
		return clz.toBytecode();
	}	

	public void addClassLoadingListener(ClassLoadingListener l) {
		
		listeners.add(l);
	}
	public void removeClassLoadingListener(ClassLoadingListener l) {
		
		listeners.remove(l);
	}	
	
	protected void addPathsToClassPath(ClassLoader loader) {
		
		JAsCoClassLoader.getSingleton().appendPathsToClassPath(loader);
		//Options.appendPathsToRuntimeClassPath(loader); uses default as well
	}
	
	public byte[] transform(ClassLoader loader, String className, Class classBeingRedefined, ProtectionDomain protectionDomain, byte[] classfileBuffer)  {
		

		if(classBeingRedefined!=null)
			return null;
		
		try  {
		
		String clazzname = className.replace('/','.') ;
		if(loader!=null)
			addPathsToClassPath(loader);
	
		CtClass clz = getTransformChecker().checkValidType(clazzname);
		if(clz==null)
			return null;
		
		JAsCoClassTransformer preTransformer = new JAsCoClassTransformer(clz);
		Logger.getInstance().showDebug("preprocessing: "+className);
		boolean result =  preTransformer.preTransform();
		
		if(!result)
			return null;
		
		
		
		byte[] code =  transformClass(clz,clz.getName(),loader);
		if(code==null) { //no aspects applicable, load preprocessed instead
			code=clz.toBytecode();
			clz.defrost();
		}
		
		fireClassLoadingEvent(clz,className);
		
		return code;
		}
		catch(Exception ex)  {
			Logger.getInstance().showFullError(ex);
		}
		catch(Throwable ex)  {
			Logger.getInstance().showFullError(ex.getMessage());
			ex.printStackTrace();
		}
		return null;
	}
	
	protected void fireClassLoadingEvent(CtClass clz, String classname) {
		
		Iterator i = listeners.iterator();
		while(i.hasNext()) {
			ClassLoadingListener l = (ClassLoadingListener) i.next();
			l.classLoaded(clz,classname);
		}
	}
}
	