/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap1;

import gnu.regexp.RE;
import jasco.options.Options;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.connector.Connector;
import jasco.runtime.connector.ISignatureMatcher;
import jasco.runtime.connector.MethodSignature;
import jasco.util.logging.ErrorOutputLogger;
import jasco.util.logging.Logger;
import jasco.util.logging.OutputLogger;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtMethod;

import com.sun.jdi.Bootstrap;
import com.sun.jdi.ReferenceType;
import com.sun.jdi.VMDisconnectedException;
import com.sun.jdi.VirtualMachine;
import com.sun.jdi.VirtualMachineManager;
import com.sun.jdi.connect.LaunchingConnector;
import com.sun.jdi.event.ClassPrepareEvent;
import com.sun.jdi.event.Event;
import com.sun.jdi.event.EventIterator;
import com.sun.jdi.event.EventQueue;
import com.sun.jdi.event.EventSet;
import com.sun.jdi.event.VMDisconnectEvent;
import com.sun.jdi.request.ClassPrepareRequest;


public class HotSwap  {
	
	private static final RuntimeBeanTransformerJavassist beantransformer = new RuntimeBeanTransformerJavassist();
	private static VirtualMachine theVM;
	private static ClassPool classPool = ClassPool.getDefault();
	//private static ClassPool nonTempClassPool;
	private static final Hashtable adaptedMethods = new Hashtable();
	//private static final Hashtable origcode = new Hashtable();
	private static String[] excludedtypes;
	private static String[] userexcludedtypes;
	private static String[] nativetypes;
	private static boolean exiting = false;
	
	public static synchronized void saveOriginal(String classname)  {
		/*
		
		try  {
		CtClass clz = classPool.get(classname);
		origcode.put(classname,clz.toBytecode());
		}catch(Exception ex)  {
			Logger.getInstance().showError("savind original failed:"+classname);
			Logger.getInstance().showError(ex);
		}
		*/
	}
	
	public static synchronized void saveAdaptedMethods(String classname, Vector adapted)  {
	
		Vector v = (Vector) adaptedMethods.get(classname);
		if (v==null)  {
			v=new Vector();
			adaptedMethods.put(classname,v);
		}
		v.addAll(adapted);
	}
	
	static  {
	
		excludedtypes = new String[]  {
			"Connector.","jasco.","javassist.","gnu.regexp."
		};
		
		nativetypes = new String[]  {
			"short","boolean","int","float","long","char","void","double"
		};
		
		userexcludedtypes = new String[]  {
			"sun.*","com.sun.*","java.*","javax.*"
		};
		String tmp = System.getProperty("jasco.hotswap.replaceexcludedtypes");
		boolean overwrite = false;
		if(tmp!=null&&!tmp.equals(""))
			overwrite = new Boolean(tmp).booleanValue();	
		
		tmp = System.getProperty("jasco.hotswap.excludedtypes");
		if(tmp!=null&&!tmp.equals(""))
			initUserExcludedTypes(tmp,overwrite);
		
		tmp = System.getProperty("jasco.hotswap.keeptemp");
		if(tmp!=null&&!tmp.equals(""))
			Options.setDeleteTempFiles(! (new Boolean(tmp).booleanValue()));
		
		
		
		tmp = System.getProperty("jasco.hotswap.trapall");
		if(tmp!=null&&!tmp.equals(""))
			Options.setTransformAllMethods(new Boolean(tmp).booleanValue());
		
	}
	
	protected static void initUserExcludedTypes(String s, boolean overwrite)  {
		
		Vector res = new Vector();
		if (!overwrite)  {
			for(int i=0;i<userexcludedtypes.length;++i)
				res.add(userexcludedtypes[i]);
		}	
		StringTokenizer tokenizer = new StringTokenizer(s,Options.PATH_SEPARATOR);
		while (tokenizer.hasMoreElements())  {
			Object o = tokenizer.nextElement();
			res.add(o); 
		}
		userexcludedtypes = new String[res.size()];
		res.toArray(userexcludedtypes);
	}
	
	public static VirtualMachine getVM()  {
	
		return theVM;
	}
	
	public static boolean suspendVMWhileSwapping()  {
	
		String tmp = System.getProperty("jasco.hotswap.suspend");
		return new Boolean(tmp).booleanValue();
	}
				
	
	public static void hotswaptargetClasses(Connector c)  {
		
		if(suspendVMWhileSwapping())
			theVM.suspend();
		Hashtable swaptable = new Hashtable();
		Enumeration enumer = c.getMethods();
		while (enumer.hasMoreElements())  {
			ISignatureMatcher sigmatcher = (ISignatureMatcher)enumer.nextElement();
			
			Iterator sigs = sigmatcher.getSignatures();
			while(sigs.hasNext()) {
				
			MethodSignature sig = (MethodSignature) sigs.next();
			String methodname = sig.getMethodname();
			if(alreadyTransformed(sig.getType(),methodname))
				continue;	

			try {
				List classes = theVM.allClasses();
				Iterator classIter = classes.iterator();
				while (classIter.hasNext())  {
					ReferenceType type = (ReferenceType) classIter.next();
					try {
					
					if(!checkValidType(type.name()))
						continue;
					if (sig.matchesClass(type.name()))  {
						byte[] code = transformClassAndMethod(type,methodname);
						if(code!=null)	
							swaptable.put(type,code);
					}
					}catch(Exception ex)  {
					Logger.getInstance().showError("Transforming failed: "+type.name());
					Logger.getInstance().showError(ex);
					}
				}
			}catch(Exception ex)  {
				Logger.getInstance().showError("Transforming failed: "+sig.getType());
				Logger.getInstance().showError(ex);
			}
			} 
				
		}
		try {
		theVM.redefineClasses(swaptable);
		}
		catch(Exception ex)  {
				Logger.getInstance().showError("Hotswapping failed for connector: "+c.getName());
				Logger.getInstance().showError(ex);
			}
			
		if(suspendVMWhileSwapping())
			theVM.resume();
	}
	
	protected static void emptyTempFolder()  {
	
		File tmp = Options.getTempDir();
		if(tmp.equals(new File(".")))
			return;
		emptyFolder(tmp);
	}
	
	protected static void emptyFolder(File dir)  {
	
		String[] files = dir.list();
		for(int i=0; i<files.length; ++i)  {
			File f = new File(dir,files[i]);
			if(f==null)
				continue;
			if (f.isDirectory()) 
				emptyFolder(f);
			f.delete();
		}
		dir.delete();
	}
	
	protected static boolean transformAspects()  {
	
		return false;
	}
	
	protected static boolean isTestType(String name)  {
	
		if(name.equals("javax.swing.MyJButton")||name.equals("sunw.demo.juggler.Juggler"))
			return true;
		else return false;	
	}	
	public static boolean checkValidType(String name) throws Exception {
		
		try {
		
			
		CtClass clz = getClassPool().get(name);
		
		
		if (Options.isJAsCoBean(clz))  {
				Logger.getInstance().showDebug("Checking valid type failed (is jasco bean): "+name);
			return false;			
		}
		
		
		if(!transformAspects()&&Options.isAspect(clz)) {
			if(Options.showDebugOutput()) 
				Logger.getInstance().showDebug("Checking valid type failed (is hook): "+name);
			return false;
		}
		
		if (!transformAspects()&&Options.isAspectBean(clz))  {
			if(Options.showDebugOutput()) 
				Logger.getInstance().showDebug("Checking valid type failed (is aspect bean): "+name);
			return false;
		}
		if (Options.isTraversalConnector(clz))  {
			if(Options.showDebugOutput()) 
				Logger.getInstance().showDebug("Checking valid type failed (is traversal): "+name);
			return false;
		}
		if (Options.isCombinationStrategy(clz))  {
			if(Options.showDebugOutput()) 
				Logger.getInstance().showDebug("Checking valid type failed (is combination strategy): "+name);
			return false;
		}
				
		}
		catch(Exception ex)  {
			if (Options.showDebugOutput())  {
				Logger.getInstance().showDebug("Checking valid type failed (exception "+ex.getClass().getName()+"): "+name);
				Logger.getInstance().showError(ex);
			}
			return false;
		}	
		
		if(isTestType(name))
			return true;
		
		if(!name.startsWith("$")&&name.indexOf("$")!=-1) {
			Logger.getInstance().showDebug("Checking valid type failed (contains $): "+name);
			return false;	
		}
		
		for(int i=0;i<excludedtypes.length;++i)  {
			if(name.startsWith(excludedtypes[i])) {
				
				Logger.getInstance().showDebug("Checking valid type failed (is excluded): "+name);
				return false;	
			}
		}
		
		
		for(int i=0;i<userexcludedtypes.length;++i)  {
			RE re = new RE(makeGNUReg(userexcludedtypes[i]));
			if(re.isMatch(makeGNUReg(name))) {
				
				Logger.getInstance().showDebug("Checking valid type failed (is user excluded): "+name);
				return false;	
			}
		}
		
		for(int i=0;i<nativetypes.length;++i)
			if (name.equals(nativetypes[i]))  {
				
				Logger.getInstance().showDebug("Checking valid type failed (is native): "+name);
				return false;
			}
			
		if(name.endsWith("[]")) {
		
			Logger.getInstance().showDebug("Checking valid type failed (is array): "+name);
			return false;
		}
		
	
		Logger.getInstance().showDebug("is a valid type: "+name);
		return true;
	}	
	
	protected static byte[] transformClassAndMethod(ReferenceType type, String methodname) throws Exception {
		
			Logger.getInstance().showOutput("Hotswapping: "+type.name());	
			//origcode.put(type.name(),getClassPool().get(type.name()).toBytecode());
			byte[] code = transfromClass(type.name(),methodname);
			return code;
	}
	
	public static void unhotswaptargetClasses(Connector unload)  {
		
	try {
		Vector toremove = new Vector();
		Enumeration adaptedMethodsEnum = adaptedMethods.keys();
		while (adaptedMethodsEnum.hasMoreElements())  {
			String matchingclassname = (String)adaptedMethodsEnum.nextElement();
			boolean classTargeted = false;
			Iterator i = ConnectorRegistry.getConnectors();
			while (!classTargeted&&i.hasNext())  {
				Connector theconnector = (Connector) i.next();
				Enumeration enumer = theconnector.getMethods();
				while (enumer.hasMoreElements())  {
					ISignatureMatcher sig = (ISignatureMatcher)enumer.nextElement();
					if (sig.matchesClass(matchingclassname))  {
						classTargeted=true;
						break;
					}	
				}
			}
			if(!classTargeted)
				toremove.add(matchingclassname);
		} 
		Hashtable swaptable = new Hashtable();
		Iterator toremoveIt = toremove.iterator();
		while (toremoveIt.hasNext()) {
			String classname = (String) toremoveIt.next();
			adaptedMethods.remove(classname);
			
			List classes = theVM.classesByName(classname);
			if (classes.size() == 0)  {
				Logger.getInstance().showError("error with unloading, class not loaded yet"+classname);
				continue;
			}
			Logger.getInstance().showOutput("Unswapping "+classname);
			//byte[] bytes = (byte[]) origcode.get(classname);
			
			try  {
				String classFileName = classname.replace('.','/');
				File f = new File(Options.getTempDir().getCanonicalPath()+File.separator+classFileName+".class");
				if(f.exists())
					f.delete();
				
				ClassPool cp = new ClassPool(null);
				cp.appendSystemPath();
				byte[] bytes = cp.get(classname).toBytecode();
			 swaptable.put(classes.get(0),bytes);
			 
			}
			catch(Exception ex)  {
				Logger.getInstance().showError("Unswapping failed: "+classname);
				Logger.getInstance().showError(ex);
			}
			
			
		}
		
		theVM.redefineClasses(swaptable);
	}
	catch(Exception ex)  {
			Logger.getInstance().showError("Unloading failed: ");
			Logger.getInstance().showError(ex);		
	}
	}
	
	public static String makeGNUReg(String s)  {
		
		return jasco.util.RegexpMatcher.makeGNUReg(s);
	}
		
	
	protected static boolean alreadyTransformed(String classname, String methodname)  {
	
		if(Options.isJAsCoBean(classname))
				return true;
		else return false;	
	}
	
	protected static String getMethodname(String method)  {
	
		
		int index = method.indexOf("(");
		if(index==-1)
			return null;
		method=method.substring(0,index);	
		index = method.lastIndexOf(".");
		if(index==-1)
			return null;
		method=method.substring(index+1);
		return method;
	}
	
	public static String removeAnnotation(String signature) {
		
		if(!signature.startsWith("@"))
			return signature;
		
		int index = signature.indexOf(" ");
		if(index!=-1) {
			return signature.substring(index+1);
		}
		else return signature;
	}
	
	protected static String getClassname(String method)  {
	    
		method=removeAnnotation(method);
		
		int index = method.indexOf(" ");
		if(index==-1)
			return null;
		
		method=method.substring(index+1);
		index = method.indexOf("(");
		if(index==-1)
			return null;
		method=method.substring(0,index);	
		index = method.lastIndexOf(".");
		if(index==-1)
			return null;
		method=method.substring(0,index);
		return method;
	}
	
	public static void run(String[] args) throws Exception {
	
		StringBuffer bf = new StringBuffer();
		for(int i=0;i<args.length;++i)
			bf.append(" "+args[i]);
		
		runDefaultHotswapmachine(bf.toString());
	}
	
	
	protected static void connectToHotSwapMachine(Process p) throws Exception {
	
		com.sun.jdi.connect.AttachingConnector connector = findAttachingConnector();
		Map arguments = connector.defaultArguments();
		Iterator entries = arguments.entrySet().iterator();
		while (entries.hasNext())  {
			Map.Entry next = (Map.Entry) entries.next();
			if (next.getKey().equals("port"))  {
				com.sun.jdi.connect.Connector.Argument arg = (com.sun.jdi.connect.Connector.Argument) next.getValue();
				arg.setValue("8000"); 
			}
			Logger.getInstance().showOutput(next.getKey()+"---"+next.getValue());
		}
		VirtualMachine vmInstance = connector.attach(arguments);
		checkVM(vmInstance);
		theVM=vmInstance;
		vmInstance.setDebugTraceMode(0);

		Options.doHotSwapJDI(true);
		Options.doTransformEvents(false);
		ConnectorRegistry.start();
		vmInstance.resume();
	}
	
	protected static class ShutdownHookThread extends Thread {
	
		ShutdownHookThread()  {
		}
		
		public void run()  {
		
			try {
			exiting=true;
			if(theVM!=null)
				theVM.exit(0);
				
			}catch(VMDisconnectedException ex)  {
				//ok, vm has already stopped
			}
			if(!Options.showDebugOutput()&&Options.deleteTempFiles())
				emptyTempFolder();
		}
	}	
			
	
	protected static void addShutdownHook()  {
	
		Runtime.getRuntime().addShutdownHook(new ShutdownHookThread());
	}
		
	
	protected static void runDefaultHotswapmachine(String args) throws Exception  {
		VirtualMachineManager vm = Bootstrap.virtualMachineManager();
		com.sun.jdi.connect.LaunchingConnector cm = findLaunchingConnector();
		Logger.getInstance().showOutput(cm.getClass().getName());
		Map res = connectorArguments(cm,args);
		Iterator entries = res.entrySet().iterator();
		while (entries.hasNext())  {
			Map.Entry next = (Map.Entry) entries.next();
			if (next.getKey().equals("main"))  {
				com.sun.jdi.connect.Connector.Argument arg = (com.sun.jdi.connect.Connector.Argument) next.getValue();
				//arg.setValue("jasco.Jasco"); 
			}
			Logger.getInstance().showOutput(next.getKey()+"---"+next.getValue());
		}
		VirtualMachine vmInstance = cm.launch(res);
		checkVM(vmInstance);
		addShutdownHook();
		
		
		vmInstance.setDebugTraceMode(0);
		redirectOutput(vmInstance);
		theVM=vmInstance;
		Options.doHotSwapJDI(true);
		Options.doTransformEvents(false);
		ConnectorRegistry.start();
		if(useRedirectClassLoader())
			redirectClassLoader();
		else interceptClassLoading(vmInstance);
		vmInstance.resume();
	}
	
	protected static boolean useRedirectClassLoader()  {
		
		boolean res = true;
		String tmp = System.getProperty("jasco.hotswap.replace.classloader");
		if(tmp!=null&&!tmp.equals(""))
			res=new Boolean(tmp).booleanValue();
		//System.out.println("replacing:"+res);
		return res;
	}
	
	protected static class EventThread extends Thread  {
	
		private EventQueue queue;
		
		EventThread(EventQueue q)  {
		
			queue=q;
			setDaemon(true);
		}
		
		public void run()  {
		
			while (true)  {
				try {
				EventSet set = queue.remove();
				EventIterator i = set.eventIterator();
				while (i.hasNext())  {
					Event ev = i.nextEvent();
					if(ev instanceof VMDisconnectEvent)
						doExit();
					if (ev instanceof ClassPrepareEvent)  {
						ClassPrepareEvent loading = (ClassPrepareEvent) ev;
						transformLoadedClass(loading.referenceType());
					}
				}
				set.resume();
				}catch(InterruptedException ex)  {
				}
				catch(VMDisconnectedException ex)  {
					doExit();
				}
			}
		}
	}
	
	protected static void doExit()  {
		if(!exiting)
			System.exit(0);
	}
	
	protected static void interceptClassLoading(VirtualMachine vmInstance)  {
	
		 ClassPrepareRequest request = vmInstance.eventRequestManager().createClassPrepareRequest();
		 
		 for(int i=0;i<excludedtypes.length;++i)
		 	 request.addClassExclusionFilter(excludedtypes[i]+".*");
		 request.enable();
		 EventThread thread = new EventThread(vmInstance.eventQueue());
		 thread.start();
	}
	

	
	protected static byte[] transfromClass(String classname,String methodname) throws Exception  {
		
		String backup = Options.getOutputDir();
		Options.setOutputDir(Options.getTempDir().getCanonicalPath());
		Vector v = (Vector) adaptedMethods.get(classname);
		if(v==null)
			v=new Vector();	
		beantransformer.setClassPool(getClassPool());	
		boolean res = beantransformer.transformBean(classname,v,true,methodname);
		if(!res)
			return null;
		/*
		System.out.println("adapted methods for"+classname);
		Iterator i = 	beantransformer.getAdaptedMethods().iterator();
		while (i.hasNext())  {
			String n = (String) i.next();
			System.out.print(n);
		}*/
		Vector adapted = (Vector) adaptedMethods.get(classname);
		if(adapted==null)
			adapted=new Vector();
		adapted.addAll(beantransformer.getAdaptedMethods());
		adaptedMethods.put(classname,adapted);
		Options.setOutputDir(backup);
		return beantransformer.getDecompiledClass().toBytecode();
		
	}
	
	public static void main(String[] args)  throws Exception{
		
		
		if (!Options.showDebugOutput())  {
			Logger.getInstance().showOutput("JAsCo HotSwap 1: Loading and trapping application, please wait...");
			Logger.getInstance().showWarning("HotSwap 1 is deprecated technology, please consider porting to Java 1.5 and HotSwap 2");
			Logger.setInstance(new ErrorOutputLogger());
		}
		else  {
			Logger.setInstance(new OutputLogger(" (VM1)")); 
			Logger.getInstance().showOutput("JAsCo HotSwap 1: Loading and trapping application, please wait...");
			Logger.getInstance().showWarning("HotSwap 1 is deprecated technology, please consider porting to Java 1.5 and HotSwap 2");
		}
		emptyTempFolder();
		Options.getTempDir().mkdirs(); //make sure temp dir exists for classpath init
		initClassPool(classPool);
		run(args);
	}
	
	protected static void initClassPool(ClassPool p) throws Exception {
		
		p.appendSystemPath();
		p.insertClassPath(Options.getTempDir().getCanonicalPath());
	}
	
	protected static ClassPool getClassPool() {
		
		try  {
		ClassPool pl = new ClassPool(null);
		initClassPool(pl);
		return pl; 
		}catch (Throwable t)  {
			return null;
		}
	}
	
	protected static void redirectClassLoader() throws Exception {
		
		
		List classes = theVM.classesByName("java.lang.ClassLoader");
		if (classes.size() == 0)  {
			Logger.getInstance().showError("ERROR: Failed hotswapping classloader, classloader not found...");
			return;
		}
		byte[] res = patchClassLoader(classPool.get("java.lang.ClassLoader"));
		if(res==null)
			return;
		ReferenceType type = (ReferenceType) classes.get(0);
		Hashtable swaptable = new Hashtable();
		swaptable.put(type,res);
		theVM.redefineClasses(swaptable);
		//activate load thread...
		new HotSwapInVM();
		
		EventThread thread = new EventThread(theVM.eventQueue());
		 thread.start();
	}
	
	protected static byte[] patchClassLoader2(CtClass clz) throws Exception {
		
		CtMethod method = null;
		CtMethod[] mts = clz.getDeclaredMethods();
		for(int i=0;i<mts.length;++i)  {
			if (mts[i].getName().equals("loadClassInternal")) {
				method = mts[i];
				break;
			}
		}
		
		
		
		if (method==null)  {
			Logger.getInstance().showError("ERROR: Failed hotswapping classloader...");
			return null;
		}
		
		String code = "";
		//code+="if(!$1.startsWith(\"sun.reflect\")&&!$1.startsWith(\"java\")&&!$1.startsWith(\"jasco\")&&!$1.startsWith(\"javassist\")&&!$1.startsWith(\"com.sun\")) {";
		code+="if(!$1.startsWith(\"jasco\")&&!$1.startsWith(\"javassist\")&&!$1.startsWith(\"java.lang\")) {";
		
		code+="java.lang.String name = \"jasco.runtime.hotswap1.TheHotSwapInVM\";";
		code+="java.lang.Class cl = java.lang.ClassLoader.getSystemClassLoader().loadClass(name);";
		
		code+="java.lang.reflect.Method[] methods = cl.getDeclaredMethods();";
		code+="java.lang.reflect.Method method;";
		code+="for(int i=0;i<methods.length;++i) {";
		code+="if(methods[i].getName().equals(\"transformLoadedClassInsideVM\")) {";
		code+="method=methods[i];";
		code+="break;";
		code+="}}";
		code+="Object[] objects = new Object[1];";
		code+="objects[0]=$1;";
	//	code+="objects[1]=$2;";
		code+="byte[] code =(byte[])method.invoke(cl.newInstance(),objects);";
	//	code+="$2=code;";
	//	code+="System.out.println($1);";
		code+="if(code.length>0)";
		code+="return defineClass($1,code,0,code.length);";
		//code+="Class cl = Class.forName(jasco.runtime.cache.HotSwap);";
		
		//code+="$2=jasco.runtime.cache.HotSwap.transformLoadedClass($1, $2);}";
		code+="}";
		method.insertBefore(code);
		return clz.toBytecode();
	} 
	
	protected static byte[] patchClassLoader(CtClass clz) throws Exception {
		
		CtMethod method = null;
		CtMethod[] mts = clz.getDeclaredMethods();
		for(int i=0;i<mts.length;++i)  {
			
			if (mts[i].getName().equals("defineClass")&&mts[i].getParameterTypes().length==5)  {
				method = mts[i];
				break;
			}
		}
			
		if (method==null)  {
			Logger.getInstance().showError("ERROR: Failed hotswapping classloader...");
			return null;
		}
		
		String code = "";
		
		
		//code+="if(!$1.startsWith(\"sun.reflect\")&&!$1.startsWith(\"jasco\")&&!$1.startsWith(\"javassist\")&&!$1.startsWith(\"com.sun\")) {";
		code+="if(!$1.startsWith(\"gnu.regexp\")&&!$1.startsWith(\"jasco\")&&!$1.startsWith(\"javassist\")&&!$1.startsWith(\"java.lang\")) {";
		
		
		code+="java.lang.String name = \"jasco.runtime.hotswap1.TheHotSwapInVM\";";
		code+="java.lang.Class cl = java.lang.ClassLoader.getSystemClassLoader().loadClass(name);";
		
		code+="java.lang.reflect.Method[] methods = cl.getDeclaredMethods();";
		code+="java.lang.reflect.Method method;";
		code+="for(int i=0;i<methods.length;++i)";
		code+="if(methods[i].getName().equals(\"transformLoadedClassInsideVM\")) {";
		code+="method=methods[i];";
		code+="break;";
		code+="}";
		code+="Object[] objects = new Object[2];";
		code+="objects[0]=$1;";
		code+="objects[1]=$2;";
		code+="$2=(byte[])method.invoke(cl.newInstance(),objects);";
		code+="$4=$2.length;";
		//code+="Class cl = Class.forName(jasco.runtime.cache.HotSwap);";
		
		//code+="$2=jasco.runtime.cache.HotSwap.transformLoadedClass($1, $2);}";
		code+="}";
		//code+="else System.out.println(\"processing2\"+$1);";
		method.insertBefore(code);
		return clz.toBytecode();
	}
	
	public HotSwap()  {
	}
	
	
	
	
	
	public static void transformLoadedClass(ReferenceType type) {
		
		
		try {
		Hashtable swaptable = new Hashtable();
		String classname = type.name();
		if(!checkValidType(classname))
			return;
		Iterator connectors = ConnectorRegistry.getConnectors();
		byte[] code = null;
		while (connectors.hasNext())  {
			Connector theconnector = (Connector) connectors.next();
			Enumeration enumer = theconnector.getMethods();
			while (enumer.hasMoreElements())  {
				ISignatureMatcher sigmatcher = (ISignatureMatcher)enumer.nextElement();

				if (sigmatcher.matchesClass(classname))  {
					if(code==null)
						;//origcode.put(classname,getClassPool().get(classname).toBytecode());
					Iterator sigs = sigmatcher.getSignatures();
					byte[] code2 = null;
					while(sigs.hasNext()) {
						
						MethodSignature sig = (MethodSignature) sigs.next();
						if(!sig.matchesClass(classname))
							continue; 
						code2 = transfromClass(classname,sig.getMethodname());
						if(code2!=null)
							code=code2;
					}
				}
			}
		}
	
		if (code!=null)  {
			Logger.getInstance().showOutput("hotswapping "+type.name());
			swaptable.put(type,code);
			theVM.redefineClasses(swaptable);
			
		}
		}
		catch(Exception ex)  {
			System.out.println("boe");
			Logger.getInstance().showError("Transforming class failed :"+type.name());
			Logger.getInstance().showError(ex);
		}
	
		/*
		ClassPool cp = ClassPool.getDefault();
 		cp.insertClassPath(new ByteArrayClassPath(classname, b));
 		CtClass cc = cp.get(classname);
		*/
	}
	
/*
	protected static void initNonTempClassPoool() throws Exception {
	
		nonTempClassPool =new ClassPool(null);
		String clp = Options.getClassPath();
		String newClp = "";
		StringTokenizer tokenizer = new StringTokenizer(clp,Options.PATH_SEPARATOR,false);
		while (tokenizer.hasMoreTokens())  {
			String path = tokenizer.nextToken();
			if(!new File(path).getCanonicalPath().equals(Options.getTempDir().getCanonicalPath()))
				newClp+=path+Options.PATH_SEPARATOR;
		}
		nonTempClassPool.appendPathList(newClp);
	}*/
	
	
	public static void redirectOutput(VirtualMachine vm) {
	
		redirectOutput(vm.process());
	}
	
	public static void redirectOutput(Process process) {
        
       Thread errThread = new StreamRedirectThread("error reader",
                                             process.getErrorStream(),
                                             System.err);
        Thread outThread = new StreamRedirectThread("output reader",
                                             process.getInputStream(),
                                             System.out);
		DirectStreamRedirectThread inThread = new DirectStreamRedirectThread("input reader",
                                             System.in, process.getOutputStream()
                                             );	 
		//inThread.setBufferSize(1);
        errThread.start();
        outThread.start();
		inThread.start();
    }
	
	
	
	protected static void checkVM(VirtualMachine theVM)  {
		
		Logger.getInstance().showOutput("checking VM...");
		if(theVM==null)
			throw new IllegalArgumentException("JASCO FATAL ERROR: Hotswap VM failed to launch!");
		try {
		if (!theVM.canRedefineClasses())  {
			throw new Exception("boe");
		}
		if (!theVM.canUnrestrictedlyRedefineClasses())  
			Logger.getInstance().showOutput("limited VM detected...");
		}catch(Throwable b)  {
			Logger.getInstance().showError("VM does not support Hotswap! exiting now...");
			if (Options.systemExitAllowed())  {
				theVM.exit(1);
				System.exit(1);
			}
		}
	}
		
	
	protected static LaunchingConnector findLaunchingConnector() {
        List connectors = Bootstrap.virtualMachineManager().allConnectors();
        Iterator iter = connectors.iterator();
        while (iter.hasNext()) {
            com.sun.jdi.connect.Connector connector = (com.sun.jdi.connect.Connector)iter.next();
            if (connector.name().equals("com.sun.jdi.CommandLineLaunch")) {
                return (LaunchingConnector)connector;
            }
        }
        throw new Error("No launching connector");
    }
	
	protected static com.sun.jdi.connect.AttachingConnector findAttachingConnector() {
        List connectors = Bootstrap.virtualMachineManager().allConnectors();
        Iterator iter = connectors.iterator();
        while (iter.hasNext()) {
            com.sun.jdi.connect.Connector connector = (com.sun.jdi.connect.Connector)iter.next();
            if (connector.name().equals("com.sun.jdi.SocketAttach")) {
                return (com.sun.jdi.connect.AttachingConnector)connector;
            }
        }
        throw new Error("No launching connector");
    }
	
	

    /**
     * Return the launching connector's arguments.
     */
    protected static Map connectorArguments(LaunchingConnector connector, String mainArgs) {
        Map arguments = connector.defaultArguments();
        com.sun.jdi.connect.Connector.Argument mainArg = 
	                   (com.sun.jdi.connect.Connector.Argument)arguments.get("main");
        if (mainArg == null) {
            throw new Error("Bad launching connector");
        }
		mainArg.setValue(mainArgs);
		
		com.sun.jdi.connect.Connector.Argument optionArg = 
		(com.sun.jdi.connect.Connector.Argument)arguments.get("options");
	    if (optionArg == null) {
		throw new Error("Bad launching connector");
	    }
	    optionArg.setValue("-classpath \""+Options.getTempDir()+Options.PATH_SEPARATOR+Options.getClassPath()+"\" "+Options.getOptionsLaunchString());

		return arguments;
    }
	
   static class StreamRedirectThread extends Thread {

    private static final int BUFFER_SIZE = 2048;
    private static final int SLEEP_TIME = 7;

    private InputStream is;
    private OutputStream os;
	private int buffer = BUFFER_SIZE;

    public StreamRedirectThread(String name, InputStream is, OutputStream os) {
        super(name);
        
		setPriority(Thread.MAX_PRIORITY - 1);
        this.is = is;
        this.os = os;
		//	setDaemon(true);
    }
	
	public void setBufferSize(int i)  {
	
		buffer=i;
	
	}

    public void run() {
        byte buf[] = new byte[buffer];
        int i;
        try {
            while ((i = is.read(buf)) > 0) {
                os.write(buf, 0, i);
				//System.out.println("blabla"+new String(buf));
                try {
                    Thread.sleep(SLEEP_TIME);
                }
                catch (InterruptedException e) {
                }
            }
        }
        catch (Exception e) {
        }
    }
   }
	
	static class DirectStreamRedirectThread extends Thread {

    private InputStream is;
    private OutputStream os;
	private static final int SLEEP_TIME = 7;
	private int buffer = BUFFER_SIZE;
	private static final int BUFFER_SIZE = 2048;

    public DirectStreamRedirectThread(String name, InputStream is, OutputStream os) {
        super(name);
        
		setPriority(Thread.MAX_PRIORITY - 1);
        this.is = is;
        this.os = os;
			//	setDaemon(true);
    }
	
    public void run() {
         byte buf[] = new byte[buffer];
        int i;
        try {
            while ((i = is.read(buf)) > 0) {
                os.write(buf, 0, i);
				os.flush();
				//System.out.println("blabla"+new String(buf));
                try {
                    Thread.sleep(SLEEP_TIME);
                }
                catch (InterruptedException e) {
                }
            }
        }
        catch (Exception e) {
        }
    }
	}
	

	
   
}


