/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.hotswap1;

import jasco.util.logging.Logger;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Iterator;
import java.util.Vector;

public class VMTCPNetwork  {
	
	private Socket socket;
	private ServerSocket serverSocket;
	private int listenport;
	private InetAddress sendaddress;
	private ListenThread listenThread;
	private Vector listeners = new Vector();
	private boolean connected = true;
	private BufferedReader stin;
	private PrintStream out;
	
	public VMTCPNetwork(int listenport, InetAddress sendaddress, boolean server)  {
	
		this.sendaddress=sendaddress;
		this.listenport=listenport;
		
		try {	
			if(server)
			startServer(listenport);
		else  {	
		socket = new Socket(sendaddress,listenport);
		stin=new BufferedReader(new InputStreamReader(socket.getInputStream()));
		out=new PrintStream(socket.getOutputStream());
		Thread listenThread = new ListenThread();
		listenThread.start();
		}
		}catch(Exception ex)  {
			Logger.getInstance().showError(ex);
		}
		
	}
	
	private void startServer(int port) throws Exception {
		
		serverSocket=new ServerSocket(port);
		ServerThread server = new ServerThread();
		server.start();
	}
	
	private class ServerThread extends Thread {
	
		private ServerThread()  {
		
		setDaemon(true);
		}
	
		public void run()  {
		
			try  {
				Socket clientSock = serverSocket.accept(); 
			//	System.out.println ("serverSocket Connected From: " + clientSock.getInetAddress ()); 
				socket=clientSock; 
				stin=new BufferedReader(new InputStreamReader(socket.getInputStream()));
				out=new PrintStream(socket.getOutputStream());
				Thread listenThread = new ListenThread();
				listenThread.start();
			}catch(Exception ex)  {
			Logger.getInstance().showError(ex);
		}
		}
	}
	
	public int getBufferLength() {
	
		return 1024;
	}
	
	public Socket getSocket() {
	
		return socket;
	}
	
	public boolean hasSocket() {
	
		return socket!=null;
	}
	
	public void sendPacket(String message) {
	
		try {
		out.println(message);
		}
		catch(Exception ex) {
			
			Logger.getInstance().showError(ex);
		}
	}

	public void packetReceived(String packet) {
		
	//	System.out.println("received");
		String message = packet;
		notifyReceive(message);
	}
		
	class ListenThread extends Thread {
		
		public 	ListenThread()  {
		
			setDaemon(true);
		}
	
		public void run() {
		while(socket.isConnected()) {
			try {
				String data = stin.readLine();
				if(data != null) packetReceived(data);
				else socket.close();
			}
			catch (Exception ioe) {
				
			}
		}
		}
	}
	
	public void addNetworkListener(NetworkListener l)  {
	
		listeners.add(l);
	}
	
	public void removeNetworkListener(NetworkListener l)  {
	
		listeners.remove(l);
	}
	
	protected void notifyReceive(String packet)  {
		
		Iterator i = listeners.iterator();
		while (i.hasNext())  {
			NetworkListener listener = (NetworkListener) i.next();
			listener.receive(packet);
		}
	}
}