/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.inline;

import jasco.runtime.aspect.IHook;
import jasco.runtime.connector.HookElement;
import jasco.runtime.transform.TransformerConstants;
import jasco.util.logging.Logger;

import java.util.Hashtable;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CodeConverter;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.CtNewMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.bytecode.Bytecode;
import javassist.bytecode.CodeAttribute;
import javassist.bytecode.CodeIterator;
import javassist.bytecode.ConstPool;
import javassist.bytecode.Descriptor;
import javassist.bytecode.MethodInfo;
import javassist.bytecode.Opcode;
import javassist.expr.ExprEditor;
import javassist.expr.FieldAccess;
import javassist.expr.MethodCall;

public class InlineReplaceGenerator  {
	
	private CtClass fieldClass;
	private String accessNextHookReplace;
	private Vector accessHooks;
	private ClassPool classPool;
	private CtMethod joinpMethod;
	private long id;
	private Vector replaceHooks;
	private String args;
	private int startForArgs;
	private static Hashtable idTable = new Hashtable();

	public InlineReplaceGenerator(CtClass fieldClass, Vector accessHooks, Vector replaceHooks, ClassPool pool, CtMethod joinpMethod, long id, String args, int startForArgs)  {
	
		this.fieldClass=fieldClass;
		this.accessHooks = accessHooks;
		this.classPool=pool;
		this.joinpMethod=joinpMethod;
		this.id=id;
		this.replaceHooks=replaceHooks;
		this.args=args;
		this.startForArgs=startForArgs;
	}
	
	public void start() throws Exception {
	
		Vector targets = new Vector();
		
		addDummyProceed(fieldClass);
		
		for(int i=0;i<replaceHooks.size();++i)  {
			HookElement cpElem = (HookElement) replaceHooks.elementAt(i);
			IHook cp = cpElem.getCutpoint();
			
			final String clzname = cp.getClass().getName();
			CtClass cpclz = classPool.get(clzname);
			targets.add(cpclz);
			
			String accessHook = (String) accessHooks.elementAt(i);
			
			CtMethod method = generateAdaptedReplaceCode(cpclz, cp);
			
			redirectFieldAccesses(method,accessHook);
			
			convertAllMethodCalls(method,cpclz,accessHook,cp);
			
			fieldClass.addMethod(method);
			
			if (i+1<replaceHooks.size())  {
				HookElement nextCpElem = (HookElement) replaceHooks.elementAt(i+1);
				IHook nextcp = nextCpElem.getCutpoint();
				
				String accessNextHook = (String) accessHooks.elementAt(i+1);
					
				addInvokeNextHook(cpclz,method,accessNextHook,nextcp);
				
			}
			else  {
				
				addInvokeCorrectProceed(cpclz,method,accessHook);
			}
		}
	}

	public void addDummyProceed(CtClass clz) throws Exception {
		
		boolean found=false;
		try  {
			CtMethod m = clz.getDeclaredMethod("proceed");
			found=true;
		}
		catch(NotFoundException ex)  {
			found=false;
		}
		
		if(found) return;
		
		CtMethod method = CtNewMethod.make("public Object proceed(jasco.runtime.JascoMethod method) {return null;}", clz);
		clz.addMethod(method);
	}
	
	
	public CtMethod copyMethodForReplace(CtMethod replMethod, String newName, CtClass clz) throws Exception  {
	
		CtClass params[] = replMethod.getParameterTypes();
		boolean isStatic = Modifier.isStatic(replMethod.getModifiers());
		
		params[1]=joinpMethod.getDeclaringClass(); //type of calledobject is correct type of executing class....
		
		/* when optimized replace, add args to the end, but have to alter all load instructions :-(
		CtClass[] paramsjp = joinpMethod.getParameterTypes();
		
		//add params of jp method to the end...
		CtClass[] newparams = new CtClass[params.length+paramsjp.length+1];
		for(int i=0;i<params.length;++i)
			newparams[i]=params[i];
		
		newparams[params.length]=params[1];//calledobject type is always at one (FIXME: hardcoded int)
		
		for(int i=params.length+1,j=0;i<newparams.length;++i,++j)
			newparams[i]=paramsjp[j];
			*/
		
		CtMethod newMethod = CtNewMethod.make(replMethod.getReturnType(), newName, params, replMethod.getExceptionTypes(), null, clz);
		if (isStatic)  {
			int modifiers = newMethod.getModifiers();
			modifiers|= Modifier.STATIC;
			newMethod.setModifiers(modifiers);
		}
		
		return newMethod;
	}
	
	protected void redirectFieldAccesses(final CtMethod method, final String accessHook) throws Exception {
	
		method.instrument(new ExprEditor()  {
			public void edit(FieldAccess access)  {
				
				try  {
				if (access.getClassName().equals(method.getDeclaringClass().getName()))  
					redirectFieldAccess(access,method,accessHook);
					
				} catch(Exception ex)  {
					Logger.getInstance().showError(ex);
				}	
			}
		});
	}
	
	protected void redirectFieldAccess(FieldAccess access, CtMethod newMethod, final String accessHook) throws Exception {
		
		if(access.isReader())
			access.replace("$_="+accessHook+"."+access.getFieldName()+";");
		else access.replace(accessHook+"."+access.getFieldName()+"=$1;");		
		
	}
	
	
	
	//TODO: ADD OTHER PROCEED METHODS!!!!!!!!!!!!!
	protected void convertProceed(CtMethod replaceCopy, CtMethod proceedMethod, String accessHook, IHook cp) throws Exception {
	
		
		if(proceedMethod.getParameterTypes().length==1)
			return;
		
		
		
		CtMethod proceedMethod2 = convertMethodCall(replaceCopy,proceedMethod,accessHook,cp);
		
		CtClass invokeparams[] = joinpMethod.getParameterTypes();
		CtClass receivedparams[] = proceedMethod.getParameterTypes();
		
	}
	
	protected void convertAllMethodCalls(CtMethod copy, CtClass cpClass, String accessHook,  IHook cp) throws Exception  {
		
		CtMethod[] methods = cpClass.getMethods();
		for(int i=0;i<methods.length;++i)  {	
			if(!methods[i].getName().equals("proceed")&&!methods[i].getName().equals(IHook.REPLACE))
				convertMethodCall(copy,methods[i],accessHook,cp);
			else if(methods[i].getName().equals("proceed"))
				convertProceed(copy,methods[i],accessHook, cp);
		}	
	}
	
	protected CtMethod convertMethodCall(CtMethod copy, CtMethod orig, String accessHook, IHook cp) throws Exception {
		
		CtMethod newMethod = CtNewMethod.copy(orig,getNewMethodName(orig.getName(),cp,id),fieldClass,null);
		String ret="";
		
		int modifier = newMethod.getModifiers();
		modifier=Modifier.clear(modifier,Modifier.ABSTRACT);
		modifier=Modifier.clear(modifier,Modifier.NATIVE);
		newMethod.setModifiers(modifier);
		
		if(newMethod.getReturnType()!=CtClass.voidType)
			ret="return ";
		String body = ret+accessHook+"."+orig.getName()+"($$);";
	
		if (orig.getName().equals("equals"))  { //strange javassist bug...
			String code = "public boolean "+getNewMethodName(orig.getName(),cp,id)+"(java.lang.Object object) {"+body+"}";
			newMethod=CtNewMethod.make(code,fieldClass); 
		}
		else newMethod.setBody(body);
		fieldClass.addMethod(newMethod);
		
		//System.out.println(orig.getName()+";"+orig.getSignature());
		//System.out.println(newMethod.getName()+";"+newMethod.getSignature());
		
		CodeConverter convertor = 	new CodeConverter();
		convertor.redirectMethodCall(orig,newMethod);
		copy.instrument(convertor);
		
		return newMethod;
	}
	
	private static int idGenerator = 0;
	
	protected static Integer getNextId()  {
	
		return new Integer(idGenerator++);
	}
	
	public static String getNewMethodName(String name, IHook cp, long id) throws Exception {
		
		Integer cpid = (Integer) idTable.get(cp);
		if (cpid==null)  {
			cpid=getNextId();
			idTable.put(cp,cpid);
		}
		
		return "_inlineJasco_"+name+"_"+cp.getClass().getName().replace('.','_')+"_"+id+"_"+cpid;
	}
	
	protected void addInvokeCorrectProceed(final CtClass cpclz, final CtMethod method, final String accessHook) throws Exception  {
		
		final String hookName=accessHook.replace('#','.');
		final String descriptor = Descriptor.of(cpclz);
		
		final String newname = "proceed";
		
		method.instrument(new ExprEditor()  {
			
			public void edit(MethodCall call)  {
				
				try  {
				//if we have empty proceed method, proceed to invoc orig behavior immediately
				if (call.getMethodName().equals(newname)&&call.getMethod().getParameterTypes().length==1)  {
					replaceCallWithOriginal(call, method, descriptor, hookName);
				}
				} catch(Exception ex)  {
					Logger.getInstance().showFullError(ex);
				}	
			}
		});
		
		replaceCorrectReturns(method.getMethodInfo().getCodeAttribute(),method);
	}
	
	protected void replaceCallWithOriginal(MethodCall call, CtMethod newMethod, String descriptorOfHook, String hookName) throws Exception {
		
		int index = call.indexOfBytecode();
		boolean isStatic = Modifier.isStatic(newMethod.getModifiers());
		//CtClass nextHookClz = classPool.get(nextHook.getClass().getName());
		
		CtMethod calledmethod = call.getMethod();
		MethodInfo methodInfo = newMethod.getMethodInfo();
		CodeAttribute ca = methodInfo.getCodeAttribute();
		ConstPool cp = methodInfo.getConstPool();
		
		CtClass params[] = newMethod.getParameterTypes();
		Bytecode code = new Bytecode(cp,ca.getMaxStack()+params.length,ca.getMaxLocals());
		
		/*
		int startForArgs = this.startForArgs-1; //called object as well, invocation on that object...
		if(startForArgs<params.length) { //jp method has arguments
			CtClass[] realParams = new CtClass[params.length-startForArgs];
			for(int i=startForArgs;i<params.length;++i)
				realParams[i-startForArgs]=params[i];
		
		
			//code.addGetstatic(fieldClass,hookName,descriptorOfHook);
		}*/
		int start=1;
		if(isStatic) 
			start=0;
		code.addLoadParameters(params,start);
		Bytecode invokecode = new Bytecode(cp,ca.getMaxStack(),ca.getMaxLocals());	
		if(!isStatic)
			invokecode.addInvokevirtual(joinpMethod.getDeclaringClass(),TransformerConstants.getNewMethodName(joinpMethod),joinpMethod.getReturnType(),joinpMethod.getParameterTypes());
		else 	invokecode.addInvokestatic(joinpMethod.getDeclaringClass(),TransformerConstants.getNewMethodName(joinpMethod),joinpMethod.getReturnType(),joinpMethod.getParameterTypes());
	
		CodeIterator i = ca.iterator();
		i.write32bit(CodeIterator.NOP,index-2);
		i.write32bit(CodeIterator.NOP,index-1);
		i.move(index);
		i.write(invokecode.get(),index);
		i.insert(index,code.get());
	}
	
	protected void replaceCorrectReturns(CodeAttribute ca, CtMethod newMethod) throws Exception  {
		
		int retinstr = CodeIterator.ARETURN;
		CtClass returnType = newMethod.getReturnType();
		if(returnType.equals(CtClass.voidType))
			retinstr=CodeIterator.RETURN;
		else if(returnType.equals(CtClass.charType))
			retinstr=CodeIterator.IRETURN;	
		else if(returnType.equals(CtClass.longType))
			retinstr=CodeIterator.LRETURN;	
		else if(returnType.equals(CtClass.intType))
			retinstr=CodeIterator.IRETURN;	
		else if(returnType.equals(CtClass.intType))
			retinstr=CodeIterator.IRETURN;	
		else if(returnType.equals(CtClass.byteType))
			retinstr=CodeIterator.IRETURN;
		else if(returnType.equals(CtClass.doubleType))
			retinstr=CodeIterator.DRETURN;	
		else if(returnType.equals(CtClass.floatType))
			retinstr=CodeIterator.FRETURN;
		else if(returnType.equals(CtClass.booleanType))
			retinstr=CodeIterator.IRETURN;	
		
		
		CodeIterator i = ca.iterator();
		while (i.hasNext())  {
			int index = i.next();
			int operand = i.byteAt(index);
			
			if(operand==Opcode.ARETURN)
				i.writeByte(retinstr,index);
		}
	}	
			
	protected void addInvokeNextHook(final CtClass clz, final CtMethod copy, String accessNextHook, final IHook nextHook) throws Exception {
		
		int index=accessNextHook.indexOf("#");
		final String hookName = accessNextHook.substring(index+1);
		final String descriptor = Descriptor.of(clz);
		
		copy.instrument(new ExprEditor()  {
			public void edit(MethodCall call)  {
				
				try  {
				//if we have empty proceed method, proceed to invoc next replace immediately!
				if (call.getMethodName().equals("proceed")&&call.getMethod().getParameterTypes().length==1)  {
					replaceCallWithNextHook(call, copy, descriptor, hookName, nextHook);
				}
				} catch(Exception ex)  {
					Logger.getInstance().showError(ex);
				}	
			}
		});
		
		replaceCorrectReturns(copy.getMethodInfo().getCodeAttribute(),copy);
	}
	
	protected CtMethod findReflectMethod(CtClass hook) throws Exception {
		
		CtMethod[] methods = hook.getMethods();
		CtMethod defaultMethod = null;
		for(int i=0;i<methods.length;++i)
			if(methods[i].getName().equals(IHook.REPLACE)) { //method found, check whether not default
				CtClass[] args = methods[i].getParameterTypes();
				if(args.length!=3)
					return methods[i];
				else {
					int	start=0;
					
					if(!args[start].getName().equals("jasco.runtime.JascoMethod"))
						return methods[i];
					else if(!args[start+1].getName().equals("java.lang.Object"))
						return methods[i];
					else if(!args[start+2].getName().equals("java.lang.Object[]"))
						return methods[i];
					else defaultMethod=methods[i];
				}
			}
		
		return defaultMethod;
	}

	public CtMethod generateAdaptedReplaceCode(CtClass cpClz, IHook cp) throws Exception {
	
		String newReplName = getNewMethodName(IHook.REPLACE,cp,id);
		
		CtMethod replMethod = findReflectMethod(cpClz);
		
		CtMethod copy = copyMethodForReplace(replMethod,newReplName,fieldClass);
		copy.setBody(replMethod,null);
		
		/*
		CtClass[] params = copy.getParameterTypes();
		
		CodeAttribute att = copy.getMethodInfo().getCodeAttribute();
		att.setMaxLocals(params.length+1);
		att.setMaxStack(params.length+1);
		copy.getMethodInfo().setCodeAttribute(att);
		*/
		
		return copy;
	}

	protected void replaceCallWithNextHook(MethodCall call, CtMethod newMethod, String descriptorOfField, String fieldName, IHook nextHook) throws Exception {
	
		//System.out.println("generating new byte code: "+nextHook._Jasco_getCutpointName()+";"+newMethod.getName()+";"+descriptorOfField);
		int index = call.indexOfBytecode();
		CtClass nextHookClz = classPool.get(nextHook.getClass().getName());
		boolean isStatic = Modifier.isStatic(newMethod.getModifiers());
		
		MethodInfo methodInfo = newMethod.getMethodInfo();
		CodeAttribute ca = methodInfo.getCodeAttribute();
		ConstPool cp = methodInfo.getConstPool();
		
		Bytecode code = new Bytecode(cp,ca.getMaxStack(),ca.getMaxLocals());
		CtClass params[] = newMethod.getParameterTypes();
		int start=1;
		if(isStatic) 
			start=0;	
		if(!isStatic) //aload0 deleted by NOP
			code.addAload(0);
		
		code.addLoadParameters(params,start);
		
		Bytecode invokecode = new Bytecode(cp,ca.getMaxStack(),ca.getMaxLocals());	
		if(isStatic)
			invokecode.addInvokestatic(fieldClass,getNewMethodName(IHook.REPLACE,nextHook,id),newMethod.getReturnType(),params);
		else invokecode.addInvokevirtual(fieldClass,getNewMethodName(IHook.REPLACE,nextHook,id),newMethod.getReturnType(),params);
	
		CodeIterator i = ca.iterator();
		i.write32bit(CodeIterator.NOP,index-2);
		i.move(index);
		i.write(invokecode.get(),index);
		i.insert(index,code.get());
	}
}


	
