/*
 * Created on 2-mrt-2005
 *
 * TODO To change the template for this generated file go to
 * Window - Preferences - Java - Code Style - Code Templates
 */
package jasco.runtime.refinements;

import jasco.options.Options;
import jasco.runtime.MethodJoinpoint;
import jasco.runtime.aspect.IHook;
import jasco.util.logging.Logger;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

/* 
 Copyright (C) 2002-2005 JAsCo Development Team
 http://ssel.vub.ac.be/jasco
 
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as
 published by the Free Software Foundation; either version 2 of the
 License, or (at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA 
 */

/**
 * @author Wim
 */

public class RefinementManager {
	
	public static final RefinementComparator REFINING_COMPARATOR = new RefinementComparator();
	
	private static final RefinementManager instance = new RefinementManager();
	
	private Vector refinements = new Vector();
	
	public static final String REFINING_PACKAGE = "jasco/generated/refinements/";
	public static final String THIS_HOOK = "thisHook";

	
	
	public static RefinementManager getInstance() {
		
		return instance;
	}
	
	private RefinementManager() {
		
		initFromClassPath();
	}
	
	public static String getRefinementPackageName() {
	
		return RefinementManager.REFINING_PACKAGE.replace('/','.').substring(0,RefinementManager.REFINING_PACKAGE.length()-1);
	}
	
	public void registerRefinement(IRefinement refinement) {
		
		refinements.add(refinement);
		Collections.sort(refinements,REFINING_COMPARATOR); 
	}
	
	public void unregisterRefinement(IRefinement refinement) {
		
		refinements.remove(refinement);
	}
	
	public Object execute(String methodname, Object[] args, Class[] fargs, MethodJoinpoint joinpoint, IHook cp) throws InvocationTargetException, NoRefinementFound {
		
		IRefinement refinement = findApplicableRefinement(cp,joinpoint,methodname);
		if(refinement == null)
			throw new NoRefinementFound(cp.getClass().getName());
		
		try {
		Object[] newArgs = new Object[args.length+2];
		newArgs[0]=cp;
		newArgs[1]=joinpoint;
		for(int i=0;i<args.length;++i)
			newArgs[i+2]=args[i];
		
		Class[] newFArgs = new Class[fargs.length+2];
		newFArgs[0]=Options.loadClass("jasco.runtime.aspect.IHook");
		newFArgs[1]=Options.loadClass("jasco.runtime.MethodJoinpoint");
		for(int i=0;i<fargs.length;++i)
			newFArgs[i+2]=fargs[i];
		
		Method themethod = fetchMethod(refinement,methodname,newFArgs);
		if(themethod == null)
			throw new NoRefinementFound(cp.getClass().getName()+"."+methodname);

		return themethod.invoke(refinement,newArgs);
		}
		catch(IllegalAccessException ex) {
			Logger.getInstance().showFullError("Finding refinement for "+cp.getClass().getName()+"."+methodname+" failed! (details: illegal access)");
		}
		catch(ClassNotFoundException ex) {
			Logger.getInstance().showFullError("Finding refinement for "+cp.getClass().getName()+"."+methodname+" failed! (details: class not found"+ex.getMessage()+")");
		}
		catch(NoSuchMethodException ex) {
			throw new NoRefinementFound(cp.getClass().getName()+"."+methodname);
		}
		catch(IllegalArgumentException ex) {
			Logger.getInstance().showFullError("Finding refinement for "+cp.getClass().getName()+"."+methodname+" failed! (details follow)");
			Logger.getInstance().showFullError(ex);
		}
		
		return null;
	}
	
	protected IRefinement findApplicableRefinement(IHook cp, MethodJoinpoint jp, String method) throws NoRefinementFound {
		
		Class cpClass = cp.getClass();
		Iterator i = refinements.iterator();
		while(i.hasNext()) {
			IRefinement refinement =  (IRefinement) i.next();
		
			Class target = jp.loadClass();
			if(jp.getCalledObject()!=null)
				target=jp.getCalledObject().getClass();
			if(refinement.isApplicable(target,cp,method))
				return refinement;
		}
		
		return null;
		
	}
	
	protected Method fetchMethod(IRefinement refinement, String methodname, Class[] fargs) throws NoSuchMethodException {
		
		Method  method = refinement.getClass().getMethod(methodname,fargs);
		return method;
	}
	
	protected Method fetchMethod(IRefinement refinement, String methodname) {
		
		Method[] methods = refinement.getClass().getMethods();
		for(int i=0;i<methods.length;++i)
			if(methods[i].getName().equals(methodname))
				return methods[i];
			
		return null;
	}
	
	protected void initFromClassPath() {
		
		Iterator i = Options.getClassPathVector().iterator();
		while (i.hasNext()) { 
		{
			String next = (String) i.next();
			
			try {
			File f = new File(next);
			if(f.isDirectory())
				loadRefinementsFromDir(f);
			else if(f.getName().endsWith(".jar"))
				loadRefinementsFromJar(f);
			}catch(IOException ex) {
				Logger.getInstance().showError(ex);
			}
		}
	}
}

	/**
	 * @param f
	 */
	private void loadRefinementsFromJar(File f) throws IOException {
		
		JarFile jar = new JarFile(f);
		
		Enumeration contents =  jar.entries();
		
		while(contents.hasMoreElements()) {
			
			JarEntry next = (JarEntry) contents.nextElement();
			if(next.getName().startsWith(REFINING_PACKAGE)&&next.getName().endsWith(".class"))
				loadRefinementFromName(next.getName());
		}
	}

	/**
	 * 
	 */
	private void loadRefinementFromName(String name) {
		
		name=name.replace('/','.');
		if(name.endsWith(".class")) {
			int index = name.lastIndexOf(".");
			name=name.substring(0,index);
		}
		try {
		Class clz = Options.loadClass(name);
		if(clz.isInterface()) //cant instantiate interfaces
			return;
		
		if(Modifier.isAbstract(clz.getModifiers())) //cant instantiate abstract classes
			return;
		
		if(isRefinement(clz)) //only take refinements into account
			this.registerRefinement((IRefinement)clz.newInstance());
		}catch(ClassNotFoundException ex) {
			
			Logger.getInstance().showFullError("Loading refinement "+name+" failed! (class not found)");
		}
		catch(IllegalAccessException ex) {
			
			Logger.getInstance().showFullError("Loading refinement "+name+" failed! (illegal access)");
		}
		catch(InstantiationException ex) {
			
			Logger.getInstance().showFullError("Loading refinement "+name+" failed! (details folow)");
			Logger.getInstance().showFullError(ex);
		}
		catch(java.lang.NoClassDefFoundError ex) {
			
			Logger.getInstance().showFullError("Loading refinement "+name+" failed! (class not found)");
		}
	}
	
	public static final boolean isRefinement(Class clz) {
		
		return IRefinement.class.isAssignableFrom(clz);
	}

	/**
	 * @param f
	 */
	private void loadRefinementsFromDir(File f) {
		
		loadRefinementsFromDir(f,"");
		
	}
	
	private void loadRefinementsFromDir(File f, String accname) {
		
		
		if(!REFINING_PACKAGE.startsWith(accname)) //not correct package
			return;
		
		File[] files = f.listFiles();
		for(int i=0;i<files.length;++i) {
			String tmpname = accname+files[i].getName();
			if(files[i].isDirectory())
				loadRefinementsFromDir(files[i],tmpname+"/");
			else if(files[i].getName().endsWith(".class")&&!accname.equals(""))
				loadRefinementFromName(tmpname);
		}
	}

	/**
	 * @param newInstance
	 * @param cutpoint
	 */
	public void addConnectorInlineInstanceFor(IHook newInstance, IHook cutpoint) {
		
		Iterator refinements = this.refinements.iterator();
		while(refinements.hasNext()) {
			IRefinement next = (IRefinement) refinements.next();
			if(next instanceof ConnectorInlineRefinement) {
				ConnectorInlineRefinement inlineRef = (ConnectorInlineRefinement) next;
				if(inlineRef.isApplicableForHook(cutpoint))
					inlineRef.addTargetCutpoint(newInstance);
			}
		}
		
	}
}
