/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import jasco.options.Options;
import jasco.util.RegexpMatcher;
import jasco.util.logging.Logger;

import java.util.Set;
import java.util.Vector;

import javassist.CtClass;
import javassist.CtMethod;
import javassist.Modifier;

public class MethodAccumulator extends ClassProcessor  {

	private RegexpMatcher methodSelection;	
	private Vector excludedmethods;
	private Vector resultingVector;
	private Set matchingMethods;
	private String debugString = "Adapting";
	private boolean acceptCallMethods = false;
	
	/*
	public MethodAccumulator(Vector excludedMethods, String selectionRE)  {
		
		setExcludedMethods(excludedMethods);
		setMethodSelectionRE(selectionRE);
	}*/
	
	public MethodAccumulator(Vector excludedMethods)  {
	
		this(excludedMethods,null); 
	}
	
	public MethodAccumulator(Vector excludedMethods,Set matchingMethods)  {
		
		setExcludedMethods(excludedMethods);
		this.matchingMethods=matchingMethods;
	}
	
	public void setDebugString(String str)  {
	
		debugString=str;
	}
	
	public MethodAccumulator()  {
	
		this(new Vector()); 
	}
	/*
	public void setMethodSelectionRE(String selection)  {
		
		if(selection==null)
			return;
		this.methodSelection=new RegexpMatcher(selection);
	}
	
	protected RegexpMatcher getMethodSelectionRE()  {
	
		return methodSelection;
	}
	*/
	protected boolean matchesSelectionMethod(CtMethod method)  {
		
		if(matchingMethods==null)
			return true;
		else return matchingMethods.contains(method.getName());
	}
	
	protected boolean doPrivateMethods()  {
	
		return jasco.options.Options.transformAllMethods();
	}
	
	protected boolean isExcluded(CtMethod aMethod)  {
		
		boolean excluded = excludedmethods.contains(aMethod.getName());
		excluded=excluded||TransformerConstants.isSystemExcludedMethod(aMethod);
		return excluded;
	}
	
	public void setExcludedMethods(Vector v)  {
		
		excludedmethods=v;
	}
	
	protected void setResult(Vector v)  {
	
		resultingVector=v;
	}
	
	public Vector getResultingMethods()  {
	
		return resultingVector;
	}
	
	protected Vector accumulateAllMethods()  {
	
		CtMethod[] methods = getTargetClass().getDeclaredMethods();	
		Vector adaptedmethods = new Vector();
		
		for(int j=0; j<methods.length; ++j)
		{   
			if(methods[j].getName().indexOf('$')==-1) //do not adapt internal methods!
			{ 
				if (isExcluded(methods[j]))  {
					Logger.getInstance().showDebug("Skipping excluded method: "+methods[j].getName());
					continue;
				}
				else if(Modifier.isAbstract(methods[j].getModifiers())) //do not adapt abstract methods
					continue;
				else if(Modifier.isNative(methods[j].getModifiers())) //do not adapt native methods
					continue;
				//else if(Modifier.isStatic(methods[j].getModifiers())) //do not adapt static methods
				//	continue;
				else if (Modifier.isPublic(methods[j].getModifiers()))
				{
					if(!matchesSelectionMethod(methods[j]))
							continue;
					Logger.getInstance().showDebug(debugString+" public method: "+methods[j].getName());
					adaptedmethods.add(methods[j]);
				}
				else if(Options.transformAllMethods()) {
					
					if(!matchesSelectionMethod(methods[j]))
						continue;
					Logger.getInstance().showDebug(debugString+" non-public method: "+methods[j].getName());
					adaptedmethods.add(methods[j]);
				}
				
			}
		}
		
		return adaptedmethods;
	}
	
	protected boolean startProcessing(CtClass clz, String classname) throws Exception  {
	
		Vector res =  accumulateAllMethods();
		setResult(res);
		if(res.size()==0)
			return false;
		else return true;
	}	
	

}