/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import java.util.Iterator;
import java.util.Vector;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CtMethod;

public abstract class MethodProcessor extends ClassProcessor  {

	
	private Vector methods;
	
	public MethodProcessor()  {
	
	}
	
	public MethodProcessor(ClassPool p)  {
		super(p);
	}
	
	public MethodProcessor(ClassProcessor g)  {
	
		super(g);
		
		if (g instanceof MethodProcessor)  {
			MethodProcessor m = (MethodProcessor) g;
			setProcessMethods(m.getProcessMethods());
		}
	}
	
	public void setProcessMethods(Vector methods)  {
	
		this.methods=methods;
	}
	
	public Vector getProcessMethods()  {
	
		return methods;
	}
	
	//returns true when changes made, false when none made
	protected boolean startProcessing(CtClass clazz, String classname) throws Exception {
	
		if(getProcessMethods()==null)
			return false;
		
		boolean result = false;
		Iterator i = methods.iterator();
		for(int methodid=0;i.hasNext();++methodid)  {
			
			CtMethod method = (CtMethod) i.next();
			boolean res = processMethod(method, methodid);
			if(!res)
				jasco.util.logging.Logger.getInstance().showDebug("preprocessing method failed "+method.getName());
			result=result||res;
		}
		return result;
	}
	
	protected abstract boolean processMethod(CtMethod method, int methodid) throws Exception;
	
			
}