/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.runtime.transform;

import jasco.util.logging.Logger;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.NotFoundException;

public class TransformerConstants  {
	
	public static final String CALL_DES =  "_Jasco_CALL_NEW_";

	public static String getCallbackClass(long theID)  {
	
		return "CallbackClass"+theID;
	}
	
	public static String getCallbackClassFullname(CtClass clazz, long id)  {
	
		return clazz.getName()+"$"+getCallbackClass(id);
	}
	
	protected static String getClallBackObject(ClassProcessor clazz, long id, CtMethod method)  {
	
		return clazz.getFieldClass().getName()+".INSTANCE";
	}
	
	public static String getMethodField(long methodid)  {
	
		return "_Jasco_JascoMethod"+methodid;
	}
	
	public static String getCallMethodField(long methodid)  {
		
			return "_Jasco_JascoMethodCALL"+methodid;
		}
	
	public static String getTrapIDField(long methodid)  {
	
		return "_Jasco_JascoMethod"+methodid+"ID";
	}
	
	public static String getNewMethodName(String method, String clz)  {
		
		if(isForCall(method)) {
			String methodname = method.substring(CALL_DES.length());
			return "_Jasco_NEW_"+methodname;
		}
		else return "_Jasco_NEW_"+clz+method;
	}
	
	public static String getNewCallMethodName(String method, String clz)  {
		
			return CALL_DES+clz+method;
	}
	
	public static boolean addedByJAsCo(CtMethod method) {
		
		return method.getName().startsWith("_Jasco");
	}
	public static boolean isForCall(String methodname) {
		
		return methodname.startsWith(CALL_DES);
	}
	
	public static String getNewCallMethodName(CtMethod method)  {
		return TransformerConstants.getNewCallMethodName(method.getName(),method.getDeclaringClass().getSimpleName());
	}
	
	public static String getNewMethodName(CtMethod method)  {
		return TransformerConstants.getNewMethodName(method.getName(),method.getDeclaringClass().getSimpleName());
	}
	
	public static String getCallbackPerMethodName(CtMethod method)  throws Exception {
	
		return "jasco.runtime.callback."+method.getDeclaringClass().getName()+"_"+method.getName()+"_"+method.getParameterTypes().length;
	}
	
	public static String getClallBackPerMethodObject(CtMethod method) throws Exception  {
	
		return getCallbackPerMethodName(method)+"#INSTANCE";
	}
	
	public static boolean isSystemExcludedMethod(CtMethod method) {
		
		if(addedByJAsCo(method))
			return true;
		
		try {
			if(method.getName().equals("hashCode")&&method.getParameterTypes().length==0)
				return true;
			
			else if(method.getName().equals("equals")&&method.getParameterTypes().length==1&&method.getParameterTypes()[0].getName().equals("java.lang.Object"))
				return true;
			else return false;
		}catch(NotFoundException ex) {
			Logger.getInstance().showError(ex);
		}
		
		return false;
			
	}
	 
}