/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
/*
 * Created on 20-okt-2004
 */
package jasco.testing;

import jasco.options.Options;
import jasco.util.CommandRunner;
import jasco.util.StreamGobbler;
import jasco.util.generators.JavaGenerator;
import jasco.util.logging.Logger;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.Writer;
import java.util.Iterator;
import java.util.Vector;

import junit.framework.TestCase;

/**
 * @author Wim
 */
public abstract class DefaultJAsCoTest extends TestCase implements TestConstants, IJAsCoTest {
	
	private static Writer globalLog;
	private static java.io.File workingDir = new File(System.getProperty("working.dir"),"testing");
	private static java.io.File classDir = new File(workingDir,"classes");
	
	private boolean succeededPrev=false;
	private Vector options = new Vector();
	private String desc;
	private String command;
	private boolean skipped=false;
	
	public DefaultJAsCoTest(String desc, String command) {
		setDescription(desc);
		this.command=command;
	}
	
	static {
		try {
			new java.io.File(getLogDir()).mkdirs();
			globalLog= new FileWriter(getLogDir()+"global.txt");
			
		}
		catch(Exception ex) {
			Logger.getInstance().showError("Failed to init global log file, aborting....");
			Logger.getInstance().showError(ex);
			System.exit(0);
		}
	}
	
	public void skipTest() {
		
		skipped=true;
	}
	
	public static String getOutputDir() {
		
		return new File(getWorkingDir(),"Generated").getAbsolutePath();
	}
	
	protected  static void  setWorkingDir(File d) {
		
		workingDir=d;
	}
	
	protected static File getWorkingDir() {
		
		return workingDir;
	}
	
	protected static File getClassDir() {
		
		return classDir;
	}
	
	protected Writer getGlobalLog() {
		
		return globalLog;
	}
	
	protected void logGlobalError(String error) {
		try {
		globalLog.write("ERROR: "+error+" IN "+getDescription());
		}catch(IOException ex) {}
	}
	
	protected static String getLogDir() {
		
		String path =  new File(workingDir, "logs").getAbsolutePath();
		if(!path.endsWith(File.separator))
			path+=File.separator;
			
		return path;
	}
	
	public String getDescription() {
		return desc;
	}
	public void setDescription(String d) {
		desc=d;
		testSucceededPreviously();
	}
	public String getCommand() {
		return command;
	}
	
	public String getOptions() {
		
		String res = "";
		Iterator i = options.iterator();
		while(i.hasNext()) {
			res+=i.next().toString()+" ";
		}
		
		return res;
	}
	
	public void addOption(String s) {
		
		options.add(s);
	}
	
	public void addOptionFirst(String s) {
		
		options.insertElementAt(s,0);
	}
	
	public void writeResult(String result) {
		
		try {
		Writer w = openDataFile(getDescription());
		w.write(result);
		w.close();
		}catch(Exception ex){}
	}
	
	public boolean succeededPreviously() {
		return succeededPrev;
	}
	
	public void testSucceededPreviously() {
		try {
		BufferedReader r = new BufferedReader(readDataFile(getDescription()));
		if(r.readLine().equals("true")) {
			succeededPrev=true;
			return;
		}
		r.close();
				
		}catch(Exception ex) {}
		succeededPrev=false;
	}
	
	public void runBare() throws Throwable {
		if(skipped)
			throw new TestSkippedException();
		boolean result = runCommand(getDescription(),getCommand(),getOptions());
		writeResult(result+"");
		assertTrue("Main test",result);
	}

	public String getName() {
		return getDescription();
	}

	public boolean startTest() {
		
		Logger.getInstance().showOutputNoNewLine("Executing "+getDescription()+" ..... ");
		if(skipped) {
			Logger.getInstance().showOutput("skipped");
			return true;
		}
		try {
			boolean result =  runCommand(getDescription(),getCommand(),getOptions());
			
			writeResult(result+"");
			
			if(result) {
				Logger.getInstance().showOutput("succes");
			}
			else Logger.getInstance().showOutput("FAILED");
			return result;
		} catch(Exception ex) {
			logGlobalError(ex.getMessage());
			System.err.println(ex.getMessage());
			ex.printStackTrace();
			Logger.getInstance().showOutput("skipped");
		}
		
		return false;
	}
	
	public boolean runCommand(String testName, String command, String options) throws Exception {
		
		StringBuffer errorResult = new StringBuffer();
		Writer output = openLogFile(testName);
		StringBuffer logResult = new StringBuffer();
		
		output.write("JAsCo testing framework "+Options.getVersion()+JavaGenerator.NEWLINE);
		output.write("command: "+command+JavaGenerator.NEWLINE);
		output.write("options: "+options+JavaGenerator.NEWLINE);
		output.write("output: "+JavaGenerator.NEWLINE);
		
		Process process = CommandRunner.getProcess(command+" "+options);
		
		StreamGobbler errorGobbler = new JAsCoTestStreamGobbler(process.getErrorStream(), StreamGobbler.ERROR, errorResult, output);

		// any output?
		StreamGobbler outputGobbler = new JAsCoTestStreamGobbler(process.getInputStream(), StreamGobbler.OUTPUT, logResult, output);

		// kick them off
		errorGobbler.start();
		outputGobbler.start();
		
		int exitVal = process.waitFor();
		
		// here we should also wait for threads "errorGobbler" and "outputGobbler" to come back
		// if not, they end up with a close()'d output when we win the race
		errorGobbler.join();
		outputGobbler.join();
		
		boolean success = false;
		success=checkError(errorResult)&&checkOutput(logResult);
		
		output.close();
		
		return success;
	}
	
	public void setCommand(String s) {
		
		command = s;
	}
	
	public boolean checkError(StringBuffer error) {
		
		return error.length()==0;
	}
	
	public boolean checkOutput(StringBuffer output) {
		
		return true;
	}
	
	public Writer openLogFile(String testName) throws Exception {
		
		FileWriter output = new  FileWriter(getLogDir()+testName+".txt");
		return output;
	}
	
	public Writer openDataFile(String testName) throws Exception {
		
		FileWriter output = new  FileWriter(getLogDir()+testName+".result");
		return output;
	}
	
	public Reader readDataFile(String testName) throws Exception {
		
		FileReader output = new  FileReader(getLogDir()+testName+".result");
		return output;
	}
}
