/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.aspectparser;

import jasco.runtime.aspect.HookContextProperty;
import jasco.util.generators.JavaGenerator;
import jasco.util.logging.Logger;

import java.util.Vector;

import javassist.CannotCompileException;
import javassist.CtClass;
import javassist.CtField;
import javassist.CtMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.expr.ExprEditor;
import javassist.expr.FieldAccess;
import javassist.expr.MethodCall;

public class ContextPropertySetter
{ 
	private ClassFileEditor editor;
	
	public ContextPropertySetter(ClassFileEditor editor) {
		this.editor=editor;
	}
	
	public ClassFileEditor getEditor() {
		
		return editor;
	}


	public void setProperty(HookContextProperty original) throws Exception  {
		
		CtClass clz = getEditor().getMainClass();
		CtClass object = getEditor().loadClass("java.lang.Object");
		
		CtMethod method = clz.getDeclaredMethod(AspectClassesGenerator.HOOKCONTEXTPROP);
		HookContextProperty result = computeContextProperty(clz,original);
	
		String code = "";
		code+="return "+AspectClassesGenerator.getMethodNameForLocalHookContextProperty(clz.getName())+"($1, "+result.toSourceCode()+");"+JavaGenerator.NEWLINE;		
		method.setBody(code);
	}

	
	protected void makeEverythingPublic(CtClass clazz) throws Exception {
	
		CtField[] fields = clazz.getDeclaredFields();
		
		for(int i=0;i<fields.length;++i)  {
			CtField field = fields[i];
			int modifiers = field.getModifiers();
			modifiers=Modifier.setPublic(modifiers);
			field.setModifiers(modifiers);
		}
		
		CtMethod[] methods = clazz.getDeclaredMethods();
		
		for(int i=0;i<methods.length;++i)  {
			CtMethod method = methods[i];
			int modifiers = method.getModifiers();
			modifiers=Modifier.setPublic(modifiers);
			method.setModifiers(modifiers);
		}
		
	}
	
	public HookContextProperty computeContextProperty(CtClass clz, HookContextProperty prop) throws Exception {
		
		ReflCalledObjectExpressionEditor reflcalledObject = new ReflCalledObjectExpressionEditor();
		NonEmptyProceedExpressionEditor proceed = new NonEmptyProceedExpressionEditor();
		ReflArgsExpressionEditor reflargs = new ReflArgsExpressionEditor();
		ReflInvokeExpressionEditor reflinvoke = new ReflInvokeExpressionEditor();
		
		CtMethod[] methods = clz.getDeclaredMethods();
		for(int i=0;i<methods.length;++i)  {
			if(methods[i].getName().startsWith("_Jasco_"))
				continue;
			if(methods[i].getName().equals("proceed"))
				continue;
		
			methods[i].instrument(reflcalledObject);
			methods[i].instrument(proceed);
			methods[i].instrument(reflargs);
			methods[i].instrument(reflinvoke);
		}
		
		HookContextProperty result = new HookContextProperty(reflargs.getFound()||reflinvoke.getFound(),prop.isDirectArgsArray(),
				prop.isDirectCalledObject(), prop.isDirectCalledMethod(), 
				reflcalledObject.getFound()||reflinvoke.getFound(),proceed.getFound());
		
		return result;
	}
	

	
	class DirectCalledObjectExpressionEditor extends MyExpressionEditor  {
	
		
   	    public void edit(FieldAccess m) throws CannotCompileException {
    	   if(m.getFieldName().equals("calledobject"))
		    	found=true;
   	    }
	
	}
	
	class ReflCalledObjectExpressionEditor extends MyExpressionEditor  {
	
		public void edit(MethodCall m) throws CannotCompileException {
		   
		   try  {
		   CtMethod method = m.getMethod();
		   if(method.getDeclaringClass().getName().equals("jasco.runtime.JascoMethod")&&method.getName().equals("getCalledObject"))
		   	found=true;
		
		   }catch(Exception ex)  {
		   	Logger.getInstance().showFullError(ex);
		   }	
   		}
	}
	
	class ReflArgsExpressionEditor extends MyExpressionEditor  {
		
		public void edit(MethodCall m) throws CannotCompileException {
		   
		   try  {
		   	CtMethod method = m.getMethod();
		   	if(method.getDeclaringClass().getName().equals("jasco.runtime.JascoMethod")&&method.getName().equals("getArgumentsArray"))
		   		found=true;
			if(method.getDeclaringClass().getName().equals("jasco.runtime.JascoMethod")&&method.getName().equals("getArgumentTypes"))
		   		found=true;
		
		   }catch(Exception ex)  {
		   	Logger.getInstance().showFullError(ex);
		   }	
   		}
	}
	
	class DirectArgsExpressionEditor extends MyExpressionEditor  {
		
		private final Vector args;
		
		public DirectArgsExpressionEditor(Vector args)  {
			this.args=args;
		}
		
   	    public void edit(FieldAccess m) throws CannotCompileException {
    	   if(args.contains(m.getFieldName()))
		   	found=true;
   		}
	}
	
	class CalledMethodExpressionEditor extends MyExpressionEditor  {
	

   	    public void edit(FieldAccess m) throws CannotCompileException {
    	   if(m.getFieldName().equals("calledmethod"))
		    	found=true;
   		}
		
		
	}
	
	class NonEmptyProceedExpressionEditor extends MyExpressionEditor  {
		
	
		
   	    public void edit(MethodCall m) throws CannotCompileException {
		   
		   try  {
		   CtMethod method = m.getMethod();
		  
		   if(method.getName().equals("proceed")&&method.getParameterTypes().length>1)
			   	found=true;
		   		
			
		   }catch(Exception ex)  {
		   	Logger.getInstance().showFullError(ex);
		   }	
   		}	
	}
	
	class ReflInvokeExpressionEditor extends MyExpressionEditor  {
		
		private CtClass methodjp;
		
		ReflInvokeExpressionEditor() {
			try{
			methodjp=getEditor().loadClass("jasco.runtime.MethodJoinpoint");
			}catch(NotFoundException ex) {
				Logger.getInstance().showFullError(ex);
			}
		}
		
   	    public void edit(MethodCall m) throws CannotCompileException {
		   
		   try  {
		   CtMethod method = m.getMethod();
		  
		   if(method.getName().startsWith("invoke")&&method.getDeclaringClass().subtypeOf(methodjp))
			   	found=true;
		   		
			
		   }catch(Exception ex)  {
		   	Logger.getInstance().showFullError(ex);
		   }	
   		}	
	}
	
	class MyExpressionEditor extends ExprEditor  {
	
		protected boolean found = false;
		
 		public boolean getFound()  {
		
			return found;
 		}
	} 
	
	
}