/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.gui;

import jasco.Jasco;
import jasco.runtime.CRChangeEvent;
import jasco.runtime.ConnectorRegistry;
import jasco.runtime.ConnectorRegistryListener;
import jasco.runtime.connector.Connector;
import jasco.runtime.connector.HookElement;
import jasco.util.AbstractDialog;
import jasco.util.AbstractFrame;
import jasco.util.BrowseField;
import jasco.util.MyFileChooser;
import jasco.util.SwingFactory;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeMap;
import java.util.Vector;

import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JPanel;
import javax.swing.JRadioButtonMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.JTree;
import javax.swing.KeyStroke;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;

public class ConnectorRegistryIntrospector extends AbstractFrame {
	
	private static final String VIEW_METH = "Sort by subject";
	private static final String VIEW_CONN = "Sort by connector";
	private static final String REMO_CONN = "Unload Connector";
	private static final String COMP_CONN = "Compile&Load Connector";
	private static final String COMP_ASPE = "Compile An Aspect Bean";
	private static final String TRAN_BEAN = "Transform Bean";
	private static final String SHOW_OPTI = "Options";
	private static final String QUIT = "quit";
	
	public static final int VIEWMODE_METH = 0;
	public static final int VIEWMODE_CONN = 1;
	public static final int VIEWMODE_HOOK = 2;
	
	private JTree connectorTree = new JTree();
	private JPanel treePanel = new JPanel();
	private JPanel infoPanel = new JPanel();
	private JScrollPane treeScroller = new JScrollPane();
	private JSplitPane main = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,treePanel,infoPanel);
	
	private IntrospectorOptionsDialog optionsDialog = new IntrospectorOptionsDialog(this);
	private TransformBeanDialog transformDialog = new TransformBeanDialog(this);
	private	CompileAspectDialog aspectCompileDialog = new CompileAspectDialog(this);
	
	private int viewMode = 1;
	
	public ConnectorRegistryIntrospector()  {
	
		super("JAsCo Introspector",640,480,true);
		
		initGUI();
		loadConnectors();
		
		ConnectorRegistry.addConnectorRegistryListener(new ConnectorRegistryListener()  {
		
			public void registryChanged(CRChangeEvent event)  {
			
				loadConnectors();
				setVisible(true);
			}
			
			public void registryUpdated()  {
				
			}
		});
		
		ActionListener listener = new MenuActionListener();  
		setJMenuBar(setupMenuBar(listener));
		
		jasco.options.Options.setSystemExitAllowed(false);
		
		setVisible(true);
	}
	
	protected void loadConnectors()  {
	
		if(viewMode==VIEWMODE_CONN)
			loadConnectorsByConn();
		else if(viewMode==VIEWMODE_METH)
			loadConnectorsByMethod();
	}	
	
	public JMenuBar setupMenuBar(ActionListener listener) {
		JMenuBar bar = new JMenuBar();
		bar.add(createFileMenu(listener));
		bar.add(createToolsMenu(listener));
		bar.add(createViewMenu(listener));
		bar.add(createLookAndFeelMenu());
		
		//JascoIntrospector.displayStackTraceInformation(new Exception("boe"),true);
		
		return bar;
	} 
	
	public JMenu createFileMenu(ActionListener listener) {
		String[]    commands = {null, QUIT};
		String[]    labels = commands;
		Icon[]      icons = new Icon[commands.length];
		int[]       mnemonics = { 0, KeyEvent.VK_Q};
		KeyStroke[] shortcuts = new KeyStroke[commands.length];
		for (int i = 0; i < commands.length; ++i)
			shortcuts[i] = KeyStroke.getKeyStroke(mnemonics[i], KeyEvent.CTRL_MASK);
		return SwingFactory.createJMenu("File", KeyEvent.VK_F, commands, shortcuts, icons, mnemonics, labels, listener, true);
	}
	
	public JMenu createToolsMenu(ActionListener listener) {
		String[]    commands = {COMP_CONN, REMO_CONN,null,TRAN_BEAN,COMP_ASPE,null,SHOW_OPTI};
		String[]    labels = commands;
		Icon[]      icons = new Icon[commands.length];
		int[]       mnemonics = { KeyEvent.VK_C, KeyEvent.VK_R,0,KeyEvent.VK_T,KeyEvent.VK_B,0,KeyEvent.VK_O};
		KeyStroke[] shortcuts = new KeyStroke[commands.length];
		for (int i = 0; i < commands.length; ++i)
			shortcuts[i] = KeyStroke.getKeyStroke(mnemonics[i], KeyEvent.CTRL_MASK);
		return SwingFactory.createJMenu("Tools", KeyEvent.VK_E, commands, shortcuts, icons, mnemonics, labels, listener, true);
	}
	
	public JMenu createViewMenu(ActionListener listener) {
	
	
		JMenu viewMenu = new JMenu("View");
		viewMenu.setMnemonic(KeyEvent.VK_V);
		JRadioButtonMenuItem method = new JRadioButtonMenuItem(VIEW_METH);
		method.addActionListener(listener);
		viewMenu.add(method);
		JRadioButtonMenuItem connector = new JRadioButtonMenuItem(VIEW_CONN,true);
		connector.addActionListener(listener);
		viewMenu.add(connector);
		
		ButtonGroup group = new ButtonGroup();
		group.add(method);
		group.add(connector);
		
		return viewMenu;
	}
	
	protected class MenuActionListener implements ActionListener {
		public void actionPerformed(ActionEvent e) {
			String c = e.getActionCommand();
			
			if (c.equals(QUIT)) {
				dispose();
				System.exit(0);
			}
			else if (c.equals(VIEW_METH))  {
				viewMode=VIEWMODE_METH;
				loadConnectors();
			}
			else if (c.equals(VIEW_CONN))  {
				viewMode=VIEWMODE_CONN;	  
				loadConnectors();
			}
			else if (c.equals(REMO_CONN))  {
				removeConnector();
			}
			else if (c.equals(COMP_CONN))  {
				compConnector();
			}
			else if (c.equals(SHOW_OPTI))  {
				optionsDialog.show(); 
			}
			else if (c.equals(TRAN_BEAN))  {
				transformDialog.setVisible(true);
			}
			else if (c.equals(COMP_ASPE))  {
				aspectCompileDialog.setVisible(true);
			}
			
			
			else System.out.println("unrecognized command");
			
			setVisible(true);
		} 

	}
	
	protected void compConnector() {
		
		CompileConnectorDialog dialog = new CompileConnectorDialog(this);
		if(dialog.isCanceled())
			return;
		
		jasco.options.Options.setOutputDir(dialog.getOutputDir());
		String file = dialog.getConnectorFile();	
		try {
			Jasco.compileConnector(file);
		} catch (FileNotFoundException e) {
			
			e.printStackTrace();
		} 
		 
	}
	
	protected void removeConnector()  {
	
		TreePath path = connectorTree.getSelectionPath();
		if(path==null)
			return;
		Object o = path.getLastPathComponent();
		if (!(o instanceof ConnectorTreeNode))  {
			showErrorDialog("You should select a connector to remove!");
			return;
		}
		
		ConnectorTreeNode node = (ConnectorTreeNode) o;
		Jasco.removeConnectorInClassPath(node.getName());
		
		infoPanel.removeAll();
		infoPanel.repaint();
		
	
	}
	
	
	
	
	protected TreeMap getConnectors()  {
	
		TreeMap treeMap = new TreeMap();
		Iterator i = ConnectorRegistry.getConnectors();
		int j = 0;
		while (i.hasNext())  {
			treeMap.put(j+"",i.next());
			j++;
		}
		return treeMap;
	}
	
	protected void initGUI()  {

	
		treeScroller.add(connectorTree);
		treePanel.setLayout(new BorderLayout());
		infoPanel.setLayout(new BorderLayout());
		
		treePanel.add(treeScroller);
		
		setContentPane(main);
	}
	
	protected void loadConnectorsByConn()  {
	
		TreeMap connectors = getConnectors();
		DefaultMutableTreeNode rootNode = new DefaultMutableTreeNode("Connectors");
		Iterator i = connectors.values().iterator();
		Vector done = new Vector();
		while (i.hasNext()) 	 {
		
			
			Connector connector = (Connector) i.next();
			
			if(done.contains(connector))
				continue;
			
			done.add(connector);
			String name = connector.getName();	
			
			ConnectorTreeNode node = new ConnectorTreeNode(connector);
			rootNode.add(node);
			
			Hashtable hookNames = new Hashtable();
			Enumeration cutPoints = connector.getHooks();
			Enumeration methods = connector.getMethods();
			int k = 0;
			while (cutPoints.hasMoreElements())  {
				HookElement element = (HookElement) cutPoints.nextElement();
				String method = methods.nextElement().toString();
				//NIET GOED
				Hashtable methodHashtable = connector.getMethodHashtable(k);
				//System.out.println(methodHashtable);
				//System.out.println(k);
				String hookName = element.getCutpoint().getClass().getName().replace('_','.');
				HookTreeNode hook = (HookTreeNode) hookNames.get(hookName);
				if (hook==null)  {
					hook = new HookTreeNode(hookName);
					hookNames.put(hookName,hook);
					node.add(hook);
					hook.setHook(element.getCutpoint());
				}
				
				MethodTreeNode methodNode = new MethodTreeNode(method);
				methodNode.setMethodHashtable(methodHashtable);
				hook.add(methodNode);
				k++;
			}
			
		}
		
		connectorTree = new JTree(new DefaultTreeModel(rootNode));
		initConnectorTree(connectorTree);
		treePanel.removeAll();
		treePanel.add(connectorTree);
	}
	
	protected void loadConnectorsByMethod()  {
		/*
		TreeMap connectors = getConnectors();
		DefaultMutableTreeNode rootNode = new DefaultMutableTreeNode("Connectors");
		Iterator i = connectors.keySet().iterator();
		
		while (i.hasNext())  {
		
			String method = (String) i.next();
			
			MethodTreeNode methodNode = new MethodTreeNode(method);
			rootNode.add(methodNode);
			Vector v = (Vector) connectors.get(method);
			Iterator j = v.iterator();
			while (j.hasNext())  {
				Connector c = (Connector) j.next();
				
				ConnectorTreeNode connectorNode = new ConnectorTreeNode(c);
				methodNode.add(connectorNode);
				
				Iterator cutPoints = c.getCutpoints();
				Iterator methods = c.getMethods();
				while (cutPoints.hasNext())  {
					CutpointElement element = (CutpointElement) cutPoints.next();
					String method2 = (String) methods.next();
					if (method.equals(method2))  {
						String hookName = element.getCutpoint().getClass().getName();
						HookTreeNode hnode = new HookTreeNode(hookName);
						connectorNode.add(hnode);
					}	
				}
					
			}
		}
		
		connectorTree = new JTree(new DefaultTreeModel(rootNode));
		initConnectorTree(connectorTree);
		treePanel.removeAll();
		treePanel.add(connectorTree);
		*/
		showErrorDialog("Not yet implemented");
			
	}
	
	protected void initConnectorTree(JTree tree)  {
	
		tree.addMouseListener(new MouseAdapter()  {
		
			public void mouseReleased(MouseEvent ev)  {
			
				reloadInfoPanel();
				setVisible(true);	
			}
		});
		//tree.setCellRenderer(new MyTreeCellRenderer());
	}
	
	public void reloadInfoPanel()  {
	
		TreePath path = connectorTree.getSelectionPath();
		if (path==null)  {
		
			infoPanel.removeAll();
			infoPanel.repaint();
		
			return;
		}
		Object o = path.getLastPathComponent();
		if (!(o instanceof AbstractTreeNode)) {
		
			infoPanel.removeAll();
			infoPanel.repaint();
			
			return;
		}
			
		AbstractTreeNode node = (AbstractTreeNode) o;
		infoPanel.removeAll();
		node.initInfoPanel(infoPanel);
		infoPanel.repaint();
		
	}	
	

			
	
	
}


class MyTreeCellRenderer  extends DefaultTreeCellRenderer{

	public Component getTreeCellRendererComponent(JTree tree, Object value, boolean sel, boolean expanded, boolean isLeaf, int row, boolean hasFocus)  {
											  
		Component res = super.getTreeCellRendererComponent(tree,value, sel,expanded,isLeaf,row,hasFocus);
		JLabel resLabel = (JLabel) res;
		if (!(value instanceof AbstractTreeNode)) 
			return resLabel;
		
		AbstractTreeNode node = (AbstractTreeNode) value;
		resLabel.setIcon(node.getIcon());
		return resLabel;
	}
}
	


class CompileConnectorDialog  extends AbstractDialog{
	
	private BrowseField output;
	private BrowseField connector;
	private MyFileChooser fileChooser;
	public final static String NEWLINE = System.getProperty("line.separator");
	
	CompileConnectorDialog(JFrame parent)  {
	
		super(parent,"Compile a connector",640,480,true);
		fileChooser= new MyFileChooser(parent);
		fileChooser.selectDir(new File(System.getProperty("user.dir")));
		output=new BrowseField(fileChooser,true);
		output.selectFile(System.getProperty("user.dir"));
		connector=new BrowseField(fileChooser,false);
		connector.selectFile(System.getProperty("user.dir"));
		setupGUI();
		pack();
		setVisible(true);
	}
	
	protected void setupGUI()  {
	
		getContentPane().setLayout(new BorderLayout());
		final JTextArea text = new JTextArea();
		JPanel top = new JPanel();
		top.setLayout(new GridLayout(2,2));
		JLabel outputL = new JLabel("Select the output directory:");
		JLabel connectorL = new JLabel("Select the connector file:");
		//text.setPreferredSize(new Dimension(100,100));
		text.setRows(15);
		text.setColumns(50);
		connector.addActionListener(new ActionListener()  {
			public void actionPerformed(ActionEvent ev)  {
				
				try {
					text.setText("");
				String f = connector.getSelectedFile();
				BufferedReader buffer = new BufferedReader(new FileReader(f));
				
				String tmp = "";
				while ((tmp=buffer.readLine()) != null) 
					text.append(tmp+NEWLINE);
				
				buffer.close();
				}
				catch(Exception ex)  {
				}
			}
		});
					
		//output.selectFile(new File(jasco.options.Options.getOutputDir()).getAbsolutePath());
		top.add(connectorL);
		top.add(connector);
		
		top.add(outputL);
		top.add(output);
		
		getContentPane().add(top,"North");
		text.setLineWrap(true);
		text.setEditable(false);
		getContentPane().add(new JScrollPane(text));
		
		JButton comp = new JButton("COMPILE");
		JButton cancel = new JButton("CANCEL");
		
		cancel.addActionListener(new ActionListener()  {
			public void actionPerformed(ActionEvent ev)  {
				
				cancelDialog();
			}
		});
		
		comp.addActionListener(new ActionListener()  {
			public void actionPerformed(ActionEvent ev)  {
				
				File f = new File(output.getSelectedFile());
				boolean b = true;
				b=b&&f.exists();
				f=new File(connector.getSelectedFile());
				b=b&&f.exists();
				
				if (!b) {
					showErrorDialog("Selected Files are not valid!!");
					return;
				}
				approveDialog();
			}
		});
		
		JPanel boe = new JPanel();
		GridLayout g = new GridLayout(1,2,10,10);
		boe.setLayout(g);
		
		boe.add(cancel);
		boe.add(comp);
		
		getContentPane().add(boe,"South");
		
	}
	
	public String getConnectorFile()  {
	
		return connector.getSelectedFile();
	}
	
	public String getOutputDir()  {
	
		return output.getSelectedFile();
	}
}

					
			
	
	
	
	
