/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.traversalparser;

import jasco.options.Options;
import jasco.runtime.aspect.IHook;
import jasco.tools.connectorparser.RefinableImplGenerator;
import jasco.tools.connectorparser.RefinableMethodGenerationFailed;
import jasco.tools.jascoparser.PImport;
import jasco.util.generators.ClassGenerator;
import jasco.util.generators.GImport;
import jasco.util.generators.JavaGenerator;
import jasco.util.generators.MethodGenerator;
import jasco.util.generators.VariableGenerator;
import jasco.util.javacompiler.CompileError;
import jasco.util.javacompiler.JavaCompiler;
import jasco.util.logging.Logger;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.lang.reflect.Modifier;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;

public class TraversalClassesGenerator {
	
	private TraversalSpec traversal;
	
	private static final String ABSTRACT_CLASS = "AbstractImplementation";
	
	private Vector lastErrors;
	
	private Hashtable aspectInitCode = new Hashtable();
	
	public TraversalClassesGenerator(TraversalSpec spec)  {
	
		this.traversal=spec;
	}
	
	public TraversalSpec getTraversal()  {
	
		return traversal;
	}
	
	public void generate()  {
		
		try  {
			generateImpl();
		}
		catch(IllegalArgumentException ex)  {
			Logger.getInstance().showError(ex.getMessage());
		}
	}
	
	protected void generateImpl()  {
		
	
		ClassGenerator traversalClass = new ClassGenerator(getTraversal().getName(),getTraversal().getPackageName());
		traversalClass.addImportPackage(new GImport("edu.neu.ccs.demeter.dj.*"));
		traversalClass.addImportPackage(new GImport("java.util.*"));
		traversalClass.addImportPackage(new GImport("jasco.runtime.aspect.*"));
		traversalClass.addImportPackage(new GImport("jasco.runtime.connector.*"));
		traversalClass.addImportPackage(new GImport("jasco.runtime.connector.traversal.*"));
		traversalClass.addImportPackage(new GImport("jasco.runtime.*"));
		traversalClass.addImportPackage(new GImport("java.lang.reflect.*"));
		traversalClass.addModifier(Modifier.PUBLIC);
		traversalClass.addInterface("jasco.runtime.connector.traversal.Traversal");
		
		VariableGenerator uniqueInstance = new VariableGenerator(getTraversal().getName(),"instance");
		uniqueInstance.addModifier(Modifier.PRIVATE);
		uniqueInstance.addModifier(Modifier.STATIC);
		uniqueInstance.setInitialiser("new "+getTraversal().getName()+"()");
		traversalClass.addVariable(uniqueInstance);
		
		VariableGenerator combstrats = new VariableGenerator("Vector","strategies");
		combstrats.addModifier(Modifier.PRIVATE);
		combstrats.setInitialiser("new Vector()");
		traversalClass.addVariable(combstrats);
		
		MethodGenerator getInstance = new MethodGenerator("getTraversal",getTraversal().getName());
		getInstance.addModifier(Modifier.PUBLIC);
		getInstance.addModifier(Modifier.STATIC);
		getInstance.setImplementation("return instance;");
		traversalClass.addMethod(getInstance);
		
		MethodGenerator getInstance2 = new MethodGenerator("getInstance",getTraversal().getName());
		getInstance2.addModifier(Modifier.PUBLIC);
		getInstance2.addModifier(Modifier.STATIC);
		getInstance2.setImplementation("return instance;");
		traversalClass.addMethod(getInstance2);
		
		VariableGenerator bwrappers = new VariableGenerator("Hashtable","beforeWrappers");
		bwrappers.addModifier(Modifier.PRIVATE);
		
		traversalClass.addVariable(bwrappers);
		
		
		VariableGenerator awrappers = new VariableGenerator("Hashtable","afterWrappers");
		awrappers.addModifier(Modifier.PRIVATE);
	
		traversalClass.addVariable(awrappers);
		VariableGenerator rwrappers = new VariableGenerator("Hashtable","aroundWrappers");
		rwrappers.addModifier(Modifier.PRIVATE);
		
		traversalClass.addVariable(rwrappers);
		VariableGenerator registry = new VariableGenerator("Vector","registry");
		registry.addModifier(Modifier.PRIVATE);
		
		traversalClass.addVariable(registry);
		
		MethodGenerator constr = new MethodGenerator(getTraversal().getName(),"");
		constr.addModifier(Modifier.PRIVATE);
		constr.setImplementation(" ");
		traversalClass.addMethod(constr);
		
		MethodGenerator getHooks = new MethodGenerator("getHooks","Iterator");
		getHooks.addModifier(Modifier.PUBLIC);
		getHooks.setImplementation(" return registry.iterator();");
		traversalClass.addMethod(getHooks);
		

		/*
		MethodGenerator getAspects = new MethodGenerator("getAspectBeans","Iterator");
		getAspects.addModifier(Modifier.PUBLIC);
		String code="";
		code+="Vector res = new Vector();";
		code+="Iterator i = registry,iterator();";
		code+="while(i.hasNext()) {";
		code+="Cutpoint o = (Cutpoint) i.next();";
		code+="res.add(o._Jasco_getGlobal());";
		code+="} return res.iterator();";
		getAspects.setImplementation(code);
		traversalClass.addMethod(getAspects);*/
		
		Iterator imports = getTraversal().getImports().getImports();
		while (imports.hasNext())  {
		
			PImport next = (PImport) imports.next();
			traversalClass.addImportPackage(new GImport(next.getImportName()));
		}
		
		generateRefinableMethods(traversalClass);
		generateMainMethod(traversalClass);
		generateCombStrats(traversalClass);
		generateJascoMethodDispatch(traversalClass);
		
		final char sepChar = File.separatorChar;
	
		try  {
			String file = Options.getOutputDir() +sepChar+traversal.getPackageName().replace('.',sepChar) + sepChar+getTraversal().getName() +".java";
		
			File filetowrite = new File(file);
			if(filetowrite.exists())
				filetowrite.delete();
			else if(filetowrite.getParent()!=null)
				new File(filetowrite.getParent()).mkdirs();	
			BufferedWriter out = new BufferedWriter(new FileWriter(filetowrite));
			out.write(traversalClass.toString());
			out.close();
			
			
			//retrieve the source and classpath
			String classpath = Options.getClassPath();
						
			//compile all files
			JavaCompiler compile = new JavaCompiler();
			compile.compile(file,classpath);
			Vector errors = compile.getErrors();
			this.lastErrors=errors;
			//translate the errors to the aspect source instead of the class source
			translateAndPrint(errors);
			
			if(Options.deleteTempFiles())
				filetowrite.delete();
			
		}
		catch(Exception e)
		{
			Logger.getInstance().showError(e.getMessage()); 
		}
	}
	
	public Iterator getLastErrors()  {
		
		if(lastErrors!=null)
			return lastErrors.iterator();
		else return null;	
	}
	
	protected void generateRefinableMethods(ClassGenerator cg)  {
		
		Iterator i = getTraversal().getHookInstances();
		while (i.hasNext())  {
			HookInstance next = (HookInstance) i.next();
			String varname = next.getVarName();
			generateRefinableMethods(next,cg);
		}
	}
	
	
	
	/**
	 * @param next
	 * @param cg
	 */
	private void generateRefinableMethods(HookInstance next, ClassGenerator cg) {
		
		Iterator j = next.getAbstractMethods();
		if(!j.hasNext())
			return;
		
		RefinableImplGenerator refGen = new RefinableImplGenerator(cg);
		
		ClassGenerator result=null;
		try {
			Vector v = new Vector();
			while(j.hasNext()) {
				AbstractMethodImplementation a = (AbstractMethodImplementation) j.next();
				v.add(a.getMethodInfo());
			}
			
			
			result = refGen.generateAbstractMethodImplemenations(next.getVarName(),v.iterator(),next.getDeclaredHookClass());
			if(result==null)
				return;
			
			String res = result.getName();
			
			addAspectInitCode(next.getVarName(),"jasco.runtime.refinements.RefinementManager.getInstance().registerRefinement(new "+res+
					"("+next.getVarName()+"));"+JavaGenerator.NEWLINE);
			
			
		} catch (RefinableMethodGenerationFailed e) {
			throw new IllegalArgumentException(e.getMessage());
		}
	}

	/**
	 * @param string
	 */
	private void addAspectInitCode(String aspect, String string) {
		aspectInitCode.put(aspect,string);
	}

	protected void generateJascoMethodDispatch(ClassGenerator cg)  {
	
		MethodGenerator combstratsMethod = new MethodGenerator("getJascoMethod","TraversalJascoMethod");
		combstratsMethod.addVariable("TraversalHookWrapper","wrapper");
		combstratsMethod.addVariable("Object","context");
		combstratsMethod.addModifier(Modifier.PUBLIC);
		
		String code = "try{";
		
		code+="Class cl = getClass();"+JavaGenerator.NEWLINE;
		code+="Class[] args = new Class[] {Class.forName(\"java.lang.Object\")};"+JavaGenerator.NEWLINE;
		code+="Method method = cl.getMethod(wrapper.getVarName()+\"JascoMethod\",args);"+JavaGenerator.NEWLINE;
		code+="return (TraversalJascoMethod) method.invoke(this,new Object[] {context});"+JavaGenerator.NEWLINE;
		code+="}catch (Exception ex) {System.err.println(ex.getMessage());ex.printStackTrace();}"+JavaGenerator.NEWLINE;
		code+="return null;"+JavaGenerator.NEWLINE;
		combstratsMethod.setImplementation(code);
		cg.addMethod(combstratsMethod);
	}
		
		
	
	protected void generateCombStrats(ClassGenerator traversalClass)  {
	
		MethodGenerator combstratsMethod = new MethodGenerator("addCombinationStrategy","void");
		combstratsMethod.addVariable("CombinationStrategy","strat");
		combstratsMethod.addModifier(Modifier.PROTECTED);
		String code = "";
		code+="strategies.add(strat);"+JavaGenerator.NEWLINE;
		combstratsMethod.setImplementation(code);
		traversalClass.addMethod(combstratsMethod);
		
		MethodGenerator filterMethod = new MethodGenerator("filterHooks","HookList");
		filterMethod.addVariable("HookList","_JascoHooks");
		filterMethod.addModifier(Modifier.PROTECTED);
		code="";
		
		code += "Iterator combinations = strategies.iterator();"+JavaGenerator.NEWLINE;
		code+="while(combinations.hasNext()) {" + JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"CombinationStrategy strategy = (CombinationStrategy) combinations.next();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"_JascoHooks=strategy.validateCombinations(_JascoHooks);"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		
		
		code+="return _JascoHooks;"+JavaGenerator.NEWLINE;
		filterMethod.setImplementation(code);
		traversalClass.addMethod(filterMethod);
	}
	
	protected String generateVariableInits()  {
		
		String code="";
		code+="strategies=new Vector();"+JavaGenerator.NEWLINE;
		code+="aroundWrappers=new Hashtable();"+JavaGenerator.NEWLINE;
		code+="beforeWrappers=new Hashtable();"+JavaGenerator.NEWLINE;
		code+="afterWrappers=new Hashtable();"+JavaGenerator.NEWLINE;
		code+="registry=new Vector();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.NEWLINE;
		
		return code;
	}
	
	protected void generateMainMethod(ClassGenerator traversalClass)  {
		
		MethodGenerator traverse = new MethodGenerator("traverse","void");
		traverse.addModifier(Modifier.PUBLIC);
		traversalClass.addMethod(traverse);
		traverse.addVariable("Object","startObject");
		traverse.addException("Throwable");
		
		String visitorname = "AdaptiveVisitor";
		
		traversalClass.addPlainJava(generateVisitor(visitorname));
		
		String code = "";
		code+=generateVariableInits();
		code+=generateHookInits(traversalClass);
		code+=generateClassGraph();
		
		Iterator i = getTraversal().getCombinationStrats();
		while (i.hasNext())  {
			String temp = (String) i.next();
			code+=temp+JavaGenerator.NEWLINE;
			
		}
		
		code+="Strategy sg = new Strategy("+getTraversal().getStrategy()+");"+JavaGenerator.NEWLINE;
    	code+="TraversalGraph tg = new TraversalGraph(sg, cg);"+JavaGenerator.NEWLINE;
		code+="tg.traverse(startObject,new "+visitorname+"(";
		
		i = getTraversal().getHookInstances();
		while (i.hasNext())  {
			HookInstance hi = (HookInstance) i.next();
			code+=hi.getVarName();
			if(i.hasNext())
				code+=", ";
		}
		
		code+="));"+JavaGenerator.NEWLINE;
		
		
		code="try{"+JavaGenerator.NEWLINE+code;
		code+="}catch(VisitorMethodException ex) {if(ex.getException() instanceof JAsCoStopTraversal) ; else throw ex.getException();}"+JavaGenerator.NEWLINE;
		
		traverse.setImplementation(code);
	}
	
	protected String generateVisitor(String name)  {
	
		ClassGenerator visitor = new ClassGenerator(name,"");
		visitor.setSuperClass("Visitor");
		
		MethodGenerator start = new MethodGenerator("start","void");
		start.addModifier(Modifier.PUBLIC);
		visitor.addMethod(start);
		start.setImplementation(" ");
		
		MethodGenerator finish = new MethodGenerator("finish","void");
		finish.addModifier(Modifier.PUBLIC);
		visitor.addMethod(finish);
		finish.setImplementation(" ");
		
		MethodGenerator constr = new MethodGenerator(name,"");
		constr.addModifier(Modifier.PUBLIC);
		visitor.addMethod(constr);
		
		String constrImpl="";
		
		Iterator i = getTraversal().getHookInstances();
		while (i.hasNext())  {
			HookInstance hi = (HookInstance) i.next();
			constr.addVariable("IHook",hi.getVarName());
			constrImpl+="this."+hi.getVarName()+"="+hi.getVarName()+";";
			VariableGenerator v = new VariableGenerator("IHook",hi.getVarName());
			v.addModifier(Modifier.PRIVATE);
			visitor.addVariable(v);
		}
		
		constr.setImplementation(constrImpl);
		
		visitor.addMethod(generateVisitorMethod("before"));
		visitor.addMethod(generateVisitorMethod("after"));
		visitor.addMethod(generateReplaceVisitorMethod());
		
		visitor.addMethod(generateStopMethod());
		
		return visitor.toString();
	}
	
	protected MethodGenerator generateStopMethod()  {
	
		MethodGenerator stop = new MethodGenerator("stopTraversal","void");
		stop.addModifier(Modifier.PUBLIC);
		stop.addVariable("Object","context");
		stop.addException("JAsCoStopTraversal");
		
		String code="";
	//	code+="try{";
		//code+="Thread.currentThread().stop();";
		code+="throw new JAsCoStopTraversal(context);";
	//	code+="}catch(Throwable t) {System.out.println(Thread.currentThread().getName());}";
		
		stop.setImplementation(code);
		return stop;
	}
		
	
	protected MethodGenerator generateReplaceVisitorMethod()  {
	
		String name=IHook.REPLACE;
		MethodGenerator before = new MethodGenerator(name,"void");
		before.addVariable("Object","context");
		before.addModifier(Modifier.PUBLIC);
		before.addException("JAsCoStopTraversal");
		
		String code="try{";
		String wrapper = name+"Wrappers";
		code+="Vector temphooks = new Vector();"+JavaGenerator.NEWLINE;
		code+="Iterator _JascoHooksI = registry.iterator();"+JavaGenerator.NEWLINE;
		code+="while(_JascoHooksI.hasNext()) {"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"IHook _JascoHook = (IHook) _JascoHooksI.next();"+JavaGenerator.NEWLINE;
		//code+=JavaGenerator.TAB+"System.out.println("+wrapper+".get(_JascoHook).getClass());"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalHookWrapper wrapper = (TraversalHookWrapper) "+wrapper+".get(_JascoHook);"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"if(wrapper==null) continue;"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"String contextclass = context.getClass().getName();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalJascoMethod jascomethod = getJascoMethod(wrapper,context); "+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"if(wrapper.matchesContext(contextclass)&&wrapper.isApplicable(jascomethod))"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"temphooks.add(wrapper);"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		
		
		code+="Collections.sort(temphooks);"+JavaGenerator.NEWLINE;
		code+="HookList hooks = new HookList();"+JavaGenerator.NEWLINE;
		code+="_JascoHooksI = temphooks.iterator();"+JavaGenerator.NEWLINE;
		code+="while(_JascoHooksI.hasNext()) {"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalHookWrapper wrapper = (TraversalHookWrapper) _JascoHooksI.next();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"IHook _JascoHook = wrapper.getHook(); "+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"hooks.addImpl(_JascoHook);"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		
		code+="hooks=filterHooks(hooks);"+JavaGenerator.NEWLINE;	
		code+="_JascoHooksI = hooks.getHooks().iterator();"+JavaGenerator.NEWLINE;
		code+="while(_JascoHooksI.hasNext()) {"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"IHook _JascoHook = (IHook) _JascoHooksI.next();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalHookWrapper wrapper = (TraversalHookWrapper) "+wrapper+".get(_JascoHook);"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalJascoMethod jascomethod = getJascoMethod(wrapper,context); "+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"_JascoHook."+name+"(jascomethod,jascomethod.getCalledObject(),jascomethod.getArgumentsArray());"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"if(jascomethod.doContinue())"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+JavaGenerator.TAB+"jascomethod.doContinue(false);"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"else stopTraversal(context);"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE; 
		
		
		code+="}catch (Exception ex) {System.err.println(ex.getMessage());ex.printStackTrace();}"+JavaGenerator.NEWLINE;
		
		
		
		
		before.setImplementation(code);	
		return before;
	}
	
	protected MethodGenerator generateVisitorMethod(String name)  {
	
	
		MethodGenerator before = new MethodGenerator(name,"void");
		before.addVariable("Object","context");
		before.addModifier(Modifier.PUBLIC);
		before.addException("JAsCoStopTraversal");
		
		
		String code="try{";
		String wrapper = name+"Wrappers";
		code+="Vector temphooks = new Vector();"+JavaGenerator.NEWLINE;
		code+="Iterator _JascoHooksI = registry.iterator();"+JavaGenerator.NEWLINE;
		code+="while(_JascoHooksI.hasNext()) {"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"IHook _JascoHook = (IHook) _JascoHooksI.next();"+JavaGenerator.NEWLINE;
		//code+=JavaGenerator.TAB+"System.out.println("+wrapper+".get(_JascoHook).getClass());"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalHookWrapper wrapper = (TraversalHookWrapper) "+wrapper+".get(_JascoHook);"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"if(wrapper==null) continue;"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"String contextclass = context.getClass().getName();"+JavaGenerator.NEWLINE;
		
		code+=JavaGenerator.TAB+"JascoMethod jascomethod = getJascoMethod(wrapper,context); "+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"if(wrapper.matchesContext(contextclass)&&wrapper.isApplicable(jascomethod))"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"temphooks.add(wrapper);"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		
		
		code+="Collections.sort(temphooks);"+JavaGenerator.NEWLINE;
		code+="HookList hooks = new HookList();"+JavaGenerator.NEWLINE;
		code+="_JascoHooksI = temphooks.iterator();"+JavaGenerator.NEWLINE;
		code+="while(_JascoHooksI.hasNext()) {"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalHookWrapper wrapper = (TraversalHookWrapper) _JascoHooksI.next();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"IHook _JascoHook = wrapper.getHook(); "+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"hooks.addImpl(_JascoHook);"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		
		code+="hooks=filterHooks(hooks);"+JavaGenerator.NEWLINE;	
		code+="_JascoHooksI = hooks.getHooks().iterator();"+JavaGenerator.NEWLINE;
		code+="while(_JascoHooksI.hasNext()) {"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"IHook _JascoHook = (IHook) _JascoHooksI.next();"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"TraversalHookWrapper wrapper = (TraversalHookWrapper) "+wrapper+".get(_JascoHook);"+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"JascoMethod jascomethod = getJascoMethod(wrapper,context); "+JavaGenerator.NEWLINE;
		code+=JavaGenerator.TAB+"_JascoHook."+name+"(jascomethod,jascomethod.getCalledObject(),jascomethod.getArgumentsArray());"+JavaGenerator.NEWLINE;
		code+="}"+JavaGenerator.NEWLINE;
		
		
		code+="}catch (Exception ex) {System.err.println(ex.getMessage());ex.printStackTrace();}"+JavaGenerator.NEWLINE;
		
		if(name.equals("before"))
			code+=IHook.REPLACE+"(context);"+JavaGenerator.NEWLINE;
		
		
		before.setImplementation(code);	
		return before;
	}
		
		
	
	//te doen: argumenten vector maken die klopt met jasco method
	//te doen: type.* wordt geparsed
	
	/*
	protected String generateVisitor(String name)  {
	
		ClassGenerator visitor = new ClassGenerator(name,"");
		visitor.setSuperClass("Visitor");
		
		MethodGenerator start = new MethodGenerator("start","void");
		start.addModifier(Modifier.PUBLIC);
		visitor.addMethod(start);
		start.setImplementation(" ");
		
		MethodGenerator finish = new MethodGenerator("finish","void");
		finish.addModifier(Modifier.PUBLIC);
		visitor.addMethod(finish);
		finish.setImplementation(" ");
		
		MethodGenerator constr = new MethodGenerator(name,"");
		constr.addModifier(Modifier.PUBLIC);
		visitor.addMethod(constr);
		
		String constrImpl="";
		
		Iterator i = getTraversal().getHookInstances();
		while (i.hasNext())  {
			HookInstance hi = (HookInstance) i.next();
			constr.addVariable("Cutpoint",hi.getVarName());
			constrImpl+="this."+hi.getVarName()+"="+hi.getVarName()+";";
			VariableGenerator v = new VariableGenerator("Cutpoint",hi.getVarName());
			v.addModifier(Modifier.PRIVATE);
			visitor.addVariable(v);
		}
		
		constr.setImplementation(constrImpl);
		
		Hashtable codetable = new Hashtable();
		Hashtable hooktable = new Hashtable();
		
		i = getTraversal().getHookInstances();
		while (i.hasNext())  {
			HookInstance hi = (HookInstance) i.next();
			
			HookInit init = (HookInit) hi.getHookInits().next();
			HookMethod method = (HookMethod) init.getMethods().next();
			
			String contextclass = init.getContextClass();
			if(contextclass.indexOf("*")!=-1)
				contextclass="java.lang.Object";
		
			String code = "";
			String argscode="Class[] args = new Class[] {";
			Iterator typesI = method.getTypes();
			while (typesI.hasNext())  {
					argscode+="Class.forName(\""+typesI.next()+"\")";
					if (typesI.hasNext())  
						argscode+=", ";
			}
			argscode+="};"+JavaGenerator.NEWLINE;
			
			String methodname = method.getName();
			if (methodname.equals("*"))  {
				methodname="getClass";
				argscode="Class[] args = new Class[] {};"+JavaGenerator.NEWLINE;
			}
			
			String jascomethod = hi.getVarName()+"Jascomethod";
			code+="JascoMethod "+jascomethod+" = null;"+JavaGenerator.NEWLINE;
			code+="{"+argscode;
			//code+="System.out.println(\"boe\");"+JavaGenerator.NEWLINE;
			code+="Class contextclass = context.getClass();"+JavaGenerator.NEWLINE;
			code+="Method contextmethod = contextclass.getMethod(\""+methodname+"\", args);"+JavaGenerator.NEWLINE;
			code+="Vector tempvalues = new Vector();"+JavaGenerator.NEWLINE;
			code+="Vector values = new Vector();"+JavaGenerator.NEWLINE;
			//code+="tempvalues.add(new Vector());"+JavaGenerator.NEWLINE;
			code+="values.add(new Vector());"+JavaGenerator.NEWLINE;
			code+="values.add(new Vector());"+JavaGenerator.NEWLINE;
			code+=jascomethod+" = new JascoMethod(context,contextmethod,values,\"visiting traversal node\",contextclass.getName());"+JavaGenerator.NEWLINE;
			code+="}"+JavaGenerator.NEWLINE;
			code+=hi.getVarName()+"._Jasco_initialise("+jascomethod+");"+JavaGenerator.NEWLINE;
			code+="_JascoHooks.addImpl("+hi.getVarName()+");"+JavaGenerator.NEWLINE;
			
			//String beforecode=code+hi.getVarName()+".before();"+JavaGenerator.NEWLINE;
			//String aftercode=code+hi.getVarName()+".after();"+JavaGenerator.NEWLINE;
			
	
			
			addImpl2Table(contextclass,hi,hooktable);
			
			codetable.put(hi,code);
		
		}
		
		Comparator beforeComparator = new Comparator()  {
		
			public int compare(Object o, Object p)  {
			
				HookInstance h1 = (HookInstance) o;
				HookInstance h2 = (HookInstance) p;
				
				return h1.getBeforePriority()-h2.getBeforePriority();
			}
		};
		Comparator afterComparator = new Comparator()  {
		
			public int compare(Object o, Object p)  {
			
				HookInstance h1 = (HookInstance) o;
				HookInstance h2 = (HookInstance) p;
				
				return h1.getAfterPriority()-h2.getAfterPriority();
			}
		};
		
		 i = hooktable.entrySet().iterator();
		while (i.hasNext())  {
			Map.Entry next = (Map.Entry) i.next();
			String contextclass = (String)next.getKey();
			Vector v = (Vector)next.getValue();
			Collections.sort(v,beforeComparator);
			
			
			visitor.addMethod(generateMethod(v.iterator(),codetable,"before",contextclass));
			Collections.sort(v,afterComparator);
			visitor.addMethod(generateMethod(v.iterator(),codetable,"after",contextclass));
			
			
		}
		
		
		
		return visitor.toString();
    } */
	
	protected MethodGenerator generateMethod(Iterator j, Hashtable table,String name, String contextclass)  {
	
		String code = "try{"+JavaGenerator.NEWLINE;
		code+="HookList _JascoHooks = new HookList();"+JavaGenerator.NEWLINE;
		
		while (j.hasNext())  {
			HookInstance hi = (HookInstance) j.next();
			String tempcode = (String) table.get(hi);
			code+=tempcode;
		}
		
		String appendix = "";
		
		
		appendix+="_JascoHooks=filterHooks(_JascoHooks);"+JavaGenerator.NEWLINE;
		
		
		
		appendix+="Iterator _JascoHooksI = _JascoHooks.getHooks().iterator();"+JavaGenerator.NEWLINE;
		appendix+="while(_JascoHooksI.hasNext()) {"+JavaGenerator.NEWLINE;
		//appendix+="System.out.println(\"boe2\");"+JavaGenerator.NEWLINE;
		appendix+=JavaGenerator.TAB+"IHook _JascoHook = (IHook) _JascoHooksI.next();"+JavaGenerator.NEWLINE;
		appendix+=JavaGenerator.TAB+"_JascoHook."+name+"();"+JavaGenerator.NEWLINE;
		appendix+="}"+JavaGenerator.NEWLINE;
		appendix+="}catch (Exception ex) {System.err.println(ex.getMessage());ex.printStackTrace();}"+JavaGenerator.NEWLINE;
		code+=appendix;
			
		MethodGenerator before = new MethodGenerator(name,"void");
		before.addVariable(contextclass,"context");
		before.addModifier(Modifier.PUBLIC);
		before.setImplementation(code);
		
		return before;
	}
	
	
	protected void addImpl2Table(String key, HookInstance code, Hashtable table)  {
	
		Vector v = (Vector)table.get(key);
		if (v==null)  {
			v=new Vector();
			table.put(key,v);
		}
		v.add(code);
	}
	
	protected String generateClassGraph()  {
	
		String code = "ClassGraph cg = new ClassGraph();"+JavaGenerator.NEWLINE;
		Iterator i = getTraversal().getImports().getImports();
		while (i.hasNext())  {
			PImport next = (PImport) i.next();
			String importString = next.getImportName(); 
			int index = importString.lastIndexOf(".");
			code+="cg.addPackage(\""+importString.substring(0,index)+"\");"+JavaGenerator.NEWLINE;
		}
	
		return code;
	}
	
	
	protected String generateHookInits(ClassGenerator cg)  {
		
		String code = "";
		Iterator i = getTraversal().getHookInstances();
		while (i.hasNext())  {
			HookInstance hi = (HookInstance) i.next();
			String hookClass = hi.getDeclaredHookClass();
			String aspectbean = hi.getAspectBean();
			String varname = hi.getVarName();
			
			
			code+=aspectbean+" "+varname+"AspectBean = new "+aspectbean+"();"+JavaGenerator.NEWLINE;
			code+=hookClass+" "+varname+" = new "+hookClass+"("+varname+"AspectBean);"+JavaGenerator.NEWLINE;
			
			code+=varname+"._Jasco_setCutpointName(\""+varname+"\");"+JavaGenerator.NEWLINE;
			
			String initRefCode = (String) aspectInitCode.get(varname);
			if(initRefCode!=null)
				code+=initRefCode;
			
			HookInit init = (HookInit) hi.getHookInits().next();
			String contextclass = init.getContextClass();
			code+="registry.add("+varname+");"+JavaGenerator.NEWLINE;
			
			if(hi.getBeforePriority()!=HookInstance.NO_EXECUTIONS)
				code+="beforeWrappers.put("+varname+", new TraversalHookWrapper(\""+contextclass+"\", "+varname+", "+"\""+varname+"\""+", "+hi.getBeforePriority()+"));"+JavaGenerator.NEWLINE;
			if(hi.getAfterPriority()!=HookInstance.NO_EXECUTIONS)
				code+="afterWrappers.put("+varname+", new TraversalHookWrapper(\""+contextclass+"\", "+varname+", "+"\""+varname+"\""+", "+hi.getAfterPriority()+"));"+JavaGenerator.NEWLINE;
			if(hi.getReplacePriority()!=HookInstance.NO_EXECUTIONS)
				code+="aroundWrappers.put("+varname+", new TraversalHookWrapper(\""+contextclass+"\", "+varname+", "+"\""+varname+"\""+", "+hi.getReplacePriority()+"));"+JavaGenerator.NEWLINE;
			
			cg.addMethod(generateJascoMethodInitMethod(hi,init));
			
			Iterator j = hi.getHookExecutions();
			while (j.hasNext())  {
				String next = (String) j.next();
				code+=next+JavaGenerator.NEWLINE;
			}
		}
		
		return code;
	}
	
	protected MethodGenerator generateJascoMethodInitMethod(HookInstance hi, HookInit init)  {
	
		HookMethod method = (HookMethod) init.getMethods().next();
		
		MethodGenerator initMethod = new MethodGenerator(hi.getVarName()+"JascoMethod","TraversalJascoMethod");
		initMethod.addModifier(Modifier.PUBLIC);
		initMethod.addVariable("Object","context");
		
		String contextclass = init.getContextClass();
		if(contextclass.indexOf("*")!=-1)
			contextclass="java.lang.Object";
	
		String code = "";
		/*
		String argscode="Class[] args = new Class[] {";
		Iterator typesI = method.getTypes();
		while (typesI.hasNext())  {
				argscode+="Class.forName(\""+typesI.next()+"\")";
				if (typesI.hasNext())  
					argscode+=", ";
		}
		argscode+="};"+JavaGenerator.NEWLINE;
			
		String methodname = method.getName();
		if(methodname.indexOf("*")!=-1) {
			methodname="getClass";
			argscode="Class[] args = new Class[] {};"+JavaGenerator.NEWLINE;
		}
		*/	
		String jascomethod = hi.getVarName()+"Jascomethod";
		code+="TraversalJascoMethod "+jascomethod+" = null;"+JavaGenerator.NEWLINE;
		code+="try {";
		//code+="System.out.println(\"boe\");"+JavaGenerator.NEWLINE;
		code+="Class contextclass = context.getClass();"+JavaGenerator.NEWLINE;
		//code+="Method contextmethod = contextclass.getMethod(\""+methodname+"\", args);"+JavaGenerator.NEWLINE;
		//code+="Vector tempvalues = new Vector();"+JavaGenerator.NEWLINE;
		//code+="Vector values = new Vector();"+JavaGenerator.NEWLINE;
		//code+="tempvalues.add(new Vector());"+JavaGenerator.NEWLINE;
		//code+="values.add(new Vector());"+JavaGenerator.NEWLINE;
		//code+="values.add(new Vector());"+JavaGenerator.NEWLINE;
		code+=jascomethod+" = new TraversalJascoMethod(context,\"visiting traversal node\",contextclass.getName());"+JavaGenerator.NEWLINE;
		code+="}catch (Exception ex) {System.err.println(ex.getMessage());ex.printStackTrace();}"+JavaGenerator.NEWLINE;
		
		code+="return "+jascomethod+";"+JavaGenerator.NEWLINE;
		
		initMethod.setImplementation(code);
		return initMethod;
	}
			
	private void translateAndPrint(Vector aErrors)
	{
	
		String source = getTraversal().getName();
		int k = 0;
		
		for (int i = 0 ; i < aErrors.size() ; i++) 
		{
			//Get error and errorstring;
			CompileError error = (CompileError)aErrors.elementAt(i);
			String errorstring = error.getErrorString();
			
			
			//Get linenumber of error in original file
			int documentationindex = errorstring.lastIndexOf("//");
			if (documentationindex==-1)  {
				//System.out.println("!!!!!!!! JAsCo System Error: !!!!!!!!");
				//System.out.println(error.toString());
				//CompileError newerror = new CompileError(source,-1,errorstring);
				//Logger.getInstance().showFullError(newerror.toString());
				
				if (!error.toString().startsWith("Note:"))  {
					Logger.getInstance().showFullError(error.toString());
					k++;
				}
				continue;
			}
			
			
			
			k++;
			int restindex = errorstring.indexOf(Options.NEWLINE,documentationindex);
			String linenumber = errorstring.substring(documentationindex + 2,restindex);
			
			int reallinenumber = (new Integer(linenumber)).intValue();
			
			
			//Get errormessage by extracting documentation
			String realerror = errorstring.substring(0,documentationindex) + Options.NEWLINE + errorstring.substring(restindex + 1); 
			
			CompileError newerror = new CompileError(source,reallinenumber,realerror);
			
			//System.out.print(newerror.toString());
			Logger.getInstance().showError(newerror);
		}
		if (k != 0)
		{
			if (k == 1)
			{
				//System.out.println(1 + " error");
				Logger.getInstance().showOutput(1 + " error");
			}
			else 
			{
				//System.out.println(aErrors.size() + " errors");
				Logger.getInstance().showOutput(k + " errors");
			}
		}
	
	}	
	/*
	private void translateAndPrint(Vector aErrors)
	{
		String source = getTraversal().getName();
		
		for (int i = 0 ; i < aErrors.size() ; i++) 
		{
			//Get error and errorstring;
			CompileError error = (CompileError)aErrors.elementAt(i);
			String errorstring = error.getErrorString();
			
			//Get linenumber of error in original file
			int documentationindex = errorstring.lastIndexOf("//");
			if (documentationindex==-1)  {
				Logger.getInstance().showFullError(error.toString());
			
				continue;
			}
			
			int restindex = errorstring.indexOf("\n",documentationindex);
			String linenumber = errorstring.substring(documentationindex + 2,restindex);
			
			int reallinenumber = (new Integer(linenumber)).intValue();
			
			
			//Get errormessage by extracting documentation
			String realerror = errorstring.substring(0,documentationindex) + "\n" + errorstring.substring(restindex + 1); 
			
			CompileError newerror = new CompileError(source,reallinenumber,realerror);
			
			Logger.getInstance().showError(newerror.toString());
		}
		if (aErrors.size() != 0)
		{
			if (aErrors.size() == 1)
			{
				Logger.getInstance().showError(1 + " error");
			}
			else 
			{
				Logger.getInstance().showError(aErrors.size() + " errors");
			}
		}
	}*/
}
		
		
