/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.tools.traversalparser;

import jasco.runtime.aspect.IHook;
import jasco.tools.jascoparser.JascoParseException;
import jasco.tools.jascoparser.JascoParser;
import jasco.tools.jascoparser.JascoParserTokenManager;
import jasco.tools.jascoparser.PImports;
import jasco.tools.jascoparser.PJavaClass;
import jasco.tools.jascoparser.PMethod;
import jasco.util.javacompiler.CompileError;
import jasco.util.logging.Logger;
/* Generated By:JavaCC: Do not edit this line. TraversalParser.java */
public class TraversalParser extends JascoParser {

	private TraversalSpec traversal = new TraversalSpec();
	
	public TraversalParser(JascoParserTokenManager tokenizer, String filename)
	{
		super(tokenizer,filename);
		traversal=new TraversalSpec();
	}	

	public TraversalSpec getTraversalSpec()
	{

		return traversal;
	}

	public void start()
	{
		try
		{
			parsePackageDeclaration();
			PImports imports = parseImportDeclaration();
			getTraversalSpec().setImports(imports); 
			parseTraversalDeclaration();
		}
		catch(JascoParseException exception)
		{
			Logger.getInstance().showError(exception.getError());
			Logger.getInstance().showOutput("1 error");
			jasco.options.Options.exit(1);

		} 
	}

	private String transformToAspectClasses(String s)
	{

		int i = s.lastIndexOf(".");
		if(i==-1)
			return s;

		return s.substring(0,i)+"_"+s.substring(i+1);
	}	

	


	public void parsePackageDeclaration()  throws JascoParseException
	{


		if (getCurrentId()!=PACKAGE)
			return;

		String packageName = "";
		getNextToken();
		packageName+=getCurrentImage();
		getNextToken();

		while (getCurrentId() != SEMICOLON)
		{

			checkDOT("package "+packageName);

			getNextToken();

			checkIDENTIFIER("package "+packageName);

			packageName+="."+getCurrentImage();
			getNextToken();
		}

		getTraversalSpec().setPackageName(packageName);	
		getNextToken();
	}

	public void parseTraversalDeclaration() throws JascoParseException
	{


		String currentParsed = "";

		CompileError error = new CompileError(getFileName(), getPreviousLine(), "keyword traversalconnector expected");
		assertToken(TRAVERSAL,getCurrentId(),error);

		currentParsed+=getCurrentImage();

		getNextToken();
		checkIDENTIFIER(currentParsed);
		currentParsed+=getCurrentImage();

		getTraversalSpec().setName(getCurrentImage());
		checkFileName(getTraversalSpec().getName());

		getNextToken();
		checkLPAREN(currentParsed);
		currentParsed+=getCurrentImage();

		getNextToken();
		error = new CompileError(getFileName(), getPreviousLine(), "String containing traversal strategy expected");
		assertToken(STRING_LITERAL,getCurrentId(),error);
		currentParsed+=getCurrentImage();

		getTraversalSpec().setStrategy(getCurrentImage());

		getNextToken();
		checkRPAREN(currentParsed);
		currentParsed+=getCurrentImage();

		getNextToken();
		checkLBRACE(currentParsed);
		currentParsed+=getCurrentImage();

		parseHookInstances();
	}

	protected void parseHookBehaviorExec(HookInstance hi, String methodname) throws JascoParseException
	{

		if(methodname.equals("before"))
			hi.setBeforePriority(getPreviousLine());
		else if(methodname.equals("after"))
			hi.setAfterPriority(getPreviousLine());
		else if(methodname.equals(IHook.REPLACE))
			hi.setReplacePriority(getPreviousLine());
		else if(methodname.equals("default"))
			hi.setDefaultPriority(getPreviousLine());	

		getNextToken();
		while (getCurrentId()!=SEMICOLON)
		{
			getNextToken();
		}
	}	

	protected void parseHookExecution(String parseString) throws JascoParseException
	{

		String hookType = parseString;
		parseString+=" "+getCurrentImage();

		int index = parseString.indexOf(".");
		if (index==-1)
		{
			CompileError error = new CompileError(getFileName(), getPreviousLine(), "illegal syntax: "+parseString);
			throw new JascoParseException(error);
		}

		String hookName = hookType.substring(0,index);
		HookInstance hi = getTraversalSpec().findInstanceForName(hookName);

		if (hi==null)
		{
			CompileError error = new CompileError(getFileName(), getPreviousLine(), "variable "+hookName+" not defined!");
			throw new JascoParseException(error);
		}

		String methodname = hookType.substring(index+1);
		if (methodname.equals("before")||methodname.equals("after")||methodname.equals("default")||methodname.equals(IHook.REPLACE))
		{
			parseHookBehaviorExec(hi,methodname);
			return;
		}

		getNextToken();
		while (getCurrentId()!=SEMICOLON)
		{
			parseString+=" "+getCurrentImage();
			getNextToken();
		}
		parseString+=" "+getCurrentImage();

		hi.addHookExecution(parseString+"//"+getPreviousLine());
	}

	protected void parseCombStrat() throws JascoParseException
	{

		String code = "addCombinationStrategy";
		while (getCurrentId()!=SEMICOLON)
		{
			code+=" "+getCurrentImage();
			getNextToken();
		} 
		code+=" "+getCurrentImage()+"//"+getPreviousLine();

		getTraversalSpec().addCombinationStrat(code);
	}

	public void parseHookInstances() throws JascoParseException
	{   

		getNextToken();
		if(getCurrentId()==RBRACE)
			return;	
		
		String hookType = getCurrentImage();
		getNextToken();
		while(getCurrentId()==DOT) {
			hookType +=getCurrentImage();
			getNextToken();
			hookType +=getCurrentImage();
			getNextToken();
		}
		String parseString=hookType;

		if (hookType.equals("addCombinationStrategy"))
		{
			parseCombStrat();
			parseHookInstances();
			return;
		}

		if (getCurrentId()==LPAREN)
		{
			parseHookExecution(parseString);
			parseHookInstances();
			return;
		}

		String newHookType = isAspect(hookType,getTraversalSpec().getImports());
		
		if (newHookType==null)
		{
			CompileError error = new CompileError(getFileName(), getPreviousLine(), "Aspect not found error : "+ hookType);
			throw new JascoParseException(error);
		}


		checkIDENTIFIER(parseString);
		parseString+=" "+getCurrentImage();
		String hookVar = getCurrentImage();

		getNextToken();
		checkASSIGN(parseString);
		parseString+=getCurrentImage();

		getNextToken();
		checkNEW(parseString);
		parseString+=getCurrentImage();

		getNextToken();
		String type2 = accumulateType(this.getTraversalSpec().getImports());
		parseString+=" "+type2+getCurrentImage();

		String newType2=isAspect(type2,getTraversalSpec().getImports());
		
		if (newType2==null)
		{
			CompileError error = new CompileError(getFileName(), getPreviousLine(), "Aspect not found error : "+ hookType);
			throw new JascoParseException(error);
		}


		if (!newType2.equals(newHookType))
		{
			CompileError error = new CompileError(getFileName(), getPreviousLine(), "incompatible types: "+ type2 + " assigned to "+hookType);
			throw new JascoParseException(error);
		}

		checkLPAREN(parseString);


		HookInstance hi = new HookInstance(newType2,hookVar);

		getNextToken();
		HookInit init = parseHookInit(parseString);
		hi.addHookInit(init);

		getNextToken();
		if (getCurrentId()==LBRACE)
		{
			parseAbstractMethodImpl(parseString,hi);
			getNextToken();
		}


		checkSEMICOLON("");

		getTraversalSpec().addHookInstance(hi);

		parseHookInstances();
	}

	protected void parseAbstractMethodImpl(String parseString, HookInstance hi) throws JascoParseException
	{

		checkLBRACE(parseString);

		getNextToken();
		PJavaClass clz  = new PJavaClass("TraversalClassRefinableMethods");
		
		while (getCurrentId() != RBRACE)
		{	
			parseJavaCode(clz,getTraversalSpec().getImports());
		}
		
		java.util.Iterator i = clz.getMethods(); 
		while(i.hasNext()) {
			PMethod method = (PMethod) i.next();
			hi.addAbstractMethod(new AbstractMethodImplementation(method.getMethodInfo()));
		}
	}

	protected HookInit parseHookInit(String prev) throws JascoParseException
	{

		String parsedString = prev;
		HookInit hookinit = new HookInit();

		while (getCurrentId()!=RPAREN)
		{
			//checkTYPE(parsedString);
			parsedString+=getCurrentImage();


				String type = "";
			HookMethod method = null;
			boolean visiting = true;
			if (getCurrentId()==VISITING)
			{
				getNextToken();
				String temptype = accumulateType(getTraversalSpec().getImports());
				parsedString+=" "+temptype;
				method = new HookMethod("*");
				type=temptype;
			}
			else
			{
				String returnType = getCurrentImage();

				getNextToken();
				String temptype = accumulateType(getTraversalSpec().getImports());

					parsedString+=" "+temptype;

				int index = temptype.lastIndexOf(".");
				if (index==-1)
				{
					CompileError error = new CompileError(getFileName(), getPreviousLine(), "non valid method: "+parsedString);
					throw new JascoParseException(error);
				}
				type = temptype.substring(0,index);
				method = new HookMethod(temptype.substring(index+1));
				visiting=false;
			}

			checkClass(type,getTraversalSpec().getImports());
			String result = getFullClassName(type, getTraversalSpec().getImports());
			if(result!=null) type=result;


			if (hookinit.getContextClass()!=null&&!hookinit.getContextClass().equals(type))
			{
				CompileError error = new CompileError(getFileName(), getPreviousLine(), "context classes for traversal input methods should be the same!"+type+" != "+hookinit.getContextClass());
				throw new JascoParseException(error);
			}
			hookinit.setContextClass(type);
			hookinit.addMethod(method);

			if (!visiting)
			{
				checkLPAREN(parsedString);
				parsedString+=getCurrentImage();

				getNextToken();
				while (getCurrentId()!=RPAREN)
				{
					String addtype = accumulateType(getTraversalSpec().getImports());
					method.addType(addtype);
					parsedString+=addtype;

					CompileError error = new CompileError(getFileName(), getPreviousLine(), " ) or , expected: "+parsedString);
					assertOrToken(getCurrentId(),COMMA,RPAREN,error);

					if (getCurrentId()==COMMA)
					{
						parsedString+=getCurrentImage();
						getNextToken();
					}
				}
				getNextToken();
			}

			CompileError error = new CompileError(getFileName(), getPreviousLine(), " ) or , expected: "+parsedString);
			assertOrToken(getCurrentId(),COMMA,RPAREN,error);
			if (getCurrentId()==COMMA)
			{
				parsedString+=getCurrentImage();
				getNextToken();
			}
		}

		if (!hookinit.getMethods().hasNext())
		{
			CompileError error = new CompileError(getFileName(), getPreviousLine(), "at least one hook  input method needed!");
			throw new JascoParseException(error);
		}


		return hookinit;	
	}

}
	
		
	
