/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;

public abstract class AbstractDialog extends JDialog {
    private boolean isCanceled = true;

    public AbstractDialog(JFrame parent, String title, int width, int height) {
        this(parent, title, width, height, true);
    }

    public AbstractDialog(JFrame parent, String title, int width, int height, boolean isModal) {
        super(parent, title, isModal);
        this.setSize(width, height);
        this.setResizable(false);
        this.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                quit();
            } 
        });
        this.centerOnScreen();
    }

    public void approveDialog() {
        isCanceled = false;
        dispose();
    } 

    public void cancelDialog() {
        dispose();
    } 

    public final void centerOnScreen() {
        Dimension frameSize = this.getSize();
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        this.setLocation(screenSize.width / 2 - frameSize.width / 2, screenSize.height / 2 - frameSize.height / 2);
    } 

    public boolean isCanceled() {
        return isCanceled;
    } 

    public void quit() {
    }

    protected final void showErrorDialog(String error) {
        this.showErrorDialog(error, "ERROR");
    } 

    protected final void showErrorDialog(String error, String title) {
        JOptionPane.showMessageDialog(this, error, title, JOptionPane.ERROR_MESSAGE);
    } 

    protected final void showInformationDialog(String error) {
        this.showInformationDialog(error, "Info");
    } 

    protected final void showInformationDialog(String error, String title) {
        JOptionPane.showMessageDialog(this, error, title, JOptionPane.INFORMATION_MESSAGE);
    } 

    protected final int showQuestionDialog(String error) {
        return this.showQuestionDialog(error, "Question : select an option");
    } 

    protected final int showQuestionDialog(String error, String title) {
        return JOptionPane.showConfirmDialog(this, error, title, JOptionPane.YES_NO_OPTION);
    } 

    protected final int showQuestionDialogWithCancel(String error) {
        return this.showQuestionDialog(error, "Question : select an option");
    } 

    protected final int showQuestionDialogWithCancel(String error, String title) {
        return JOptionPane.showConfirmDialog(this, error, title, JOptionPane.YES_NO_CANCEL_OPTION);
    } 

    protected final void showWarningDialog(String error) {
        this.showWarningDialog(error, "Warning");
    } 

    protected final void showWarningDialog(String error, String title) {
        JOptionPane.showMessageDialog(this, error, title, JOptionPane.WARNING_MESSAGE);
    } 

}


