/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util;

import java.util.TreeMap;
import java.util.Vector;

public class CommandLine  {
	
	private Vector args = new Vector();
	private String[] cl ;
	private int minArgs = 1;
	private TreeMap flags = new TreeMap();
	
	public CommandLine(String[] cl, int minArgs)  {
		this.cl=cl;
		this.minArgs=minArgs;
	}
	
	public CommandLine(String[] cl)  {
	
		this(cl,1);
	}
	
	public boolean parse()  {
		
		args=new Vector();
		flags=new TreeMap();
		for(int i=0; i<cl.length;++i) {
			
			if(isFlag(cl[i]))
				addFlag(cl[i]);
			else addArg(cl[i]);
		}
		
		return checkConsistency();	
	}
	
	public TreeMap getFlags()  {
	
		return flags;
	}
	
	public Vector getArgs()  {
	
		return args;
	}
	
	protected void addFlag(String s)  {
		
		try {
		int i=s.indexOf(":");
		String key = s.substring(1);
		String value="";
		if (i>0)  {
			key = s.substring(1,i);
			value = s.substring(i+1);
		}
		flags.put(key,value);
		}
		catch(RuntimeException ex)  {			
			System.out.println(s);
			throw ex;
		}
	}
	
	protected boolean checkConsistency()  {
	
		return args.size()>=minArgs;
	}
	
	protected boolean isFlag(String s)  {
	
		return s.startsWith("-");
	}
	
	protected void addArg(String s)  {
	
		args.add(s);
	}
}
		
		