/*
  Copyright (C) 2002-2005 JAsCo Development Team
  http://ssel.vub.ac.be/jasco

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation; either version 2 of the
  License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
 */
package jasco.util.generators;

import java.lang.reflect.Modifier;

public class VariableGenerator extends JavaGenerator {
    private String type;
    private String initialiser = "";

    public VariableGenerator(String type, String name) {
        super(name);
        this.type = type;
    }

    public String getType() {
        return type;
    } 

    public String getInitialiser() {
        return initialiser;
    } 

    public void setInitialiser(String initialiser) {
        this.initialiser = initialiser;
    } 

    public boolean hasInitialiser() {
        return !initialiser.equals("");
    } 

    public String toString() {
        String code = super.toString();
        code+=getAnnotationsString();
        if (Modifier.isPrivate(getModifiers()))
            code += "private ";
        else if (Modifier.isProtected(getModifiers()))
            code += "protected ";
        else if (Modifier.isPublic(getModifiers()))
            code += "public ";

        if (Modifier.isStatic(getModifiers()))
            code += "static ";

        if (Modifier.isFinal(getModifiers()))
            code += "final ";

        code += getType() + " ";
        code += getName();

        if (hasInitialiser())
            code += " = " + getInitialiser();

        code += ";";
        
        if(hasLineNo())
        	code+=" //"+getLineNo();

        return code;
    } 
	
	public String toTestString() {
        String code = super.toString();
        if (Modifier.isPrivate(getModifiers()))
            code += "private ";
        else if (Modifier.isProtected(getModifiers()))
            code += "protected ";
        else if (Modifier.isPublic(getModifiers()))
            code += "public ";

        if (Modifier.isStatic(getModifiers()))
            code += "static ";

        if (Modifier.isFinal(getModifiers()))
            code += "final ";
		
		int index = getType().lastIndexOf(".");
		String type  =getType();
		if (index!=-1)  {
			type=type.substring(index+1);
		}
		
        code += type + " ";
        code += getName();

        if (hasInitialiser())
            code += " = " + getInitialiser();

        code += ";";

        return code;
    } 

}


